/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include <inttypes.h>
#include "tranalyzer.h"
#include "basicLayer4CalcStatistics.h"
#include "packetCapture.h"
#include "main.h"

char* get_plugin_name() {
	return "basicLayer4CalcStatistics";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	pluginBasicLayer4CalcStatisticsIndex = pPluginIndex;
#ifdef __APPLE__
	mainHashMap = getMainHashMap();
#endif // __APPLE__
	// initialize the statistics variables
	basicStatsPerFlow = (basicStatsPerFlow_t*) malloc(sizeof(basicStatsPerFlow_t) * mainHashMap->hashChainTableSize);
	memset(basicStatsPerFlow, '\0', sizeof(basicStatsPerFlow_t) * mainHashMap->hashChainTableSize); // statistics for flows
}

binary_value_t* printHeader() {
	binary_value_t *bl4_main_bv = NULL;
	binary_value_t *bv = bv_new_bv("Number of transmitted packets", "numPktsSnt", 0, 1, bt_uint_64);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Number of received packets", "numPktsRcvd", 0, 1, bt_uint_64);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Number of transmitted bytes", "numBytesSnt", 0, 1, bt_uint_64);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Number of received bytes", "numBytesRcvd", 0, 1, bt_uint_64);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

#if BASIC_PKT_STATS != 0
	bv = bv_new_bv("Minimum layer3 packet size", "minPktSz", 0, 1, bt_uint_16);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Maximum layer3 packet size", "maxPktSz", 0, 1, bt_uint_16);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Average packet load ratio", "avePktSize", 0, 1, bt_float);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);
	
	bv = bv_new_bv("Send packets per second", "pktps", 0, 1, bt_float);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Send bytes per second", "bytps", 0, 1, bt_float);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Packet stream asymmetry", "pktAsm", 0, 1, bt_float);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);

	bv = bv_new_bv("Byte stream asymmetry", "bytAsm", 0, 1, bt_float);
	bl4_main_bv = bv_append_bv(bl4_main_bv, bv);
#endif // BASIC_PKT_STATS

	return bl4_main_bv;
}


void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	memset(&basicStatsPerFlow[flowIndex], '\0', sizeof(basicStatsPerFlow_t));
	
#if BASIC_PKT_STATS != 0
	// init variables that record a minimum
	basicStatsPerFlow[flowIndex].minL3PacketSize = 65535;
#endif // BASIC_PKT_STATS
}


// basicLayer4CalcStatistics is called whenever we received an IP packet and are processing it on the layer 4 level
void claimLayer4Information(packet_t *packet, unsigned long flowIndex) { 

	basicStatsPerFlow_t* basicStatsFlowP = &basicStatsPerFlow[flowIndex];

	uint16_t ipLength = packet->packetLength; 
	
	
#if BASIC_PKT_STATS != 0
	// check if packet is smaller than smallest packet we've observed so far
	if (ipLength < basicStatsFlowP->minL3PacketSize) {
		basicStatsFlowP->minL3PacketSize = ipLength;
	}

	// check if packet is larger than largest packet we've observed so far 
	if (ipLength > basicStatsFlowP->maxL3PacketSize) {
		basicStatsFlowP->maxL3PacketSize = ipLength;
	}
#endif // BASIC_PKT_STATS
	
	// update basic statistics
	basicStatsFlowP->numTransPackets++; // depends on frag and PACKETLENGTH status
	basicStatsFlowP->numTransBytes += ipLength;

#if FORCE_MODE == 1
	flow_t *flowP = &flows[flowIndex];
	if (basicStatsFlowP->numTransPackets > 10) { 
		flowP->status |= RMFLOW;
		rm_flows[num_rm_flows++] = flowP;
	}
#endif // FORCE_MODE == 1

}

// print the statistics for the flow at position flowIndex into the file flowFile
void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	flow_t *flowP = &flows[flowIndex];
	basicStatsPerFlow_t* basicStatsFlowP = &basicStatsPerFlow[flowIndex];

	int64_t tempNumOfPackets, tempNumOfBytes; // two vars to store the num of packets and bytes of the opposite flow

	// get info from opposite flow
	if (flowP->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		tempNumOfPackets = basicStatsPerFlow[flowP->oppositeFlowIndex].numTransPackets;
		tempNumOfBytes = basicStatsPerFlow[flowP->oppositeFlowIndex].numTransBytes;
	} else {
		tempNumOfPackets = 0;
		tempNumOfBytes = 0;
	}

#if BASIC_PKT_STATS != 0
	float packet_sym_ratio, byte_sym_ratio; // 2 vars for packet- and byte asymmetry calculation
	float packetsPerSec, bytesPerSec;
	float avePktSize;

	// calc per seconds stats
	if (flowP->duration.tv_sec == 0 && flowP->duration.tv_usec == 0) {
		packetsPerSec = 0.0f;
		bytesPerSec = 0.0f;
	} else {
		// calc packets per second
		packetsPerSec = (float)(basicStatsFlowP->numTransPackets) / ( (float)flowP->duration.tv_sec + (float)flowP->duration.tv_usec/1000000.0f );
		// calc bytes per second
		bytesPerSec = (float)(basicStatsFlowP->numTransBytes) / ( (float)flowP->duration.tv_sec + (float)flowP->duration.tv_usec/1000000.0f );
	}

	// calculate asymmetry of packets send and received
	if (tempNumOfPackets > 0) {
		packet_sym_ratio = ((float)basicStatsFlowP->numTransPackets - (float)tempNumOfPackets) / ((float)basicStatsFlowP->numTransPackets + (float)tempNumOfPackets);
	} else {
		packet_sym_ratio = 1.0f;
	}

	// calculate asymmetry of bytes send and received
	if (tempNumOfBytes > 0) {
		byte_sym_ratio = ((float)basicStatsFlowP->numTransBytes - (float)tempNumOfBytes) / ((float)basicStatsFlowP->numTransBytes + (float)tempNumOfBytes);
	} else {
		byte_sym_ratio = 1.0f;
	}

	if (basicStatsFlowP->numTransPackets > 0) {
		avePktSize = (float)basicStatsFlowP->numTransBytes / (float)basicStatsFlowP->numTransPackets;
	} else {
		avePktSize = 0;	
	}
#endif // BASIC_PKT_STATS

	outputBuffer_append(main_output_buffer, (char*) &basicStatsFlowP->numTransPackets, sizeof(uint64_t));
	outputBuffer_append(main_output_buffer, (char*) &tempNumOfPackets, sizeof(uint64_t));
	outputBuffer_append(main_output_buffer, (char*) &basicStatsFlowP->numTransBytes, sizeof(uint64_t));
	outputBuffer_append(main_output_buffer, (char*) &tempNumOfBytes, sizeof(uint64_t));
#if BASIC_PKT_STATS != 0
	outputBuffer_append(main_output_buffer, (char*) &basicStatsFlowP->minL3PacketSize, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &basicStatsFlowP->maxL3PacketSize, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &avePktSize, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &packetsPerSec, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &bytesPerSec, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &packet_sym_ratio, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &byte_sym_ratio, sizeof(float));
#endif // BASIC_PKT_STATS

}


void onApplicationTerminate() {
	if (basicStatsPerFlow) free(basicStatsPerFlow); basicStatsPerFlow = NULL;
}

