/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * descriptiveStatistics.h
 * calculates various statistics for a single flow
 */

#ifndef DESCRIPTIVESTATISTICS_H_
#define DESCRIPTIVESTATISTICS_H_

// includes
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <inttypes.h>

// local includes
#include "hashTable.h"
#include "redBlackTree.h"
#include "main.h"
#include "outputBuffer.h"
#include "packetSizeInterArrivalTimeHisto.h"
#include "binaryValue.h"
#include "getter.h"

// User defines
#define ENABLE_PACKETSIZE_CALCULATION 1 // enables the calculation of the statistics for the packet sizes. Default is one.
#define ENABLE_IAT_CALCULATION 1 // enables the calculation of the statistics for the inter arrival times. Default is zero.

/*
 * The structure for the basic descriptive statistics for each target.
 * A target can be, for example, the bytelength, the inter arrival time for each
 * packet
 * the descriptive statistics for each target are:
 */
typedef struct dStats_s {
	uint64_t numOfPackets; // the number of packets used to calculate these statistics
	float minimum;				// the minimum value
	float maximum;				// the maximum value
	float mean;					// the mean value
	float lowerQuartile;		// the lower quartile
	float median;				// the median value
	float upperQuartile;		// the upper quartile
	float iqd;					// the inter quartile distance = the distance between the lower and the upper quartile
	float mode;					// the mode = the value with the most occurrence
	float range;				// the range = maximum value - minimum value
	float stddev;				// the standard deviation of the values
	float stdrob;				// the robust standard deviation = the minimum of the standard deviation and the 0.7413'th iqd
	float skewness;				// the skewness of the values
	float excess;				// the excess kurtosis (= kurtosis - 3) of the values
} dStats_t;

/* variables */
dStats_t dStats_packetLengths, dStats_IATs;
packetSizeInterArrivalTimeTrees_t *dStats_currTree;
long counter;
unsigned long pluginDescriptiveStatisticsIndex;

// vars for calc the quartiles
float quartilePosLower, quartilePosMedian, quartilePosUpper;
unsigned long carryOverLower, carryOverMedian, carryOverUpper; // vars to carry over the target value if a quartile lies in between two target values and therefore between values of two nodes in a tree
float carryOverLowerIAT, carryOverMedianIAT, carryOverUpperIAT; // vars to carry over the target value for IAT
float carryOverVariance;
unsigned long actNumPacketsForMode; // a var to store the act maximum Pnum of packets for determining the mode
unsigned long dStatsBuffer_numPackets; // a var to store some calc'ed vals
unsigned long *dStats_actNumPackets;
long dStats_i; // a var for loops
long dStats_actBin, dStats_actNumPacketsIAT;
float dStats_temp; // a temp variable for multiple use

#ifdef __APPLE__
//global variable 
extern packetSizeInterArrivalTimeTrees_t* packetSizeInterArrivalTime_trees;
extern psiat_val_t* psiat_vals;
extern redBlackTreeNodePool_t* packetSizeInterArrivalTime_treeNodePool;
extern outputBuffer_t *main_output_buffer;
#endif // __APPLE__
/* functions */
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
char* get_dependencies();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

#endif /* DESCRIPTIVESTATISTICS_H_ */
