/*
 * entropy.c
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 * Co-Author: Mischa Lehmann, RDNIZ
 */

#include "entropy.h"
#include <inttypes.h>

collectedPayloadPerFlow_t *collectedPayloadPerFlow;


static int entropy_calc(uint32_t num_bytes,ENT_DATATYPE *byte_count,float *retentropy);

char* get_plugin_name() {
	return "entropy";
}

char* get_plugin_version() {
	return "0.5.6";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	pluginEntropyIndex = pPluginIndex;
#ifdef __APPLE__
	mainHashMap = getMainHashMap();
#endif // __APPLE__
	size_t data_size = sizeof(collectedPayloadPerFlow_t) * mainHashMap->hashChainTableSize;

#if ENT_MK_FRAGS > 0
	if(data_size > ENT_MAX_DATA_SIZE){
		printf("entropy - initialize(): Data field to large! (%d kbytes)\n",(int)(data_size/1024));
		exit(-1);
	}
#endif // ENT_MK_FRAGS

	//reserve memory for collected flow data
	collectedPayloadPerFlow = (collectedPayloadPerFlow_t*) malloc(data_size);
	memset(collectedPayloadPerFlow, '\0', sizeof(collectedPayloadPerFlow_t) * mainHashMap->hashChainTableSize); // statistics for flows
}

binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;

	binary_value_t *act_bv = bv_new_bv("Payload entropy", "PyldEntropy", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Payload_Character ratio", "PyldChRatio", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

//	act_bv = bv_new_bv("Pyld_Char_nonChar_ratio", "PyldChnChRatio", 0, 1, bt_float);
//	main_bv = bv_append_bv(main_bv, act_bv);
        
	act_bv = bv_new_bv("Payload_Binary ratio", "PyldBinRatio", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

#if ENT_ALPHA_D == 1
	act_bv = bv_new_bv("Payload length", "Pyldlen", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Payload histogram", "PyldHisto", 1, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // ENT_ALPHA_D == 1

#if ENT_MK_FRAGS > 0
	act_bv = bv_new_bv("Entropy bins", "EntropyBins", 0, (uint32_t)ENT_MAX_FRAG_COUNT, ENT_ARGS);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif
	return main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
	memset(&collectedPayloadPerFlow[flowIndex], 0x0, sizeof(collectedPayloadPerFlow_t));
}

void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	uint16_t i, snapIPPayloadLen=0, offset = 0;
	const uint8_t *pld;

	collectedPayloadPerFlow_t *collectedPayloadFlowPtr = &collectedPayloadPerFlow[flowIndex];

	//Layer-4-Header
	
	pld = packet->layer7Header;
	snapIPPayloadLen = packet->snapL7Length;

	for ( i = 0; i < snapIPPayloadLen; i++) {
#if ENT_MK_FRAGS > 0
		if (collectedPayloadFlowPtr->frag_size >= ENT_FRAG_SIZE){ // is fragment full?
			// calculate entropy
			if ( entropy_calc(collectedPayloadFlowPtr->frag_size, collectedPayloadFlowPtr->frag_binCount, &(collectedPayloadFlowPtr->frag_entropy[collectedPayloadFlowPtr->frag_count])) < 1 ) {
				fprintf(stderr, "entropy - claimLayer4Information(): failed entropy calculation on full frame!\n");
				exit(-1);
			}
			collectedPayloadFlowPtr->frag_count++;	// goto next fragment
			collectedPayloadFlowPtr->frag_size = 0; // reset fragment-size

			// rest byte_count
			memset(collectedPayloadFlowPtr->frag_binCount, 0x0, ENT_MAXPBIN*sizeof(ENT_DATATYPE));
		}

		if (collectedPayloadFlowPtr->frag_count < ENT_MAX_FRAG_COUNT) { // has still entropy fragments to fill
			collectedPayloadFlowPtr->frag_binCount[pld[i]]++;
			collectedPayloadFlowPtr->frag_size++;
		}
#endif
		collectedPayloadFlowPtr->binCount[pld[i]]++; // Increment counter for overall-entropy!
	}

	collectedPayloadFlowPtr->numBytes += snapIPPayloadLen; // only the true snaped payload
}

void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	int i;
	uint32_t actCount, numBytesTotal, pChar = 0, pBinary = 0;
	float p, ratio = 65535, crp = -1.0, brp = -1, entropy = 0.0;

	collectedPayloadPerFlow_t *collectedPayloadFlowPtr = &collectedPayloadPerFlow[flowIndex];
	numBytesTotal = collectedPayloadFlowPtr->numBytes;

	if ( numBytesTotal > 15) { // sqr(m) = 16, m: alphabet = 256
		for ( i = 0; i < ENT_MAXPBIN; i++ ) {
			actCount = collectedPayloadFlowPtr->binCount[i];
			if (actCount) {
				p = (float)actCount / (float)numBytesTotal;
				entropy += p * log(p);
				if ( i == 10 || i == 13 || ( i >= 31 && i <= 176 ) ) pChar += actCount;
				if ( i < 10 ) pBinary += actCount;
			}
		}
		entropy /= -log(ENT_MAXPBIN); //  Normalize to base 256 

		crp = (float)pChar / (float)numBytesTotal;
		brp = (float)pBinary / (float)numBytesTotal;
//		if ( crp < 1.0 )  ratio = crp / ( 1.0 - crp ); 
	} else entropy = -1.0f;

	outputBuffer_append(main_output_buffer, (char*) &entropy, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &crp, sizeof(float));
//	outputBuffer_append(main_output_buffer, (char*) &ratio, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &brp, sizeof(float));

#if ENT_ALPHA_D == 1
	outputBuffer_append(main_output_buffer, (char*) &numBytesTotal, sizeof(uint32_t));

	actCount = ENT_MAXPBIN;	
	outputBuffer_append(main_output_buffer, (char*) &actCount, sizeof(uint32_t));
	for (i = 0; i < ENT_MAXPBIN; i++) { // print payload distribution
		actCount = collectedPayloadFlowPtr->binCount[i];
		outputBuffer_append(main_output_buffer, (char*) &actCount, sizeof(uint32_t));
        }
#endif // ENT_ALPHA_D == 1

#if ENT_MK_FRAGS > 0
	
	if ( collectedPayloadFlowPtr->frag_size > 0 ) { // is there data in the current fragment?
		//  yes then calculate last entropy-fragment..
		if ( entropy_calc(collectedPayloadFlowPtr->frag_size, collectedPayloadFlowPtr->frag_binCount, &(collectedPayloadFlowPtr->frag_entropy[collectedPayloadFlowPtr->frag_count])) < 0 ) {
			fprintf(stderr,"entropy - onFlowTerminate(): failed calculating entropy!\n"); // Can happen...
			exit(-1);
		}
		collectedPayloadFlowPtr->frag_count++;	// goto next fragment
		collectedPayloadFlowPtr->frag_size = 0; // reset fragment-size
	}

	for (i = 0; i < ENT_MAX_FRAG_COUNT; i++) { // print fragments

		if (i < collectedPayloadFlowPtr->frag_count) entropy = collectedPayloadFlowPtr->frag_entropy[i]; // is there a fragment for this index?
		else entropy = -1.0f;

		outputBuffer_append(main_output_buffer, (char*)&entropy, sizeof(float));
	}
#endif
}


static inline int entropy_calc(uint32_t num_bytes, ENT_DATATYPE *byte_count, float *retentropy) {
	unsigned int i;
	float entropy = 0.0f, p, actCount;

	if ( num_bytes < 16 ) return 0; // sqr(m) = 8, m: alphabet = 256

	for ( i = 0; i < ENT_MAXPBIN; i++ ) {
		actCount = byte_count[i];
		if (actCount) {
			p = (float)actCount / (float)num_bytes;

			entropy += p * log(p);
		}
	}

	entropy /= -log(ENT_MAXPBIN); //  byte entropy, ln(256)

	(*retentropy) = entropy;

	return 1;
}

void onApplicationTerminate() {
	if (collectedPayloadPerFlow) {
		free(collectedPayloadPerFlow); 
		collectedPayloadPerFlow = NULL;
	}
}
