/*
 * global_statistics.c
 *
 * Autor: Torben Ruehl, RUAG (SWITZERLAND) Ltd., RDNIZ
 *
 * Copyright: 2013 by RUAG (Switzerland) Ltd.
 */

#include "global_statistics.h"

struct timeval gs_timer;
uint64_t timer_el_cnt;
uint64_t gs_num_flows;
FILE *gs_file;

#define timer_check(a, b) ((a.tv_sec - b.tv_sec) * 1000000 + (a.tv_usec - b.tv_usec))

// Global variables 
#ifdef __APPLE__
extern hashMap_t* mainHashMap;
extern outputBuffer_t* main_output_buffer;
#endif // __APPLE__

// plugin variables
unsigned long plugin_global_statistics_index;

void gs_write_out(uint64_t id,struct timeval timer, gs_stat_t gs_stats) {
    fprintf(gs_file, "\n%lu\t%lu.%06lu\t%lu\t%lu\t%lu\t%lu\t%lu\t%lu\t%lu", id,timer.tv_sec, timer.tv_usec, gs_stats.packets_sec, gs_stats.bytes_sec, gs_stats.flows_started, gs_stats.flows_stopped, gs_stats.current_flows_min, gs_stats.current_flows_max, gs_stats.timer_changes);
}

void add_time(struct timeval *timer, uint64_t value) {
    timer->tv_sec += value / 1000000;
    timer->tv_usec += value % 1000000;
    if (timer->tv_usec >= 1000000) {
#ifdef GS_DEBUG
        printf("overflow! %lu ->", timer->tv_usec);
#endif
        timer->tv_usec -= 1000000;
        timer->tv_sec += 1;
#ifdef GS_DEBUG
        printf(" %lu\n", timer->tv_usec);
#endif
    }
}

// local functions
void check_timer(struct timeval *timer, struct timeval act_time) {
    if (timer->tv_sec) {
        int64_t elapsed_time = timer_check(act_time, (*timer));
        while (elapsed_time >= GS_SAMPLE_TIME) {

            timer_el_cnt +=1;
#ifdef GS_DEBUG
            fprintf(stdout, "timer elapsed! id:%lu, timer was: %lu:%06lu, time is: %lu:%06lu\n", timer_el_cnt, timer->tv_sec, timer->tv_usec, act_time.tv_sec, act_time.tv_usec);
#endif
            // Write the stuff out
            gs_write_out(timer_el_cnt, *timer, gs_stats);

            memset(&gs_stats, 0, sizeof(gs_stat_t)); // reset the stats struct

            // fill max and min vals with currents
            gs_stats.current_flows_min = gs_num_flows;
            gs_stats.current_flows_max = gs_num_flows;


            add_time(timer, GS_SAMPLE_TIME);
            elapsed_time = timer_check(act_time, (*timer));
        }
    }
    else {
        timer->tv_sec = act_time.tv_sec;
        timer->tv_usec = act_time.tv_usec;

#ifdef GS_DEBUG
        printf("setting timer to %lu:%06lu\n", timer->tv_sec, timer->tv_usec);
#endif
    }
}

// Tranalyzer listener functions
char* get_plugin_name() {
    return "global_statistics";
}

char* get_plugin_version() {
    return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
    return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
    return 5;
}


void initialize(unsigned long pPluginIndex) {
    plugin_global_statistics_index = pPluginIndex;
#ifdef __APPLE__
    mainHashMap = getMainHashMap();
    baseFileName = getBaseFileName();
    specialOutput = getSpecialOutput();
#endif // __APPLE__
    char gs_file_name[MAX_FILENAME_LEN], *temp;

    memset(&gs_stats, 0, sizeof(gs_stat_t));

    gs_timer.tv_sec = 0;
    gs_timer.tv_usec = 0;

    timer_el_cnt = 0;
    gs_num_flows = 0;

    if (baseFileName) {
	temp = baseFileName;
    } else if (captureFileName) {
		temp = captureFileName;
    } else if (captureListFileName) {
		temp = captureListFileName;
    } else if (captureInterface) {
		temp = captureInterface;
    } else terminate();

    strncpy(gs_file_name, temp, strlen(temp)+1);
    strncat(gs_file_name, "_gs.txt", MAX_FILENAME_LEN-strlen(temp));

    if ((gs_file = fopen(gs_file_name, "wt")) == NULL) {
        fprintf(stderr, "ERROR! Couldn't open global statistics file '%s'!!!\n", gs_file_name);
        exit(EXIT_FAILURE);
    }

    fprintf(gs_file, "id\ttime\tpackets\tbytes\tflows_started\tflows_stopped\tflows_min\tflows_max\ttimeout_changes");
}


void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
#ifdef __APPLE__
    flows = getFlows();
#endif // __APPLE__
    flow_t *FlowP = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
    ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
    ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

    gs_stats.flows_started += 1;

    gs_num_flows += 1;
    if (gs_num_flows > gs_stats.current_flows_max) gs_stats.current_flows_max = gs_num_flows;

}


void claimLayer2Information(packet_t* packet) {
    // check timer
    check_timer(&gs_timer, packet->pcapHeader->ts);
    gs_stats.packets_sec += 1;
    gs_stats.bytes_sec += packet->pcapHeader->caplen;


}


/*void claimLayer3Information(packet_t* packet) {
#if IPV6_ACTIVATE == 1
    ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
    ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

    // your code

}*/


void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
#ifdef __APPLE__
    flows = getFlows();
#endif // __APPLE__
    flow_t *flow_p = &flows[flowIndex];

#if IPV6_ACTIVATE == 1
    ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
    ipHeader_t *ip_header = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

}

void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
    flows = getFlows();
    main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
    flow_t *flow_p = &flows[flowIndex];

    gs_stats.flows_stopped += 1;
    gs_num_flows -= 1;
    if (gs_num_flows < gs_stats.current_flows_min) gs_stats.current_flows_min = gs_num_flows;
}

void onApplicationTerminate() {
    timer_el_cnt++;
    gs_write_out(timer_el_cnt, gs_timer, gs_stats);
}

