/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#include "tranalyzer.h"
#include "icmpDecode.h"

char* get_plugin_name() {
	return "icmpDecode";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	pluginICMPDecodeIndex = pPluginIndex;
	mainHashMap = getMainHashMap();

	numEchoRequests = 0;
	numEchoReplies = 0;
	numSourceQuench = 0;
	numTraceroutes = 0;
	numOfICMPPackets = 0;
	memset(numDestUnreach, '\0', sizeof(numDestUnreach));
	memset(numTimeExceeded, '\0', sizeof(numTimeExceeded));
	memset(numRedirect, '\0', sizeof(numRedirect));

	memset(num_icmp6, '\0', sizeof(num_icmp6));

	icmpDecodeFlow = malloc(sizeof(icmpDecodeFlow_t) * mainHashMap->hashChainTableSize);
	memset(icmpDecodeFlow, '\0', sizeof(icmpDecodeFlow_t) * mainHashMap->hashChainTableSize);
	
}


binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;

#if IPV6_ACTIVATE == 1
	binary_value_t *act_bv = bv_new_bv("ICMP Aggregated type H (>128), L(<32) & code bit field", "icmpTypH_TypL_Code", 0, 3, bt_hex_32, bt_hex_32, bt_hex_16);
#else // IPV6_ACTIVATE == 0
	binary_value_t *act_bv = bv_new_bv("ICMP Aggregated type (<32) & code bit field", "icmpType_Code", 0, 2, bt_hex_32, bt_hex_16);
#endif // IPV6_ACTIVATE	
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("ICMP Echo reply/request success ratio", "icmpEchoSuccRatio", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	return main_bv;
}



void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	icmpDecodeFlow_t *icmpDecodeP = &icmpDecodeFlow[flowIndex];
	memset(icmpDecodeP, '\0', sizeof(icmpDecodeFlow_t));
}



void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
#if IPV6_ACTIVATE == 1
	if (packet->layer4Type != 58) return; // Only allow ICMP messages to pass through here
	else if ( packet->ip6FragHdr && (packet->ip6FragHdr->frag_off & FRAG6ID_N) ) return; // only icmp and 1. frag packet will be processed	
#else // IPV6_ACTIVATE == 0
	if ( packet->layer4Type != 1 ) return; // Only allow ICMP messages to pass through here
	else if ( (((ipHeader_t*)packet->layer3Header)->ip_off) & FRAGID_N ) return; // only icmp and 1. frag packet will be processed	
#endif // IPV6_ACTIVATE	

	uint8_t icmpType = packet->layer4Header->icmpHeader.type;
	uint8_t icmpCode = packet->layer4Header->icmpHeader.code;
	icmpDecodeFlow_t *icmpDecodeP = &icmpDecodeFlow[flowIndex];
	
	numOfICMPPackets++; // count only unfragmented icmp packets, only useful for troubleshooting
	
#if IPV6_ACTIVATE == 1
	if (icmpType < 160 && icmpCode < ICMP6_CODEFIELD) { 
		if (icmpType < ICMP6_TYPEFIELD ) icmpDecodeP->icmp_type_bfieldL |= (1 << icmpType); 
		else icmpDecodeP->icmp_type_bfieldH |= (1 << (icmpType - 128)); 

		icmpDecodeP->icmp_code_bfield |= (1 << icmpCode); 
	}

	if ( icmpType < 255 && icmpCode < 8) {
		num_icmp6[icmpType][icmpCode]++;
	}
	if ( icmpType == 128 ) { 
		icmpDecodeP->echoRequestsSent++;
		numEchoRequests++;
	}
	else if ( icmpType == 129 ) {
		icmpDecodeP->echoRepliesSent++;
		numEchoReplies++;
	}
 
#else // IPV6_ACTIVATE == 0

	if (icmpType > ICMP4_TYPEFIELD-1 || icmpCode > ICMP4_CODEFIELD-1) return; // largest icmp code is ICMP_PREC_CUTOFF or NR_ICMP_UNREACH with a size of 15
	icmpDecodeP->icmp_type_bfieldL |= (1 << icmpType); 
	icmpDecodeP->icmp_code_bfield |= (1 << icmpCode); 

//	count code

	switch (icmpType) {
		case ICMP4_ECHO:
			numEchoRequests++;
			icmpDecodeP->echoRequestsSent++;
			break;
		case ICMP4_ECHOREPLY:
			numEchoReplies++;
			icmpDecodeP->echoRepliesSent++;
			break;
		case ICMP4_SOURCE_QUENCH:
			numSourceQuench++;
			break;
		case ICMP4_DEST_UNREACH:
			numDestUnreach[icmpCode]++;
			break;
		case ICMP4_TIME_EXCEEDED:
			numTimeExceeded[icmpCode]++;
			break;
		case ICMP4_REDIRECT:
			numRedirect[icmpCode]++;
			break;
		case ICMP4_TRACEROUTE:
			numTraceroutes++;
			break;
	}
#endif // IPV6_ACTIVATE	
}



void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	float tmp = 0;
	unsigned long revFlowIndex = flows[flowIndex].oppositeFlowIndex;
	icmpDecodeFlow_t *icmpDecodeP = &icmpDecodeFlow[flowIndex];
	icmpDecodeFlow_t *icmpDecodePO = &icmpDecodeFlow[revFlowIndex];
	
#if IPV6_ACTIVATE == 1
	outputBuffer_append(main_output_buffer, (char*) &icmpDecodeP->icmp_type_bfieldH, sizeof(uint32_t));
#endif // IPV6_ACTIVATE	
	outputBuffer_append(main_output_buffer, (char*) &icmpDecodeP->icmp_type_bfieldL, sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &icmpDecodeP->icmp_code_bfield, sizeof(uint16_t));

	if ( revFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) {
		if ( icmpDecodeP->echoRequestsSent != 0 ) tmp = (float)icmpDecodePO->echoRepliesSent / (float)icmpDecodeP->echoRequestsSent;
		else if ( icmpDecodePO->echoRepliesSent != 0 ) tmp = -1.0f * (float)icmpDecodePO->echoRepliesSent;
	}
	outputBuffer_append(main_output_buffer, (char*) &tmp, sizeof(float));
}



void onApplicationTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	baseFileName = getBaseFileName();
	numOfPackets = *(getNumOfPackets());
#endif // __APPLE__
	float tmp = 0;
	FILE *icmpStatsFile;
	char icmpStatsFileName[MAX_FILENAME_LEN+1], *temp; // filename of the icmp statistics file

	if (baseFileName) {
                temp = baseFileName;
        } else if (captureFileName) {
                temp = captureFileName;
        } else if (captureListFileName) {
                temp = captureListFileName;
        } else if (captureInterface) {
                temp = captureInterface;
	}

	strncpy(icmpStatsFileName, temp, MAX_FILENAME_LEN);
	strncat(icmpStatsFileName, ICMP_SUFFIX, sizeof(ICMP_SUFFIX));
	
	// open protocol statistics file
	if ((icmpStatsFile = fopen(icmpStatsFileName, "w")) == NULL) {
		fprintf(stderr, "icmpDecode: Error, Can't open file %s in write mode\n", icmpStatsFileName);
		terminate();
	}

	if ( numEchoRequests ) tmp = (float)numEchoReplies / (float)numEchoRequests;

	fprintf(icmpStatsFile, "Total # of ICMP messages: \t%lu\n", numOfICMPPackets);
	fprintf(icmpStatsFile, "ICMP / Total traffic percentage[%%]: \t%5.3f\n", 100.0f*(float)numOfICMPPackets/(float)numOfPackets);
	fprintf(icmpStatsFile, "Echo reply / request ratio: \t%5.3f\n\n", tmp);

	fprintf(icmpStatsFile, "Type\tCode\t# of Messages\tRelative Frequency [%%]\n");
	
#if IPV6_ACTIVATE == 0
	fprintf(icmpStatsFile, "%s\t - \t%lu\t%5.3f\n", "ICMP_ECHOREQUEST", numEchoRequests, 100.0f*(float)numEchoRequests/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t - \t%lu\t%5.3f\n", "ICMP_ECHOREPLY", numEchoReplies, 100.0f*(float)numEchoReplies/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t - \t%lu\t%5.3f\n", "ICMP_SOURCE_QUENCH", numSourceQuench, 100.0f*(float)numSourceQuench/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t - \t%lu\t%5.3f\n", "ICMP_TRACEROUTE", numTraceroutes, 100.0f*(float)numTraceroutes/(float)numOfICMPPackets);
	
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_NET_UNREACH", numDestUnreach[0], 100.0f*(float)numDestUnreach[0]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_HOST_UNREACH", numDestUnreach[1], 100.0f*(float)numDestUnreach[1]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_PROT_UNREACH", numDestUnreach[2], 100.0f*(float)numDestUnreach[2]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_PORT_UNREACH", numDestUnreach[3], 100.0f*(float)numDestUnreach[3]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_FRAG_NEEDED", numDestUnreach[4], 100.0f*(float)numDestUnreach[4]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_SR_FAILED", numDestUnreach[5], 100.0f*(float)numDestUnreach[5]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_NET_UNKNOWN", numDestUnreach[6], 100.0f*(float)numDestUnreach[6]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_HOST_UNKNOWN", numDestUnreach[7], 100.0f*(float)numDestUnreach[7]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_HOST_ISOLATED", numDestUnreach[8], 100.0f*(float)numDestUnreach[8]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_NET_ANO", numDestUnreach[9], 100.0f*(float)numDestUnreach[9]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_HOST_ANO", numDestUnreach[10], 100.0f*(float)numDestUnreach[10]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_NET_UNR_TOS", numDestUnreach[11], 100.0f*(float)numDestUnreach[11]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_HOST_UNR_TOS", numDestUnreach[12], 100.0f*(float)numDestUnreach[12]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_PKT_FILTERED", numDestUnreach[13], 100.0f*(float)numDestUnreach[13]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_PREC_VIOLATION", numDestUnreach[14], 100.0f*(float)numDestUnreach[14]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_DEST_UNREACH", "ICMP_PREC_CUTOFF", numDestUnreach[15], 100.0f*(float)numDestUnreach[15]/(float)numOfICMPPackets);

	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_REDIRECT", "ICMP_REDIR_NET", numRedirect[0], 100.0f*(float)numRedirect[0]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_REDIRECT", "ICMP_REDIR_HOST", numRedirect[1], 100.0f*(float)numRedirect[1]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_REDIRECT", "ICMP_REDIR_NETTOS", numRedirect[2], 100.0f*(float)numRedirect[2]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_REDIRECT", "ICMP_REDIR_HOSTTOS", numRedirect[3], 100.0f*(float)numRedirect[3]/(float)numOfICMPPackets);

	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_TIME_EXCEEDED", "ICMP_EXC_TTL", numTimeExceeded[0], 100.0f*(float)numTimeExceeded[0]/(float)numOfICMPPackets);
	fprintf(icmpStatsFile, "%s\t%s\t%lu\t%5.3f\n", "ICMP_TIME_EXCEEDED", "ICMP_EXC_FRAGTIME", numTimeExceeded[1], 100.0f*(float)numTimeExceeded[1]/(float)numOfICMPPackets);

	fprintf(icmpStatsFile, "%s\t - \t%lu\t%5.3f\n", "ICMP_TRACEROUTE", numTraceroutes, 100.0f*(float)numTraceroutes/(float)numOfICMPPackets);

#else // IPV6_ACTIVATE == 1
	uint32_t i, j;
	for ( i = 0; i < 255 ; i++ )
		for ( j = 0; j < 8 ; j++ )
			if ( num_icmp6[i][j] ) 
				fprintf(icmpStatsFile, "%d\t%d\t%lu\t%5.3f\n",i ,j , num_icmp6[i][j], 100.0f*(float)num_icmp6[i][j]/(float)numOfICMPPackets);
#endif // IPV6_ACTIVATE

	fclose(icmpStatsFile);
	
	if ( icmpDecodeFlow ) free(icmpDecodeFlow);
	icmpDecodeFlow = NULL;
	
}

