/*      openDPIwrapper.c
*       Torben Rühl, 	Mischa Lehmann,	        Stefan Burschka
*       		ducksource@duckpond.ch
*       Version:0.5.6
*
*	Opendpi-Tranalyzer plugin
*       Require:
*		- see header
*
*
*	Licence:
*       You're allowed to edit and publish my source in all of your free and open-source projects.
*       Please send me an e-mail if you'll implement this source in one of your commercial or proprietary projects.
*       Leave this Header untouched!
*
*       Warranty:
*       Warranty void if signet is broken
*       =================== / /===================
*       [       Warranty   / /   Signet          ]
*       ==================/ /=====================
*	!!NOWAY!!
*
*/

/*==============================*/
/*             TODO's           */
/*==============================*/
/*
*
*
*
*
*
*/

/*==============================*/
/*           includes           */
/*==============================*/

#include "openDPIwrapper.h"

/*==============================*/
/*enums declaration + definition*/
/*==============================*/

/*==============================*/
/*     global declarations      */
/*==============================*/

static struct ipoque_detection_module_struct *ipoque_wrapper_struct;
static u32 openDPI_tick_resolution = 1000; // use millisecond resolution in OpenDPI
static unsigned int sizeof_id_struct; // size of the ipoque_id_struct

// mapping
static hashMap_t *openDPI_flowHashMap = NULL;

#if OPENDPI_MAKE_STATS > 0
static unsigned long detected_protocols[IPOQUE_MAX_SUPPORTED_PROTOCOLS];	// count of flows of specific protocol
static unsigned long detected_protocols_size[IPOQUE_MAX_SUPPORTED_PROTOCOLS];	// count of flow-data for specific protocol
static unsigned long detected_protocols_size_l4[IPOQUE_MAX_SUPPORTED_PROTOCOLS];// count of flow-data for specific protocol (l4 payload only)
#endif

#if OPENDPI_OUTPUT_AS_STRING != 0 ||  OPENDPI_MAKE_STATS > 0
static char *prot_long_str[] = {IPOQUE_PROTOCOL_LONG_STRING};			// long stringname of protocol
#endif

// DEBUG
#if OPENDPI_DEBUG > 0
static IPOQUE_PROTOCOL_BITMASK debug_messages_bitmask;
#endif

/*==============================*/
/*    function declarations     */
/*==============================*/

/*
*	funcname(args): <desc>
*	Arguments:
*		- <arg1> : <desc1> 
*/

/*
*	odpi_mk_hl_helper: make hashtable lookup struct
*	Arguments:
*		- flowIndex : index of flow for lookup
*		- hl_helper : the returing helper-struct
*/
static int odpi_mk_hl_helper(unsigned long flowIndex,openDPI_hash_lookup_helper_t *hl_helper);

/*
*	malloc_wrapper: the opendpi-malloc wrapper
*	Arguments:
*		- size : size of malloc bytes
*/
static void *malloc_wrapper(unsigned long size);

/*
*	free_wrapper: the opendpi-free wrapper
*	Arguments:
*		- freeable : the field for freeing
*/
static void free_wrapper(void *freeable);

#if OPENDPI_DEBUG > 0
/*
*	debug_printf: opendpi debug function
*	Arguments:
*		- id : id of flow
*		- log_level : the ipoque log level
*		- format : the format of printstring
*		- ... : arguments for printing
*/
static void debug_printf(u32 protocol, void *id_struct, ipq_log_level_t log_level, const char *format, ...);
#endif

/*==============================*/
/*     function definitions     */
/*==============================*/

char* get_plugin_name() {
	return "openDPIwrapper";
}

char* get_plugin_version() {
	return "0.5.6";
}

char* get_dependencies(){
#if OPENDPI_MAKE_STATS > 0
	return "basicLayer4CalcStatistics";
#else
	return "";
#endif
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	IPOQUE_PROTOCOL_BITMASK all;
	unsigned int i;

	// set openDPI protocol detection to all
#if OPENDPI_DEBUG > 0
	IPOQUE_BITMASK_SET_ALL(debug_messages_bitmask);
	ipoque_wrapper_struct = ipoque_init_detection_module(openDPI_tick_resolution, malloc_wrapper, debug_printf);
#else
	ipoque_wrapper_struct = ipoque_init_detection_module(openDPI_tick_resolution, malloc_wrapper, NULL);
#endif

	if(ipoque_wrapper_struct == NULL){
		fprintf(stderr,"ERROR: OpenDPI -initialization failed\n");
		exit(EXIT_FAILURE);
	}

	// enable all protocols in openDPI
	IPOQUE_BITMASK_SET_ALL(all);
	ipoque_set_protocol_detection_bitmask2(ipoque_wrapper_struct, &all);

	// init hashmap
	openDPI_flowHashMap = hashTable_initialize(1.0, sizeof(openDPI_hash_lookup_helper_t));
	openDPI_flowIndex = calloc(mainHashMap->hashChainTableSize,sizeof(unsigned long));
	openDPI_flows = (openDPI_flow *)calloc(mainHashMap->hashChainTableSize,sizeof(openDPI_flow)); // malloc space for the flows..
	
	sizeof_id_struct = ipoque_detection_get_sizeof_ipoque_id_struct(); // get size of id-struct

	for(i=0;i<mainHashMap->hashChainTableSize;i++){
		// malloc space for each flow
		openDPI_flows[i].ipoque_flow = (struct ipoque_flow_struct*)malloc(ipoque_detection_get_sizeof_ipoque_flow_struct()); // a flow-state machine
		openDPI_flows[i].lower_ip_states = (struct ipoque_id_struct*)malloc(sizeof_id_struct); // a lower ip state-machine
		openDPI_flows[i].upper_ip_states = (struct ipoque_id_struct*)malloc(sizeof_id_struct); // a upper ip state-machine
	}

#if OPENDPI_MAKE_STATS > 0
	memset(&(detected_protocols),0x00, IPOQUE_MAX_SUPPORTED_PROTOCOLS * sizeof(unsigned long));
	memset(&(detected_protocols_size),0x00, IPOQUE_MAX_SUPPORTED_PROTOCOLS * sizeof(unsigned long));
	memset(&(detected_protocols_size_l4),0x00, IPOQUE_MAX_SUPPORTED_PROTOCOLS * sizeof(unsigned long));
#endif

}

binary_value_t* printHeader() {
	binary_value_t *dpi_main_bv = NULL;

#if OPENDPI_OUTPUT_AS_STRING != 0
	dpi_main_bv = bv_append_bv(dpi_main_bv, bv_new_bv("Open DPI classification", "openDPI", 0, 1, bt_string_class));
#else
	dpi_main_bv = bv_append_bv(dpi_main_bv, bv_new_bv("Open DPI classification", "openDPI", 0, 1, bt_uint_32));
#endif

#if OPENDPI_MAKE_NN_OUTPUT > 0
	dpi_main_bv = bv_append_bv(dpi_main_bv, bv_new_bv("Open DPI NN-Training", "NNopenDPI", 1, 1, bt_int_8)); // for each protocol a int_8 which is the protocol-class
#endif

	return dpi_main_bv;
}

void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {

	// check if already an entry exists, that would mean there is already an opposite flow
	if (flows[flowIndex].oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		openDPI_flowIndex[flowIndex] = openDPI_flowIndex[flows[flowIndex].oppositeFlowIndex]; // there is a opposite flow.. (map it)
	} else {
		// there is no opposite flow
		// fill hash lookup helper struct
		openDPI_hash_lookup_helper_t hl_helper;
		memset(&hl_helper,0x0,sizeof(openDPI_hash_lookup_helper_t)); // do this because of valgrind.. XD
		odpi_mk_hl_helper(flowIndex, &hl_helper);

		// lookup the hash table
		if ((openDPI_flowIndex[flowIndex] = hashTable_lookup(openDPI_flowHashMap, (char*) &hl_helper)) == HASHTABLE_ENTRY_NOT_FOUND){ // no entry found..
			openDPI_flowIndex[flowIndex] = hashTable_insert(openDPI_flowHashMap, (char*) &hl_helper); // not found, create a new one			
			openDPI_flow *flow = &(openDPI_flows[openDPI_flowIndex[flowIndex]]); // get flow..

#if OPENDPI_DEBUG > 0
			printf("Create new flow: flowIndex:%u -> OpenDPI:%u\n",flowIndex,openDPI_flowIndex[flowIndex]);
#endif

			// init
			flow->protocol = IPOQUE_PROTOCOL_UNKNOWN; // reset protocol
#if OPENDPI_MAKE_STATS > 0
			flow->flow_size = 0;
#endif
			memset(flow->ipoque_flow,0x0,sizeof(struct ipoque_flow_struct));
			memset(flow->lower_ip_states,0x0,sizeof(struct ipoque_id_struct));
			memset(flow->upper_ip_states,0x0,sizeof(struct ipoque_id_struct));
		} else {
			printf("onFlowGenerated(): Impossible State!\n");
		}
	}
}

void claimLayer4Information(packet_t *packet, unsigned long flowIndex){
	void *src, *dst;
	openDPI_flow *flow = &(openDPI_flows[openDPI_flowIndex[flowIndex]]);

	u64 timeStamp = ((uint64_t) packet->pcapHeader->ts.tv_sec) * openDPI_tick_resolution + packet->pcapHeader->ts.tv_usec / (1000000 / openDPI_tick_resolution);
	u64 payload_len = packet->snapL2Length; // sizeof ethernet payload
#if OPENDPI_MAKE_STATS > 0
	flow->flow_size += payload_len; // increment size of flow
#endif

	// get subscriber state machines
	if (packet->layer3Header->ipHeader.ip_src.s_addr < packet->layer3Header->ipHeader.ip_dst.s_addr) {
		src = (void*)flow->lower_ip_states;
		dst = (void*)flow->upper_ip_states;
	} else {
		src = (void*)flow->upper_ip_states;
		dst = (void*)flow->lower_ip_states;
	}

	if(payload_len > 0){
		flow->protocol = ipoque_detection_process_packet(
			ipoque_wrapper_struct, // the ipoque dataholding struct
			flow->ipoque_flow, // the ipoque_flow
			(u_char*)&(packet->layer3Header->ipHeader),
			payload_len,
			timeStamp,
			src,
			dst);
#if OPENDPI_DEBUG > 0
	}else{
		fprintf(stdout, "Flow[%lu]: snaplen - etherheader <= 0!\n",flowIndex);
#endif
	}

#if OPENDPI_DEBUG > 0
	if (flow->protocol) {
		char *ip_as_char;
		if (packet->layer3Header->ipHeader.ip_src.s_addr < packet->layer3Header->ipHeader.ip_dst.s_addr) {
			ip_as_char = inet_ntoa(packet->layer3Header->ipHeader.ip_dst);
			fprintf(stdout, "Found protocol: %s->", ip_as_char);

			ip_as_char = inet_ntoa(packet->layer3Header->ipHeader.ip_src);
			fprintf(stdout, "%s: %"PRIu16"\n", ip_as_char, flow->protocol);
		} else {
			ip_as_char = inet_ntoa(packet->layer3Header->ipHeader.ip_src);
			fprintf(stdout, "Found protocol: %s->", ip_as_char);

			ip_as_char = inet_ntoa(packet->layer3Header->ipHeader.ip_dst);
			fprintf(stdout, "%s: %"PRIu16"\n", ip_as_char, flow->protocol);
		}

		fprintf(stdout, "packetNo: %lu, indexes: %lu", numOfPackets, flowIndex);
		if (flows[flowIndex].oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
			fprintf(stdout, " && %lu\n\n", flows[flowIndex].oppositeFlowIndex);
		} else {
			fprintf(stdout, "\n\n");
		}
	}
#endif
}


void onFlowTerminate(unsigned long flowIndex) {
	openDPI_hash_lookup_helper_t hl_helper;
	openDPI_flow *flow = &(openDPI_flows[openDPI_flowIndex[flowIndex]]);
	// get protocolnumber
	uint32_t protocol_num = flow->protocol;
	unsigned long hashtable_remove_ret = 0;

#if OPENDPI_MAKE_NN_OUTPUT > 0
	unsigned int i;
	uint32_t max_protos = IPOQUE_MAX_SUPPORTED_PROTOCOLS;
#endif

#if OPENDPI_OUTPUT_AS_STRING != 0
	char *tmpstr;
#endif

	if (protocol_num > IPOQUE_LAST_IMPLEMENTED_PROTOCOL){
		protocol_num = 0; // protocol unknown
#if OPENDPI_MAKE_STATS > 0
	}else{
		basicStatsPerFlow_t *flowl4stats = &(basicStatsPerFlow[flowIndex]); // this is why we need basicLayer4Statustucs.h

		detected_protocols[protocol_num]++;
		detected_protocols_size[protocol_num]+=flow->flow_size;
		detected_protocols_size_l4[protocol_num]+=flowl4stats->numTransBytes;
	}
#else
	} // no else!
#endif

#if OPENDPI_OUTPUT_AS_STRING != 0
	tmpstr = malloc(strlen(prot_long_str[protocol_num])+1);
	memcpy(tmpstr,prot_long_str[protocol_num],strlen(prot_long_str[protocol_num]));
	tmpstr[strlen(prot_long_str[protocol_num])] = '\0';
	outputBuffer_append(main_output_buffer,tmpstr, strlen(prot_long_str[protocol_num])+1);
	free(tmpstr);
#else
	outputBuffer_append(main_output_buffer, &(protocol_num), sizeof(uint32_t));
#endif

	// fill hash lookup helper struct
	memset(&hl_helper,0x0,sizeof(openDPI_hash_lookup_helper_t)); // do this because of valgrind.. XD
	odpi_mk_hl_helper(flowIndex,&hl_helper);

	if (flows[flowIndex].oppositeFlowIndex == HASHTABLE_ENTRY_NOT_FOUND) { // no opposite flow -> remove that flow from hashtable
		hashtable_remove_ret = hashTable_remove(openDPI_flowHashMap, (char*) &hl_helper);  // do remove
	} else {
		if (flowIndex < flows[flowIndex].oppositeFlowIndex) { // if this is the first seen flow remove that one too.
			hashtable_remove_ret = hashTable_remove(openDPI_flowHashMap, (char*) &hl_helper);
		}
	}
	
	if(hashtable_remove_ret == HASHTABLE_ENTRY_NOT_FOUND){
		fprintf(stderr,"openDPIwrapper - onFlowTerminate(): Cant remove hashtable entry!\n");
		return;
	}

#if OPENDPI_MAKE_NN_OUTPUT > 0
	outputBuffer_append(main_output_buffer, &max_protos, sizeof(uint32_t)); // repeating count

	for(i=0;i<max_protos;i++){
		int8_t match = OPENDPI_NN_MATCH;
		int8_t nomatch = OPENDPI_NN_NOMATCH;


		if(protocol_num == i){ // protocol matches
			outputBuffer_append(main_output_buffer, (char*)&(match), sizeof(int8_t));
		}else{ // no not this protocol
			outputBuffer_append(main_output_buffer, (char*)&(nomatch), sizeof(int8_t));
		}
	}
#endif
}

void onApplicationTerminate() {
	char openDPIFileName[MAX_FILENAME_LEN+1]; // filename of the flows file
	unsigned int i;

#if OPENDPI_MAKE_STATS > 0
	/* Write statistic file */
	// file
	FILE *openDPIFile;

	// setup output file name
	strncpy(openDPIFileName, baseFileName, MAX_FILENAME_LEN);
	strncat(openDPIFileName, "_openDPI.txt", MAX_FILENAME_LEN-strlen(baseFileName));

	if ((openDPIFile = fopen(openDPIFileName, "w")) == NULL) {
			fprintf(stderr, "Error: Can't open file %s in write mode\n", openDPIFileName);
	} else {
		// count detected Protocols:
		int i;
		unsigned long num_of_protocols = 0, size_of_protocols = 0, size_of_protocols_l4 = 0;
		unsigned long num_of_protocols_wun = 0, size_of_protocols_wun = 0, size_of_protocols_l4_wun; // num and sizeof protocols without unknown
		for (i = 0; i < IPOQUE_MAX_SUPPORTED_PROTOCOLS; i++) {
			num_of_protocols += detected_protocols[i];
			size_of_protocols += detected_protocols_size[i];
			size_of_protocols_l4 += detected_protocols_size_l4[i];
		}

		// get size and num of protocols without unknown
		num_of_protocols_wun = num_of_protocols - detected_protocols[0];
		size_of_protocols_wun = size_of_protocols - detected_protocols_size[0];
		size_of_protocols_l4_wun = size_of_protocols_l4 - detected_protocols_size_l4[0];

		fprintf(openDPIFile,
		"Open DPI detected protocols\n"
		"---------------------------\n\n"
		"Identified Flows: %lu "
		"[%lu bytes l2] "
		"[%lu bytes l4] "
		"(%.2f%% of all Flows) "
		"(%.2f%% bytes of l2-Data) "
		"(%.2f%% bytes of l4-Data) "
		"\n\n",
		num_of_protocols_wun,
		size_of_protocols_wun,
		size_of_protocols_l4_wun,
		(float) (num_of_protocols_wun / (float) num_of_protocols) * 100,
		(float) (size_of_protocols_wun / (float) size_of_protocols) * 100,
		(float) (size_of_protocols_l4_wun / (float) size_of_protocols_l4) * 100);

		fprintf(openDPIFile,"No.\tName\tQuantity(flows)\tQuantity(bytes l2)\tProbability(flows)\tProbability(bytes l2)\tProbability(bytes l4)\n");

		// Print out detected protocols
		for (i = 0; i < IPOQUE_MAX_SUPPORTED_PROTOCOLS; i++) {
			if (detected_protocols[i] > 0) {
				fprintf(openDPIFile, "\n%d\t%s\t%lu\t%lu\t%f\t%f\t%f",
				i,
				prot_long_str[i],
				detected_protocols[i],
				detected_protocols_size[i],
				(float) detected_protocols[i] / (float) num_of_protocols,
				(float) detected_protocols_size[i] / (float) size_of_protocols,
				(float) detected_protocols_size_l4[i] / (float) size_of_protocols_l4);
			}
		}
	}
#endif
	ipoque_exit_detection_module(ipoque_wrapper_struct, free_wrapper);

	free(openDPI_flowHashMap);

	for(i=0;i<mainHashMap->hashChainTableSize;i++){
		free(openDPI_flows[i].ipoque_flow);
		free(openDPI_flows[i].lower_ip_states);
		free(openDPI_flows[i].upper_ip_states);
	}
	free(openDPI_flows);

	free(openDPI_flowIndex);
}

static int odpi_mk_hl_helper(unsigned long flowIndex, openDPI_hash_lookup_helper_t *hl_helper) {
#if IPV6_ACTIVATE == 1 
        ipAddr_t lower_ip, upper_ip;
#else // IPV6_ACTIVATE == 0
	in_addr_t lower_ip, upper_ip;
#endif // IPV6_ACTIVATE
	unsigned short lower_port;
	unsigned short upper_port;
	flow_t *actFlow = &flows[flowIndex];
	/*
	 * Be aware:
	 * an openDPI flow is a complete connection between the src and dst IP,
	 * that means BOTH directions, not only one direction as in tranalyzer.
	 * Therefore, we have to check which is the lower and which is the upper IP.
	 */

#if IPV6_ACTIVATE == 1
	if (actFlow->srcIP.IPv6L[0] < actFlow->dstIP.IPv6L[0]) {
		lower_ip = actFlow->srcIP;
		upper_ip = actFlow->dstIP;
	} else {
		lower_ip = actFlow->dstIP;
		upper_ip = actFlow->srcIP;
	}
#else // IPV6_ACTIVATE == 0
	if (actFlow->srcIP.s_addr < actFlow->dstIP.s_addr) {
		lower_ip = actFlow->srcIP.s_addr;
		upper_ip = actFlow->dstIP.s_addr;
	} else {
		lower_ip = actFlow->dstIP.s_addr;
		upper_ip = actFlow->srcIP.s_addr;
	}

#endif // IPV6_ACTIVATE

	if (actFlow->srcPort < actFlow->dstPort){
		lower_port = actFlow->srcPort;
		upper_port = actFlow->dstPort;
	} else {
		lower_port = actFlow->dstPort;
		upper_port = actFlow->srcPort;
	}

	hl_helper->lower_ip = lower_ip;
	hl_helper->upper_ip = upper_ip;
	hl_helper->vlan_id = actFlow->vlanID;
	hl_helper->lower_port = lower_port;
	hl_helper->upper_port = upper_port;
	hl_helper->layer_4_protocol = actFlow->layer4Protocol;

	return 1;
}

static void *malloc_wrapper(unsigned long size) {
	return malloc(size);
}

static void free_wrapper(void *freeable) {
	free(freeable);
}

#if OPENDPI_DEBUG > 0
static void debug_printf(u32 protocol, void *id_struct, ipq_log_level_t log_level, const char *format, ...){
        if (IPOQUE_COMPARE_PROTOCOL_TO_BITMASK(debug_messages_bitmask, protocol) != 0) {
                const char *protocol_string;
                const char *file;
                const char *func;
                u32 line;
                va_list ap;
                va_start(ap, format);
		char *prot_short_str[] = { IPOQUE_PROTOCOL_SHORT_STRING };

                protocol_string = prot_short_str[protocol];

                ipoque_debug_get_last_log_function_line(ipoque_wrapper_struct, &file, &func, &line);

                printf("OpenDPI-DEBUG> %s:%s:%u Prot: %u, level: %u:", file, func, line, protocol_string,
                           log_level);

                vprintf(format, ap);
		printf("\n");

                va_end(ap);
        }
}
#endif
