/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * packetSizeInterArrivalTimeHisto.c
 */


/* includes */
#include "tranalyzer.h"
#include "main.h"

/* local includes */
#include "packetSizeInterArrivalTimeHisto.h"

void rekursiveDestroyIATTree(redBlackNode_t *node, redBlackTreeNodePool_t *treeNodePool);


char* get_plugin_name() {
	return "packetSizeInterArrivalTimeHisto";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

// new flexible float bin definition
/*
uint32_t iat2binf( float iat ) {
	int32_t i;
	float f;
	for ( i = 0; i < IATSECMAX; i++ ) {
		f = iat - IATBinBf[i];
		if ( f > 0.0f ) return f * IATBinWif[i] + IATBinNfu[i];
	}
	return IATSECMAX;
}


float bin2iatf(uint32_t bin) {
	int32_t i;
	for (i = 0; i < IATSECMAX; i++) {
                if ( bin < IATBinNu[i+1] ) return (bin - IATBinBf[i]) * IATBinWf[i] + IATBinBf[i];
	}
	return IATBinNu[IATSECMAX];
}
*/

// flexible uint bins

uint32_t iat2bin( struct timeval iat ) {
	uint32_t i, k;
	k = iat.tv_sec * IATNORM + iat.tv_usec / IATNORM;
	for (i = 0; i < IATSECMAX; i++) {
		if ( k < IATBinBu[i+1] ) return (k - IATBinBu[i]) / IATBinWu[i] + IATBinNu[i];
	}
	return IATBinNu[IATSECMAX];
}

int32_t bin2iat(uint32_t bin) {
	uint32_t i;
	for (i = 0; i < IATSECMAX; i++) {
                if ( bin < IATBinNu[i+1] ) return (bin - IATBinNu[i]) * IATBinWu[i] + IATBinBu[i];
	}
	return IATBinBu[IATSECMAX];
}


void printTree_inOrder(redBlackNode_t *tree) {
	
	if (tree->leftChild) printTree_inOrder(tree->leftChild);
	fprintf(stdout, "[%"PRId32":%"PRIu32"]\t", tree->value, psiat_vals[tree - &packetSizeInterArrivalTime_treeNodePool->nodePool[0]].numOfPackets);
	if (tree->rightChild) printTree_inOrder(tree->rightChild);
}

void initialize(unsigned long pPluginIndex) {
#ifdef __APPLE__
		mainHashMap = getMainHashMap();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	uint32_t i;
	packetSizeInterArrivalTimeHistoIndex = pPluginIndex;
	
#if PRINT_HISTO != 0 || PRINT_HISTO_IN_SEPARATE_FILE != 0
	psiat_output_buffer = outputBuffer_initialize(main_output_buffer->size_of_buffer);
#endif /* PRINT_HISTO != 0 || PRINT_HISTO_IN_SEPARATE_FILE != 0 */

	if ((packetSizeInterArrivalTime_treeNodePool = redBlackTree_initTreeNodePool(mainHashMap->hashChainTableSize * HISTO_NODEPOOL_FACTOR)) == NULL) {
		fprintf(stdout, "!!! ERROR: Couldn't initialize Tree Node Pool for packetSizeInterArrivalHisto plugin !!!\n");
		terminate();
	}

	packetSizeInterArrivalTime_trees = malloc(mainHashMap->hashChainTableSize * sizeof(packetSizeInterArrivalTimeTrees_t));
	psiat_vals = malloc(packetSizeInterArrivalTime_treeNodePool->pool_size * sizeof(psiat_val_t));

	memset(packetSizeInterArrivalTime_trees, '\0', mainHashMap->hashChainTableSize * sizeof(packetSizeInterArrivalTimeTrees_t));
	memset(psiat_vals, '\0', packetSizeInterArrivalTime_treeNodePool->pool_size * sizeof(psiat_val_t));

#if PRINT_HISTO_IN_SEPARATE_FILE != 0
	// setup output file name
	strncpy(histoFileName, baseFileName, MAX_FILENAME_LEN);
	strncat(histoFileName, HISTO_FILE_SUFFIX, MAX_FILENAME_LEN-strlen(baseFileName));

	// open histo output file
	if ((histoFile = fopen(histoFileName, "w")) == NULL) {
		fprintf(stderr, "Error: Can't open file %s in write mode! Maybe incorrect permissions?\n", histoFileName);
		terminate();
	}

	// print comment line in histo file
	fprintf(histoFile, "# Packet Size Inter Arrival Time Histogram file of %s", baseFileName);
#endif /* PRINT_HISTO_IN_SEPARATE_FILE != 0 */
/*
	IATBinNu[0] = 0;
	for (i = 1; i <= IATSECMAX; i++) {
		IATBinNu[i] = (IATBinBu[i] - IATBinBu[i-1]) / IATBinWu[i-1] + IATBinNu[i-1];
        }
*/

	
}

void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	packetSizeInterArrivalTimeTrees_t *pSIAT = &packetSizeInterArrivalTime_trees[flowIndex];
	// cleanup
	if (pSIAT->packetTree) {
		rekursiveDestroyIATTree( pSIAT->packetTree, packetSizeInterArrivalTime_treeNodePool );
		redBlackTree_destroy( pSIAT->packetTree, packetSizeInterArrivalTime_treeNodePool );
		memset( pSIAT, '\0', sizeof(packetSizeInterArrivalTimeTrees_t) );
	}
}

#if PRINT_HISTO == 1 
binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;
#if HISTO_PRINT_BIN == 1
	binary_value_t *act_bv = bv_new_bv( "Packetsize Inter Arrival Time bin histogram", "MSb:Ps_IatBin_Cnt_PsCnt_IatCnt", 1, 5, bt_uint_16, bt_uint_32, bt_uint_32, bt_uint_32, bt_uint_32);
#else
	binary_value_t *act_bv = bv_new_bv( "Packetsize min Inter Arrival Time of bin histogram", "MSt:Ps_Iat_Cnt_PsCnt_IatCnt", 1, 5, bt_uint_16, bt_uint_32, bt_uint_32, bt_uint_32, bt_uint_32);
#endif // HISTO_PRINT_BIN 
	main_bv = bv_append_bv(main_bv, act_bv);
	return main_bv;
}
#endif // PRINT_HISTO

void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
	redBlackNode_t *currentPacketNode, *currentIATNode;
	struct timeval currentIAT;
	int32_t i;
	uint8_t entryExists;

	packetSizeInterArrivalTimeTrees_t *pSIAT = &packetSizeInterArrivalTime_trees[flowIndex];

	currentPacketNode = redBlackTree_search_insert( pSIAT->packetTree, packetSizeInterArrivalTime_treeNodePool, (int32_t)packet->packetLength, INSERT_NODE, &entryExists );

	if (currentPacketNode == NULL) {
		fprintf(stderr, "Could not insert new tree node! Reason: ");
		if (packetSizeInterArrivalTime_treeNodePool->numOfFreeNodes == 0)
			fprintf(stderr, "No free treenodes left, try to increase the treenode pool size.\n");
		else
			fprintf(stderr, "Unknown\n");
		return;
	}

	pSIAT->numOfPackets++;
	currPacketTreeBucket = currentPacketNode - &packetSizeInterArrivalTime_treeNodePool->nodePool[0];
	if (!currentPacketNode->parent) pSIAT->packetTree = currentPacketNode;
#if REDBLACKTREE_ENABLE_ROTATION != 0
	else {
		// if the tree was rotated at its root, we have to change the information in the current packetSizeInterArrivalTime tree
		while( pSIAT->packetTree->parent) {
			pSIAT->packetTree = pSIAT->packetTree->parent;
		}
	}
#endif // REDBLACKTREE_ENABLE_ROTATION != 0

	if (entryExists == 0) {
		memset(&psiat_vals[currPacketTreeBucket], '\0', sizeof(psiat_val_t));
		psiat_vals[currPacketTreeBucket].numOfPackets = 1;
	} else {
		psiat_vals[currPacketTreeBucket].numOfPackets++;
	}

	// get IAT
	if ( pSIAT->lastPacketTime.tv_sec ) { // marker for flow start
		timersub(&packet->pcapHeader->ts, &pSIAT->lastPacketTime, &currentIAT);
	} else {
		currentIAT.tv_sec = 0;
		currentIAT.tv_usec = 0;
	}

	// update last packet seen time
	pSIAT->lastPacketTime.tv_sec = packet->pcapHeader->ts.tv_sec;
	pSIAT->lastPacketTime.tv_usec = packet->pcapHeader->ts.tv_usec;

	// store iat
	i = iat2bin(currentIAT);
	pSIAT->numOfpacketsInTimeBin[i]++;

	currentIATNode = redBlackTree_search_insert( psiat_vals[currPacketTreeBucket].iat_tree, packetSizeInterArrivalTime_treeNodePool, i, INSERT_NODE, &entryExists);

	if (currentIATNode == NULL) {
		fprintf(stderr, "Could not insert new tree node! Reason: ");
		if (packetSizeInterArrivalTime_treeNodePool->numOfFreeNodes == 0) fprintf(stderr, "No free treenodes left, try to increase the treenode pool size.\n");
		else fprintf(stderr, "Unknown\n");
		return;
	}

	currIATTreeBucket = currentIATNode - &packetSizeInterArrivalTime_treeNodePool->nodePool[0];
		if (!currentIATNode->parent)
			psiat_vals[currPacketTreeBucket].iat_tree = currentIATNode;
#if REDBLACKTREE_ENABLE_ROTATION != 0
		else {
			// if the tree was rotated at its root, we have to change the information in the current packet tree
			while(psiat_vals[currPacketTreeBucket].iat_tree->parent) {
				psiat_vals[currPacketTreeBucket].iat_tree = psiat_vals[currPacketTreeBucket].iat_tree->parent;
			}
		}
#endif // REDBLACKTREE_ENABLE_ROTATION != 0

		if (entryExists == 0) {
			memset(&psiat_vals[currIATTreeBucket], '\0', sizeof(psiat_val_t));
			psiat_vals[currIATTreeBucket].numOfPackets = 1;
		} else {
			psiat_vals[currIATTreeBucket].numOfPackets++;
		}
}

void rekursivePrintIAT(redBlackNode_t *node, int32_t packetSize, uint32_t numPacketsPS, packetSizeInterArrivalTimeTrees_t *tree, outputBuffer_t *buffer) {
	if (node->leftChild) rekursivePrintIAT(node->leftChild, packetSize, numPacketsPS, tree, buffer);

	currIATTreeBucket = node - &packetSizeInterArrivalTime_treeNodePool->nodePool[0];

	// print NO semicolon if it is the first bin printed out
	if (psiat_is_first == 0) outputBuffer_append(buffer, ";", 1);
	else psiat_is_first = 0;

#if HISTO_PRINT_BIN == 1
	buffer->act_pos += snprintf(&(buffer->buffer[buffer->act_pos]), buffer->size_of_buffer - buffer->act_pos,
			"%"PRId32"_%"PRId32"_%"PRIu32"_%"PRIu32"_%"PRIu32"",
			packetSize,
			node->value,
			psiat_vals[currIATTreeBucket].numOfPackets,
			numPacketsPS,
			tree->numOfpacketsInTimeBin[node->value]);
#else // HISTO_PRINT_BIN == 0
	buffer->act_pos += snprintf(&(buffer->buffer[buffer->act_pos]), buffer->size_of_buffer - buffer->act_pos,
			"%"PRId32"_%"PRId32"_%"PRIu32"_%"PRIu32"_%"PRIu32"",
			packetSize,
			bin2iat(node->value),
			psiat_vals[currIATTreeBucket].numOfPackets,
			numPacketsPS,
			tree->numOfpacketsInTimeBin[node->value]);
#endif // HISTO_PRINT_BIN
	if (node->rightChild) rekursivePrintIAT(node->rightChild, packetSize, numPacketsPS, tree, buffer);
}

void rekursivePrintIAT_binary(redBlackNode_t *node, int32_t packetSize, uint32_t numPacketsPS, packetSizeInterArrivalTimeTrees_t *tree, outputBuffer_t *buffer) {
	if (node->leftChild) rekursivePrintIAT_binary(node->leftChild, packetSize, numPacketsPS, tree, buffer);

	currIATTreeBucket = node - &packetSizeInterArrivalTime_treeNodePool->nodePool[0];

	psiat_counter++;
	outputBuffer_append(buffer, (char*) &packetSize, sizeof(uint16_t));

#if HISTO_PRINT_BIN == 1
	outputBuffer_append(buffer, (char*) &node->value, sizeof(uint32_t));
#else // HISTO_PRINT_BIN == 0
	uint32_t tempVar = bin2iat(node->value);
	outputBuffer_append(buffer, (char*) &tempVar, sizeof(uint32_t));
#endif // HISTO_PRINT_BIN

	outputBuffer_append(buffer, (char*) &psiat_vals[currIATTreeBucket].numOfPackets, sizeof(uint32_t));
	outputBuffer_append(buffer, (char*) &numPacketsPS, sizeof(uint32_t));
	outputBuffer_append(buffer, (char*) &tree->numOfpacketsInTimeBin[node->value], sizeof(uint32_t));

	if (node->rightChild) rekursivePrintIAT_binary(node->rightChild, packetSize, numPacketsPS, tree, buffer);
}

void rekursivePrintPacketSize(redBlackNode_t *node, packetSizeInterArrivalTimeTrees_t *tree, outputBuffer_t *buffer) {
	if (node->leftChild) rekursivePrintPacketSize(node->leftChild, tree, buffer);

	currPacketTreeBucket = node - packetSizeInterArrivalTime_treeNodePool->nodePool;
	rekursivePrintIAT(psiat_vals[currPacketTreeBucket].iat_tree, node->value, psiat_vals[currPacketTreeBucket].numOfPackets, tree, buffer);

	if (node->rightChild) rekursivePrintPacketSize(node->rightChild, tree, buffer);
}

void rekursivePrintPacketSize_binary(redBlackNode_t *node, packetSizeInterArrivalTimeTrees_t *tree, outputBuffer_t *buffer) {
	if (node->leftChild) rekursivePrintPacketSize_binary(node->leftChild, tree, buffer);

	currPacketTreeBucket = node - packetSizeInterArrivalTime_treeNodePool->nodePool;
	rekursivePrintIAT_binary(psiat_vals[currPacketTreeBucket].iat_tree, node->value, psiat_vals[currPacketTreeBucket].numOfPackets, tree, buffer);

	if (node->rightChild) rekursivePrintPacketSize_binary(node->rightChild, tree, buffer);
}


void rekursiveDestroyIATTree(redBlackNode_t *node, redBlackTreeNodePool_t *treeNodePool) {
	if (node->leftChild) rekursiveDestroyIATTree(node->leftChild, treeNodePool);
	if (node->rightChild) rekursiveDestroyIATTree(node->rightChild, treeNodePool);

	currPacketTreeBucket = node - packetSizeInterArrivalTime_treeNodePool->nodePool;
	redBlackTree_destroy(psiat_vals[currPacketTreeBucket].iat_tree, treeNodePool);
}


void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	packetSizeInterArrivalTimeTrees_t *pSIAT = &packetSizeInterArrivalTime_trees[flowIndex];

	if ( pSIAT->packetTree != NULL ) {

#if HISTO_DEBUG != 0
		redBlackTree_printTree( pSTAT->packetTree, 5);
		printTree_inOrder( pSTAT->packetTree );
		fprintf(stdout, "\n\n");
#endif // HISTO_DEBUG

#if PRINT_HISTO != 0 || PRINT_HISTO_IN_SEPARATE_FILE != 0

#endif /* PRINT_HISTO != 0 || PRINT_HISTO_IN_SEPARATE_FILE != 0*/

#if PRINT_HISTO == 1 
		psiat_counter = 0;
		// reset the psiat buffer
		outputBuffer_reset(psiat_output_buffer);
		// print in buffer
		rekursivePrintPacketSize_binary(pSIAT->packetTree, &packetSizeInterArrivalTime_trees[flowIndex], psiat_output_buffer);
		outputBuffer_append(main_output_buffer, (char*) &psiat_counter, sizeof(uint32_t));
		outputBuffer_append(main_output_buffer, psiat_output_buffer->buffer, psiat_output_buffer->act_pos);
#endif // PRINT_HISTO

#if PRINT_HISTO_IN_SEPARATE_FILE == 1
		// reset flag
		psiat_is_first = 1;
		// reset the psiat buffer
		outputBuffer_reset(psiat_output_buffer);
		// print in buffer
		rekursivePrintPacketSize(pSIAT->packetTree, pSIAT, psiat_output_buffer);
		// print in histo file
		fputs("\n", histoFile);
		fputs(psiat_output_buffer->buffer, histoFile);
#endif // PRINT_HISTO_IN_SEPARATE_FILE

#if HISTO_EARLY_CLEANUP != 0
		// cleanup
		rekursiveDestroyIATTree( pSIAT->packetTree, packetSizeInterArrivalTime_treeNodePool );
		redBlackTree_destroy( pSIAT->packetTree, packetSizeInterArrivalTime_treeNodePool );
		memset(pSIAT, '\0', sizeof(packetSizeInterArrivalTimeTrees_t));
#endif // HISTO_EARLY_CLEANUP


	} else fprintf(stdout, "ERROR: Flow with number %lu has no tree!\n", flowIndex);
}

void onApplicationTerminate() {
	if (packetSizeInterArrivalTime_treeNodePool->nodePool) free(packetSizeInterArrivalTime_treeNodePool->nodePool); packetSizeInterArrivalTime_treeNodePool->nodePool = NULL;
	if (packetSizeInterArrivalTime_treeNodePool) free(packetSizeInterArrivalTime_treeNodePool); packetSizeInterArrivalTime_treeNodePool = NULL;
	if (packetSizeInterArrivalTime_trees) free(packetSizeInterArrivalTime_trees); packetSizeInterArrivalTime_trees = NULL;
	if (psiat_vals) free(psiat_vals); psiat_vals = NULL;

#if (PRINT_HISTO == 1 || PRINT_HISTO_IN_SEPARATE_FILE == 1)
	outputBuffer_destroy(psiat_output_buffer);
#endif // (PRINT_HISTO == 1 || PRINT_HISTO_IN_SEPARATE_FILE == 1)

#if PRINT_HISTO_IN_SEPARATE_FILE == 1
	fclose(histoFile);
#endif // PRINT_HISTO_IN_SEPARATE_FILE
}

#ifdef __APPLE__
redBlackTreeNodePool_t* getPacketSizeInterArrivalTimeTreeNodePool(){
	return packetSizeInterArrivalTime_treeNodePool;
}

packetSizeInterArrivalTimeTrees_t* getPacketSizeInterArrivalTimeTrees(){
	return packetSizeInterArrivalTime_trees;
}

psiat_val_t* getPsiatVals(){
	return psiat_vals;
}
#endif // __APPLE__
