/*	regex_pattern_matcher.c
*	Mischa Lehmann
*	ducksource@duckpond.ch
*	Version:0.2
*
*	Tranalyzer regex matching engine
*	Require:
*		- see header
*
*
*	Licence:
*	You're allowed to edit and publish my source in all of your free and open-source projects.
*	Please send me an e-mail if you'll implement this source in one of your commercial or proprietary projects.
*	Leave this Header untouched!
*
*	Warranty:
*	Warranty void if signet is broken
*	=================== / /===================
*	[       Warranty   / /   Signet          ]
*	==================/ /=====================
*	!!NOWAY!!
*
*/

/*==============================*/
/*             TODO's           */
/*==============================*/
/*
*
*	- Default values for pattern have to be defines
*	- ATM we're using a default character table during pcre_compile (from locales) maybe we've to change this
*	- Make sure that pcre can be used in a multithreadded environment see:
*		From PCRE api doc...
*		MULTITHREADING
*			The PCRE functions can be used in  multi-threading  applications,  with
*			the  proviso  that  the  memory  management  functions  pointed  to  by
*			pcre_malloc, pcre_free, pcre_stack_malloc, and pcre_stack_free, and the
*			callout function pointed to by pcre_callout, are shared by all threads.
*
*			The  compiled form of a regular expression is not altered during match-
*			ing, so the same compiled pattern can safely be used by several threads
*			at once.
*
*
*
*/

/*==============================*/
/*           includes           */
/*==============================*/

#include "regex_pattern_matcher.h"

/*==============================*/
/*enums declaration + definition*/
/*==============================*/

/*==============================*/
/*     global declarations      */
/*==============================*/

/*==============================*/
/*    function declarations     */
/*==============================*/

/*
*	funcname(args): <desc>
*	Arguments:
*		- <arg1> : <desc1> 
*/

#if REGEX_DEBUG > 0
/*
*	dumpPatternById(regexData,pattern_id): dump (print) the pattern with the given id
*	Arguments:
*		- regexData : the regex holding struct
*		- pattern_id : the id of the pattern to be printed
*/
static int dumpPatternById(regexdata_t *regexData,unsigned int pattern_id);

/*
*	dumpPattern(regexData,pattern): dump (print) the given pattern 
*	Arguments:
*		- regexData : the regex holding struct
*		- pattern_id : the pattern for dumping
*/
static int dumpPattern(regexdata_t *regexData,pattern_t *pattern);
#endif

/*
*	stringReplace: Replace occurences of "search" with "replace"
*	Arguments:
*		- search : searchstring 
*		- replace : replacement string 
*		- string : pointer to string for relacement (sting has to be malloced)
*		- replaceall : true := replace all occurences, false:= only once
*/
static int stringReplace(char *search, char *replace, char **string, bool replaceall); 

/*
*	execCmd(pattern,rexFlow): Execute a command for flow 
*	Arguments:
*		- pattern : the pattern which is linked to the command
*		- rexFlow : regex flow information
*/
static int execCmd(pattern_t *pattern,rexFlow_t *rexFlow);

/*
*	getPattern(pattern_id,regexData,ret_pattern): Get the pattern represented with id=pattern_id
*	Arguments:
*		- pattern_id : the pattern's id to search for
*		- regexData : the regex holding struct
*		- ret_pattern : pointer to the returning pattern, if NULL then dont return any pattern (used for pattern existence checking)
*/
static inline int getPattern(unsigned int pattern_id, regexdata_t *regexData,pattern_t **ret_pattern);

/*
*	getPatternMask(pattern_id,regexData,ret_mask): Get the mask of the pattern with id=pattern_id
*	Arguments:
*		- pattern_id : the pattern's id to search
*		- regexData : the regex holding data struct
*		- ret_mask : pointer to the returning pattern
*/
static int getPatternMask(unsigned int pattern_id,regexdata_t* regexData,char **ret_mask);

/*
*	studyRegex(pattern,options): Study the regex pattern with the given options
*	Arguments:
*		- pattern : The pattern to study
*		- options : options for studying
*/
static int studyRegex(pattern_t *pattern,int options);

/*
*	compileRegex(pattern,options): Compile a regex rule with the provided options
*	Arguments:
*		- pattern : The pattern to compile
*		- options : The copmilation options
*/
static int compileRegex(pattern_t *pattern,int options);

/*
*	parseFile(fd,line_c,regexData,linebuf,linebuf_size): Parse a regex file
*	Arguments:
*		- fd : open filedescriptor to file
*		- line_c : amount of lines in the file
*		- regexData : the regex struct to be filled
*		- linebuf : a linebuffer with enough space for the biggest line + 1 (because of trailing '\0')
*		- linebuf_size : size of the linebuffer
*/
static int parseFile(FILE *fd, unsigned int line_c,regexdata_t *regexData,char *linebuf,unsigned int linebuf_size);

/*
*	loadFile(regexData): load a file and prepare for parsing
*	Arguments:
*		- regexData : the regex data struct to be filled
*/
static int loadFile(regexdata_t *regexData);

/*
*	matchPattern(rexFlow,data,data_len,regexData): try to match all the loaded patterns against data 
*	Arguments:
*		- rexFlow : holds the regex flow
*		- data : the data to match
*		- data_len : length of the data
*		- regexData : the regexData struct
*/
static int matchPattern(rexFlow_t *rexFlow,char *data,int data_len,regexdata_t *regexData);

/*
*	openFlowOutFile(rexFlow,regexData): open a output file for rexFlow
*	Arguments:
*		- rexFlow : the regex flow which is assigned to the output file
*		- regexData : the regex data struct
*/
static int openFlowOutFile(rexFlow_t *rexFlow,regexdata_t *regexData);

/*
*	writePacket(rexFlow,packet): write a packet to a flow output file
*	Arguments:
*		- rexFlow : the regex flow
*		- packet : the packet data to write
*		- regexData : the regex data struct
*/
static int writePacket(rexFlow_t* rexFlow, packet_t* packet,regexdata_t *regexData);


/*
*	Tranalyzer pluin functions
*/
char* get_plugin_name() {
	return REGEX_PLUGINNAME;
}

char* get_plugin_version() {
	return REGEX_PLUGIN_VERSION;
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

char* get_dependencies() {
	return "";
}


void initialize(unsigned long pPluginIndex){
	char *regex_file;
	size_t regexfile_path_len = strlen(getenv("HOME"))+strlen(REGEX_CONFIG_FILE) + 1 + 1; // +1 because of '/' +1 because of '\0'
	size_t flowfiles_path_len = strlen(getenv("HOME"))+strlen(REGEX_FLOWDATA_FILES) + 1 +1; // +1 because of '/' +1 because of '\0'

	// set plugin index
	plugin_REGEX_PLUGINNAME_Index = pPluginIndex;

	// init regexData with default values
	regexData.initialzed = false;
	regexData.rexFlowTable = NULL;
	regexData.patterns_c = 0;
	regexData.max_patterns_c = 0;
	regexData.patterns = NULL;
	regexData.mask_size = 0;
	regexData.flowfiles_path = NULL;
	regexData.flowfiles_path_len = 0;
	regexData.regexfile_path = NULL;
	regexData.regexfile_path_len = 0;

	// check if environment is correct
	if (strlen(getenv("HOME")) <= 0){
		printerr("initialize(): $HOME environment variable not set",0);
		return;
	}

	// malloc space for regex flows
	regexData.rexFlowTable = malloc(mainHashMap->hashChainTableSize * sizeof(rexFlow_t));
	if(regexData.rexFlowTable == NULL){
		printerr("initialize(): rexFlowTable malloc failed!",0);
		return;
	}

	//Build file & folderpaths

	// the regex-file
	// set pathlength of filename
	regexData.regexfile_path_len=regexfile_path_len;
	// copy together full pathname
	regexData.regexfile_path = malloc( (regexfile_path_len) * sizeof(char)); // malloc space 
	memset(regexData.regexfile_path,'\0',regexfile_path_len * sizeof(char));
	strncpy(regexData.regexfile_path, getenv("HOME"), regexfile_path_len); // copy home dir to path
	regexData.regexfile_path[strlen(regexData.regexfile_path)] = '/'; // append '/'
	strncat(regexData.regexfile_path, REGEX_CONFIG_FILE, strlen(REGEX_CONFIG_FILE)); // concatinate both strings
	regexData.regexfile_path[regexfile_path_len - 1] = '\0'; // \0 at the ond of the string

	// set pathlength of filename
	regexData.flowfiles_path_len=flowfiles_path_len;
	// the flowfile-output directory
	regexData.flowfiles_path = malloc( (flowfiles_path_len)*sizeof(char)); // malloc space
	memset(regexData.flowfiles_path,'\0',flowfiles_path_len * sizeof(char));
	strncpy(regexData.flowfiles_path, getenv("HOME"), flowfiles_path_len); // copy home dir to path
	regexData.flowfiles_path[strlen(regexData.flowfiles_path)] = '/'; // append '/'
	strncat(regexData.flowfiles_path, REGEX_FLOWDATA_FILES, strlen(REGEX_FLOWDATA_FILES)); // concatinate both strings
	regexData.flowfiles_path[flowfiles_path_len - 1] = '\0';

#if REGEX_DEBUG > 0
	// print pathes out
	printf("Regex file:%s\n",regexData.regexfile_path);
	printf("Flow files:%s\n",regexData.flowfiles_path);
#endif

	// make directory
	util_recursive_mkdir(regexData.flowfiles_path); 

	// load regex pattern
	if(loadFile(&(regexData)) < 0) {
		printerr("initialize(): File loading error\n",0);
		return;
	}

	// regexData is now initialized
	regexData.initialzed = true;
}

binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;
	// we only have on header at this time.
	main_bv = bv_append_bv(main_bv, bv_new_bv(REGEX_LONG_NAME,REGEX_SHORT_NAME,1,1,bt_uint_32));
	return main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex){
	unsigned int idstr_len;

	// only do something if initialized
	if(regexData.initialzed == false){
		return;
	}

	// get the current rexFlow
	rexFlow_t *rexFlow = &(regexData.rexFlowTable[flowIndex]);

	// init rexFlow with defaults
	rexFlow->idstr=NULL; 			// make idstr later
	rexFlow->mask = NULL;			// make mask field later
	rexFlow->dumpToFile=false;  		// dont dump this flow to a file per default
	rexFlow->patternMatchCount=0;   	// amount of matched patterns
	rexFlow->flowOutFile_init=false;	// output file not initialized
	rexFlow->flowOutFile_path=NULL;		// No path setted
	rexFlow->flowOutFile=NULL;		// No filedescriptor 

	// make id string
	idstr_len = (floor(flowIndex/10) +1 + 1); // get flowId-stringlength (+1 ; [0-9] char +1: \0 is strending)
	rexFlow->idstr = malloc(idstr_len*sizeof(char));
	memset(rexFlow->idstr,'\0',idstr_len); // memset to '\0'
	snprintf(rexFlow->idstr,idstr_len,"%lu",flowIndex); // convert to string

	// malloc mask field
	rexFlow->mask = malloc(regexData.mask_size);
	memset(rexFlow->mask,0x0,regexData.mask_size); // empty mask
}

void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	rexFlow_t *rexFlow; // this regex flow
	char *payloadStart = (char*)packet->layer7Header; // payload data TODO: Start matching l2 data
	unsigned int l7length = packet->snapL7Length; // only match l7 content

	// only do something if initialized
	if(regexData.initialzed == false){
		return;
	}

	// get current regex flow
	rexFlow = &(regexData.rexFlowTable[flowIndex]);

	// does the flow has data?
	if(l7length > 0){ // has l7 data
		// start matching all the rules..
		if(matchPattern(rexFlow,payloadStart,l7length,&(regexData))<0){
			printerr("claimLayer4Information(): pattern matching failed",0);
			return;
		}
	}

	// write the information to output file
	if(rexFlow->dumpToFile == true) { // as soon as the dump flag is set
		if(writePacket(rexFlow,packet,&(regexData))<0){
			printerr("claimLayer4Information(): Writing packet failed",0);
			return;
		}
	}
}

void onFlowTerminate(unsigned long flowIndex) {
	rexFlow_t *rexFlow;

	// only do something if initialized
	if(regexData.initialzed == false){
		return;
	}

	// the the current regex flow
	rexFlow = &(regexData.rexFlowTable[flowIndex]);

	// start writing output buffer
	// TODO: patternMatchCount is unsigned int but it should be something like uint32_t
	outputBuffer_append(main_output_buffer,(char*)&(rexFlow->patternMatchCount),sizeof(rexFlow->patternMatchCount)); // write repeating count 

	if(rexFlow->patternMatchCount > 0) { // has this flow matches
		unsigned int i;
		unsigned int patterns_written = 0; // count the number of patterns so that we can break out of loop and saave some time
		// check each pattern for command to execute
		// TODO: Make this faster via ptr in rexFlow_t or something similar
		for(i=0;i<regexData.patterns_c;i++){
			pattern_t *pattern=&(regexData.patterns[i]); // get next pattern

			if(util_cmp_bit_mask(rexFlow->mask,pattern->mask,regexData.mask_size) == true){ // this pattern matched

				// write pattern id to output buffer
				// TODO: pattern->id is unsigned int but it should be something like uint32_t
				outputBuffer_append(main_output_buffer,(char*)&(pattern->id),sizeof(pattern->id));

				if(pattern->command_init == true){ // this pattern has a command assigned
					if(execCmd(pattern,rexFlow)<0){ // execute command
						printerr("onFlowTerminate(): exeCmd failed",0);
						return;
					}
				}

				// one pattern more written
				patterns_written++;
				if(rexFlow->patternMatchCount <= patterns_written){ // all patterns written?
					break; // we can stop printing
				}
			}
		}
	}

	//free rexFlow-fields
	//TODO: only flag them somehow so we dont need to malloc again... 
	free(rexFlow->idstr);
	free(rexFlow->mask);
}

void onApplicationTerminate() {
	// only do something if initialized
	if(regexData.initialzed == false){
		return;
	}

	// TODO: CLEANUP
}

/*
* static functions
*/

#if REGEX_DEBUG > 0
static int dumpPatternById(regexdata_t *regexData,unsigned int pattern_id){
	pattern_t *pattern;

	// search for the pattern
	if(getPattern(pattern_id,regexData,&(pattern))<0){
		printerr("dumpPatternById(): failed getting pattern",0);
		return -1;
	}

	//dump it
	if(dumpPattern(regexData,pattern)<0){
		printerr("dumpPatternById(): failed dumping pattern",0);
		return -1;
	}

	return 1;	
}

static int dumpPattern(regexdata_t *regexData,pattern_t *pattern){

	// print it!
	// TODO: this is not really well performing as i'm calling printf for each line.. but i did this for formatting reasons -> see printf macro in utils.h
	printf(	"===== Pattern [%d] =====\n",	pattern->id);				 		// the pattern id
	printf(	"regex: [%u] : %s\n",		pattern->regex_len,pattern->regex); 			// print regex
	printf( "compRegex: %s\n",		(pattern->regexCompiled_init ? "True" : "False"));   	// print compiled regex flag 
	printf( "studied: %s\n",		(pattern->studyRegex_init ? "True" : "False")); 	// print study regex flag
	printf( "dodump : %s\n",		(pattern->dodump ? "True" : "False"));			// print dodump flag
	printf(	"cmd_init: %s\n",		(pattern->command_init ? "True" : "False"));		// print command init flag

	// print command if there is one:
	if(pattern->command_init == true){ // as a comand assigned
		printf("cmd [%u] : %s\n",pattern->command_len,pattern->command); // print cmd
	}

	// print mask
	printf("mask: ");
	util_dump_bit_mask(pattern->mask,regexData->mask_size);
	fprintf(stdout,"\n");

	// print prev_mask
	printf("prev_mask: ");
	util_dump_bit_mask(pattern->prev_mask,regexData->mask_size);
	fprintf(stdout,"\n");

	return 1;
}
#endif

static int stringReplace(char *search, char *replace, char **string, bool replaceall){
	char *str = (*string);
	char *replaceStart;
	int replace_add = strlen(replace) - strlen(search); // additional required memory for each replacement of the string

	// reaplace_add cannot be negative..
	if(replace_add < 0){
		replace_add = 0;
	}

#if REGEX_DEBUG > 0
	printf(	"Replacing: \"%s\" with \"%s\" [%s]\n",
			search,
			replace,
			(replaceall ? "all" : "once")
		);
#endif

	// repeat: replace all occurences of string? 
	do{
		// locate searchstring
		replaceStart = strstr(str, search);

		// searchstart found
		if( replaceStart != NULL ){
			unsigned int newstring_len = strlen(str)+replace_add + 1; // add more memory to replacestring
			unsigned int tempstring_len = strlen(replaceStart) - strlen(search) + 1; // size of tempstring +1 ('\0')
			unsigned int replacePos = replaceStart - str; // get position of replace start

			// malloc space
			char *tempString = (char*) malloc(tempstring_len * sizeof(char));
			char *newstring = (char*) malloc(newstring_len * sizeof(char));

			// malloc was successful?
			if(tempString == NULL || newstring == NULL) {
				printerr("stringReplace(): Malloc of tempbuffers failed",0);
				return -1;
			}

			memset(newstring,'\0',newstring_len); // memset newstring to all \0

			// start replacing the string..
			// first part goto newstring
			memcpy(newstring,str,replacePos); 
			memcpy(tempString,replaceStart+strlen(search),tempstring_len); // backup ending to tempstring

			// insert replacement
			memcpy(newstring+replacePos, replace,strlen(replace));
		
			// append temstring '\0' comes from tempString
			memcpy(newstring+replacePos+strlen(replace),tempString,tempstring_len);

			// cleanup
			free(tempString);
			free(str);
			str = newstring; // set newstring as str
		}
	}while( replaceStart != NULL // found a substring..
		&& replaceall == true ); // do i need to continue searching (replaceall)

	(*string) = str; // return the string

	return 1;
}

static int execCmd(pattern_t *pattern,rexFlow_t *rexFlow){
	char *cmd = malloc(pattern->command_len);

	//assemble command 
	//$$ as placeholder for filepath
	//## as placeholder for filenumber

	// copy command
	memcpy(cmd,pattern->command,pattern->command_len);

	// repalce ## in command..
	if(stringReplace("##",rexFlow->idstr,&cmd,true)<0){
		printerr("execCmd(): Replacing string failed",0);
		return -1;
	}

	// is there even a flow out file defined for this flow?
	if(rexFlow->flowOutFile_init == true){
		// and again for $$
		if(stringReplace("$$",rexFlow->flowOutFile_path,&cmd,true)<0){
			printerr("execCmd(): Replacing string failed",0);
			return -1;
		}
	}

#if REGEX_DEBUG > 0
	//Print command string..
	printf("Command:\"%s\"\n",cmd);
#endif

	// do execute command
	// TODO: I'm pretty sure that there is a better way than system
	//system(cmd);

	free(cmd);

	return 1;
}

static inline int getPattern(unsigned int pattern_id, regexdata_t *regexData,pattern_t **ret_pattern){
	unsigned int i;
	for(i=0;i<regexData->patterns_c;i++){ // try all the patterns
		pattern_t *pattern = &(regexData->patterns[i]); // get next pattern
		if(pattern_id == pattern->id){ // same id?
			if(ret_pattern != NULL){ // only return something if not NULL was provided as ret_pattern
				(*ret_pattern) = pattern; // return pattern
			}
			return 1; // return pattern found 
		}
	}

	return -1; // pattern not found
}

static int getPatternMask(unsigned int pattern_id,regexdata_t* regexData,char **ret_mask){
	// search for pattern with id
	pattern_t *pattern;
	if(getPattern(pattern_id,regexData,&pattern)<0){
		printerr("getPatternMask(): Pattern not found",0);
		return -1;
	}
	(*ret_mask) = pattern->mask; // assign mask

	return 1; // pattern found
}

static int studyRegex(pattern_t *pattern,int options){
	pcre_extra *newStudiedRegex;
	char *error_msg;

	// is there already a studied regex pattern
	if(pattern->studyRegex_init == true){
		pcre_free_study(pattern->studyRegex); // free the studied regex
		pattern->studyRegex_init = false; // no studied regex anymore
	}

	// study regex-pattern 
	newStudiedRegex = pcre_study(pattern->regexCompiled,options,(const char**)&error_msg);

	if(newStudiedRegex == NULL){
		printerr("studyRegex(): Regex studying failed",0);
		fprintf(stderr,	"message: %s\n", error_msg);
		return -1;
	}

	// assign everything to pattern
	pattern->studyRegex = newStudiedRegex; // assign regex to pattern
	pattern->studyRegex_init = true; // set initialzed flag to true

	return 1;
}

static int compileRegex(pattern_t *pattern,int options){
	pcre *newRegex;
	char *error_msg;
	int error_offset;

	// is there already a compiled regex pattern?
	if(pattern->regexCompiled_init == true){
		// free the old pattern using the pcre-free function
		(*pcre_free)(pattern->regexCompiled); // do the pcre-free
		pattern->regexCompiled_init = false; // no longer initialzed
	}

	// compile regex-pattern 
	// TODO: atm we're using a default character table maybe change this
	newRegex = pcre_compile(pattern->regex,options,(const char**)&(error_msg),&(error_offset),NULL);

	if(newRegex == NULL){
		printerr("compileRegex(): Regex compilation failed",0);
		fprintf(stderr,	"message: %s\n"
				"character: %d\n",
				error_msg,
				error_offset
			);
		return -1;
	}

	// assign everything to pattern
	pattern->regexCompiled = newRegex; // assign regex to pattern
	pattern->regexCompiled_init = true; // set initialzed flag to true

	return 1;
}

static int parseFile(FILE *fd, unsigned int line_c,regexdata_t *regexData,char *linebuf,unsigned int linebuf_size){
	unsigned int i;				// linecount
	bool in_pattern=false;			// has a rule started?
	pattern_t *curr_pattern=NULL;		// pointer to the current pattern
	unsigned int curr_pattern_num=0;	// the current pattern number 

	// general purpose variables which can be used for parsing
	int i1,i2,i3;
	unsigned int ui1,ui2,ui3;
	char* s1 = malloc(linebuf_size*sizeof(char));
	char* s2 = malloc(linebuf_size*sizeof(char));
	char* s3 = malloc(linebuf_size*sizeof(char));

	// parse all the lines
	for(i=0;i<line_c;i++){
		unsigned int j;

		// read line to buffer
		if(fgets((void*)linebuf,(size_t)linebuf_size,fd)!=linebuf){
			printerr("parseFile(): lineread failed",0);
			return -1;
		}

		// we are in a pattern
		if(in_pattern == true){
			curr_pattern = &(regexData->patterns[curr_pattern_num]); // get pointer to the current pattern
		}

		// try to find a meaning for this line 
		if(sscanf(linebuf,REGEX_FILE_FORMAT_RULEDEF,&ui1,s1) == 2){ // Ruleheader -> init new rule (ui1 = RuleId, s1 = the rule)
			pattern_t * new_pattern; // pointer to the new pattern location

#if REGEX_DEBUG > 0
			// new pattern: dump the last pattern
			if(in_pattern == true){ // if there was already a previous pattern
				if(dumpPattern(regexData,curr_pattern)<0){
					printerr("parseFile(): dumpPattern failed",0);
					return -1;
				}
			}
#endif

			// new pattern -> has enough space for more patterns?
			if(regexData->patterns_c >= regexData->max_patterns_c){ 
				printerr("parseFile(): Not enough space for more patterns",0);
				return -1;
			}
			
			// search if pattern with the same number already exist
			if(getPattern(ui1,regexData,NULL)>0){ // >0 beacuse getPattern should return an error
				printerr("parseFile(): Found a rule id conflict!",0);
				return -1;
			}

			// initialize new pattern
			new_pattern=&(regexData->patterns[regexData->patterns_c]); // location of new pattern
			new_pattern->id = ui1; // the id
			new_pattern->regex_len = strlen(s1)+1; // length of the regex
			new_pattern->regex = strndup(s1,new_pattern->regex_len-1); // copy regex

			// check if regex copy worked
			if(new_pattern->regex == NULL){
				printerr("parseFile(): strndup of regex failed",0);
				return -1;
			}

			// compile regex now
			new_pattern->regexCompiled_init = false; // set compiled regex as not there
			if(compileRegex(new_pattern,REGEX_COMPILE_FLAGS)<0){ // do compile with no 
				printerr("parseFile(): regex compilation failed",0);
				return -1;
			}

			// make bitmask of this rule
			new_pattern->mask = malloc(regexData->mask_size*sizeof(char)); // malloc space for mask
			memset(new_pattern->mask,0x0,regexData->mask_size*sizeof(char)); // reset mask to all zero
			if(new_pattern->mask == NULL){
				printerr("parseFile(): malloc of mask failed!",0);
				return -1;
			}

			// set bit in mask
			if(util_set_bit_mask(new_pattern->mask,0x1,regexData->patterns_c)<0){
				printerr("parseFile(): failed setting mask",0);
				return -1;
			}

			// init prev mask 
			new_pattern->prev_mask = malloc(regexData->mask_size*sizeof(char)); // malloc space for mask
			if(new_pattern->prev_mask == NULL){
				printerr("parseFiile(): malloc of prev_mask failed",0);
				return -1;
			}
			memset(new_pattern->prev_mask,0x0,regexData->mask_size*sizeof(char)); // init with zero

			// initialze other fields with defaults values
			new_pattern->studyRegex_init = false;
			new_pattern->studyRegex = NULL;
			new_pattern->command_init = false;
			new_pattern->command = NULL;
			new_pattern->command_len = 0;
			new_pattern->dodump = false;
			
			// set parsing information
			in_pattern = true; // we are in a pattern now
			curr_pattern_num = regexData->patterns_c; // our current pattern number

			// in general we have one pattern more
			regexData->patterns_c++; 

		}else if( in_pattern == true // are we in a pattern definiton?
			&& sscanf(linebuf,REGEX_FILE_FORMAT_DUMP,&ui1) == 1){

			// shall the rule cause a filedump?
			if(ui1 > 1){
				curr_pattern->dodump = true; // set the filedump flag to true
			}

		}else if( in_pattern == true // are we in a patter definiton?
			&& sscanf(linebuf,REGEX_FILE_FORMAT_COMPILE,&i1) == 1){

			// recompile the regex pattern
			if(compileRegex(curr_pattern,i1)<0){ // provice arguments
				printerr("parseFile(): regex compilation failed",0);
				return -1;
			}

		}else if( in_pattern == true // are we in a patter definiton?
			&& sscanf(linebuf,REGEX_FILE_FORMAT_STUDY,&i1) == 1){

			// do study the regex now!
			if(studyRegex(curr_pattern,i1)<0){
				printerr("parseFile(): Studying of pattern failed!",0);
				return -1;
			}

		}else if( in_pattern == true // are we in a patter definiton?
			&& sscanf(linebuf,REGEX_FILE_FORMAT_CMD,s1) == 1){

			// was there already a comamnd initialized?
			if(curr_pattern->command_init == true){
				free(curr_pattern->command); // free the old command
			}

			// save this command template
			curr_pattern->command_len = strlen(s1)+1;
			curr_pattern->command = strndup(s1,curr_pattern->command_len-1);
			// check if stdup worked
			if(curr_pattern->command == NULL){
				printerr("parseFile(): command strndup failed",0);
				return -1;
			}

			// set command as intialized
			curr_pattern->command_init = true;

		}else if( in_pattern == true // are we in a pattern definiton?
			&& sscanf(linebuf,REGEX_FILE_FORMAT_PREV,&ui1) == 1){
			pattern_t *prev_pattern;

			// get the required prev pattern
			if(getPattern(ui1,regexData,&(prev_pattern))<0){
				printerr("parseFile(): previous pattern not found -> Make sure that the are defined before using them as dependency!",0);
				return -1;
			}

			// make previous mask a dependency of this mask
			if(util_or_bit_mask(curr_pattern->prev_mask,prev_pattern->mask,curr_pattern->prev_mask,regexData->mask_size)<0){
				printerr("parseFile(): or operation on mask |= prev->mask failed",0);
				return -1;
			}

		}else{
#if REGEX_DEBUG > 0
			printf("Ignoring line[%u]: \"%s\"",i,linebuf);
#endif
		}
	}

#if REGEX_DEBUG > 0
	// last pattern: dump this pattern
	if(in_pattern == true){ // if there is even a pattern..
		if(dumpPattern(regexData,curr_pattern)<0){
			printerr("parseFile(): dumpPattern failed",0);
			return -1;
		}
	}
#endif

	// free the General purpose-vars
	free(s1);
	free(s2);
	free(s3);

	return 1;
}

static int loadFile(regexdata_t *regexData){
	FILE *fd;			// the fd for the regex-pattern-file
	int c;				// a general purpose buffer character
	char *linebuf;			// the linebuffer
	unsigned int line_c = 1;	// number of lines in the file
	unsigned int pattern_count = 0; // amount of regex patterns found in file
	unsigned int max_linesize = 0; 	// maximum linelength (for linebuffer)

	// open regex-file
	if((fd = fopen(regexData->regexfile_path,"r")) == NULL){
		printerr("loadFile(): Unable to open regex file!",4);
		return -1;
	}
	rewind(fd); // go to begining of file

	// count patterns / get linesize / get linecount
	do{
		unsigned int linesize = 0; // linesize

		c = fgetc(fd);
		if(((char)c) == REGEX_FILE_FORMAT_RULEDEF[0]){ // is this a new pattern?
			pattern_count++;
		}

		// go to next line
		while( 	c != EOF 
		 	&& c != '\n'){
			linesize++; // increment linesize
			c = fgetc(fd); // get next char
		}

		// longest line?
		if(max_linesize < linesize){
			max_linesize = linesize; // yep
		}

		line_c++;
	}while(c != EOF);
	line_c--; 	// decrement linecount again -1 (due to algorithm) 
	rewind(fd); // go to start again


	// at least some pattern found?
	if(pattern_count <= 0){
		printerr("loadFile(): No pattern found!",0);
		return -1;
	}

	// all the lines empty?
	if(max_linesize <= 0){
		printerr("loadFile(): Lines empty!",0);
		return -1;
	}
	max_linesize++; // +1 because of trailing \0 which will appear as soon as we use fgets/scanf and stuff

#if REGEX_DEBUG > 0
	printf(	"Found patterns: %u / "
		"Number of lines: %u / "
		"Max linesize: %u\n",
		pattern_count,
		line_c,
		max_linesize
		);
#endif

	// malloc space for patterns
	regexData->max_patterns_c = pattern_count; // save patterns count
	regexData->patterns = malloc(pattern_count*sizeof(pattern_t));
	if(regexData->patterns == NULL){
		printerr("loadFile(): Pattern malloc failed",0);
		return -1;
	}

	// get size of masking fileds
	regexData->mask_size = ceil(pattern_count/8.0f);
	if(regexData->mask_size <= 0){
		printerr("loadFile(): Mask of size 0 not valid!",0);
		return -1;
	}

	// malloc space for linebuffer
	linebuf = malloc(max_linesize*sizeof(char)); // malloc the linebuffer 
	if(linebuf == NULL){
		printerr("loadFile(): linebuffer malloc failed",0);
		return -1;
	}

	// parse the file now
	if(parseFile(fd,line_c,regexData,linebuf,max_linesize)<0){
		printerr("loadFile(): Fileparsing failed",0);
		return -1;
	}

	// local free's & cleanup
	free(linebuf);
	
	if(fclose(fd)!=0){
		printerr("loadFile(): closing of file failed",4);
		return -1;
	}

	return 1;
}

static int matchPattern(rexFlow_t *rexFlow,char *data,int data_len,regexdata_t *regexData){
	unsigned int i;
	// try to match each pattern
	for(i=0;i<regexData->patterns_c;i++){
		pattern_t *pattern = &(regexData->patterns[i]); // get the next pattern

		// but only if:
		if( util_cmp_bit_mask(rexFlow->mask,pattern->mask,regexData->mask_size) == false  // it didn't match before
			&& util_cmp_bit_mask(rexFlow->mask,pattern->prev_mask,regexData->mask_size) == true // previous required rules matched 
		 ){
			bool matched = true; // do we have a match here -> will be set to false if not
			int pcre_execReturn = 0;
			// TODO: Parse options atm 0x0
			// TODO: Extract substrings in ovector maype append them to tranalayzer output
			pcre_execReturn = pcre_exec(	pattern->regexCompiled,  // the compiled regex
							pattern->studyRegex, // the studied regex
							data, data_len,  // the data
							0x0, // no offset from start
							0x0, // no options
							NULL, // no overctor TODO: Dont know if this works because doc says: "smalles size: (n+1)*3"
							0x0 // no ovector -> size = 0 ( We are not interested in substrings yet)
							);
			// was there an error?
			if (pcre_execReturn < 0){ 
				// handle errorcodes of pcre_exec
				switch (pcre_execReturn){
					case PCRE_ERROR_NULL:
						printerr("matchPattern(): NULL ptr as data / ovector",0);
						return -1;
					break;
					case PCRE_ERROR_BADOPTION:
						printerr("matchPattern(): Bad options passed to pcre_exec",0);
						return -1;
					break;
					case PCRE_ERROR_BADMAGIC:
						printerr("matchPattern(): Missing magic number",0);
						return -1;
					break;
					case PCRE_ERROR_UNKNOWN_OPCODE:
						printerr("matchPattern(): Random opcode detected",0);
						return -1;
					break;
					case PCRE_ERROR_NOMEMORY:
						printerr("matchPattern(): Ovector memory not big enough",0);
						return -1;
					break;
					case PCRE_ERROR_NOSUBSTRING:
						printerr("matchPattern(): PCRE_ERROR_NOSUBSTRING this should __REALLY__ not happen here!",0);
						return -1;
					break;
					case PCRE_ERROR_MATCHLIMIT:
						printerr("matchPattern(): Match limit exceeded",0);
						return -1;
					break;
					case PCRE_ERROR_CALLOUT:
						printerr("matchPattern(): PCRE_ERROR_CALLOUT this should __REALLY__ not happen here!",0);
						return -1;
					break;
					case PCRE_ERROR_BADUTF8:
						printerr("matchPattern(): Not utf8 data presented",0);
						return -1;
					break;
					case PCRE_ERROR_PARTIAL:
						printerr("matchPattern(): Partial match detected",0);
						return -1;
					break;
					case PCRE_ERROR_BADPARTIAL:
						printerr("matchPattern(): PCRE_ERROR_BADPARTIAL no longer in use -> this should __REALLY__ not happen here!",0);
						return -1;
					break;
					case PCRE_ERROR_INTERNAL:
						printerr("matchPattern(): Internal pcre error",0);
						return -1;
					break;
					case PCRE_ERROR_BADCOUNT:
						printerr("matchPattern(): Ovecsize cannot be negative",0);
						return -1;
					break;
					case PCRE_ERROR_RECURSIONLIMIT:
						printerr("matchPattern(): match_limit_recursion limit has reached",0);
						return -1;
					break;
					case PCRE_ERROR_BADNEWLINE:
						printerr("matchPattern(): Bad newline option was given",0);
						return -1;
					break;
					case PCRE_ERROR_BADOFFSET:
						printerr("matchPattern(): startoffset cannot be negative or greater than the length of the subject",0);
						return -1;
					break;
					case PCRE_ERROR_SHORTUTF8:
						printerr("matchPattern(): Truncateed utf character at end detected",0);
						return -1;
					break;
					case PCRE_ERROR_RECURSELOOP:
						printerr("matchPattern(): Recursive loop in pattern",0);
						return -1;
					break;
					case PCRE_ERROR_JIT_STACKLIMIT:
						printerr("matchPattern(): Stack for jit processing not long enough",0);
						return -1;
					break; 
					case PCRE_ERROR_BADMODE:
						printerr("matchPattern(): Pattern compiled for a wrong library type",0);
						return -1;
					break;
					case PCRE_ERROR_BADENDIANNESS:
						printerr("matchPattern(): Endianness mismatch in  pattern detected",0);
						return -1;
					break;
					default: // wooops.. not a documented error..
						printerr("matchPattern(): unknown error!",0);
						return -1;
					break;
					case PCRE_ERROR_NOMATCH: // No match
						matched = false;
					break;
				}
			}
			

			if(matched == true){ // rule matched!
#if REGEX_DEBUG > 0
				// print debug message
				printf("Pattern[%u] matched Flow[%s]\n",pattern->id,rexFlow->idstr);
#endif
				// toogle flags for this pattern
				if(util_or_bit_mask(rexFlow->mask,pattern->mask,rexFlow->mask,regexData->mask_size)<0){ // or the bits!
					printerr("matchPattern(): Toogling flag failed",0);
					return -1;

				}

				// increment pattern match count
				rexFlow->patternMatchCount++;

				//start dumping of flow?
				if(pattern->dodump == true){ // yes
					rexFlow->dumpToFile = true; // then start dumping..
				}

				// as soon as the rule machted try to match all the other rules recursively
				if(matchPattern(rexFlow,data,data_len,regexData)<0){
					printerr("matchPattern(): recursive pattern match failed",0);
					return -1;
				}
			}
		}
	}

	return 1; // matching process completed
}

static int openFlowOutFile(rexFlow_t *rexFlow,regexdata_t *regexData) {
	size_t full_path_len = 	strlen(regexData->flowfiles_path)   	// base path (to dir)
				+ strlen(REGEX_FLOWDATA_FILEPREFIX)  	// Prefix
				+ strlen(rexFlow->idstr)		// the id of the flow
				+ strlen(REGEX_FLOWDATA_FILESUFFIX) 	// filesuffix with random ending
				+ 1; 					// +1 (becasue of '\0')
	char *full_path = malloc(full_path_len*sizeof(char));
	memset(full_path,'\0',full_path_len); // set all to '\0'

	// concatinate filepath step by step
	strncat(full_path,regexData->flowfiles_path,regexData->flowfiles_path_len); 	// base dir
	strncat(full_path,REGEX_FLOWDATA_FILEPREFIX,strlen(REGEX_FLOWDATA_FILEPREFIX)); // prefix 
	strncat(full_path,rexFlow->idstr,strlen(rexFlow->idstr)); 			// the id of the flow
	strncat(full_path,REGEX_FLOWDATA_FILESUFFIX,strlen(REGEX_FLOWDATA_FILESUFFIX));	// the id of the flow
	full_path[full_path_len-1] = '\0';						// '\0'

#if REGEX_DEBUG > 0
	printf("Opening file: \"%s\"\n",full_path);
#endif

	// try to open the file
	rexFlow->flowOutFile=fopen(rexFlow->flowOutFile_path,"w");
	if(rexFlow->flowOutFile == NULL){ // worked?
		printerr("openFlowOutFile(): Cannot open file",4);
		return -1;
	}

	// flowoutfile is now initialized
	rexFlow->flowOutFile_path=strdup(full_path); // i know i dont noeed to stdup here  but i'd really like to free full_path at the end.. (For  my personal statisfaction)
	if(rexFlow->flowOutFile_path==NULL){
		printerr("openFlowOutFile(): strdup of full_path failed",0);
		return -1;
	}

	// the file is initialized now!
	rexFlow->flowOutFile_init=true;

	// free path again
	free(full_path);

	return 1;
}

static int writePacket(rexFlow_t* rexFlow, packet_t* packet,regexdata_t *regexData) {
	char *payloadStart = (char*)packet->layer7Header;
	unsigned int l7length = packet->snapL7Length;

	// Output file not initialized yet
	if(rexFlow->flowOutFile_init == true){
		if(openFlowOutFile(rexFlow,regexData)<0){
			printerr("writePacket(): opening flow out file failed",0);
			return -1;
		}
	}

	// TODO: write to file..

	return 1;
}
