/*
 * SMTP Detector
 * Copyright (C), RUAG Defence, RUAG Electronics AG, Switzerland
 * Author: Torben Ruehl, RDNIZ
 *
 * Design concept is based on the RFCs 2821 and 5321, both named Simple Mail
 * Transfer Protocol.
 *
 */

#include "smtpDetector.h"
#include "packetCapture.h"
#include "tranalyzer_utils.h"

#if SMTP_DEBUG != 0
char debug_ipstring[INET6_ADDRSTRLEN];
#endif /* SMTP_DEBUG */

/* prototypes */
void process_client_payload(char* payload, int16_t payload_length, unsigned long flowIndex);
void process_server_payload(char* payload, int16_t payload_length, unsigned long flowIndex);
void smtp_mail_rcpt_to_output_string(unsigned long flowIndex);
void smtp_add_email_address(unsigned long flowIndex, int type, char* name, char* domain);
void smtp_flow_free_pointer(unsigned long flowIndex);

char* get_plugin_name() {
	return "smtpDetector";
}

char* get_plugin_version() {
	return "0.5.0";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	plugin_smtpDetector_Index = pPluginIndex;
	// init memory field
	smtp_flow = malloc(sizeof(smtp_flow_t) * mainHashMap->hashChainTableSize);
}

binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;
#if SMTP_OUTPUT_COMPLEXITY < 3 // TODO: Change to 1 after implementation of complexity level 2
	binary_value_t *act_bv = bv_new_bv("SMTP MAIL FROM->RCPT TO", "SMTP_M->R", 0, 1, bt_string);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif /* SMTP_OUTPUT_COMPLEXITY == 1 */

#if SMTP_OUTPUT_COMPLEXITY == 2
#endif /* SMTP_OUTPUT_COMPLEXITY == 2 */

#if SMTP_OUTPUT_COMPLEXITY == 3
	binary_value_t *act_bv;
	act_bv = bv_new_bv("SMTP Role of sender in connection", "SMTP_Role", 0, 1, bt_uint_8);
	main_bv = bv_append_bv(main_bv, act_bv);
	act_bv = bv_new_bv("SMTP MAIL FROM->RCPT TO", "SMTP_M->R", 1, 2, 0, 0);
	act_bv = bv_add_sv_to_bv(act_bv,0, 1, 2, bt_string, bt_string);
	act_bv = bv_add_sv_to_bv(act_bv,1, 1, 2, bt_string, bt_string);
	main_bv = bv_append_bv(main_bv, act_bv);
	act_bv = bv_new_bv("SMTP Error flags", "SMTP_err_flgs", 0, 1, bt_hex_8);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif /* SMTP_OUTPUT_COMPLEXITY == 3 */

	return main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
	// reset struct
	memset(&smtp_flow[flowIndex], '\0', sizeof(smtp_flow_t));
#if SMTP_STRICT_CHECK != 0
	smtp_flow[flowIndex].type = SMTP_TYPE_NO_SMTP;

	if (flows[flowIndex].layer4Protocol == IPPROTO_TCP) {
		if (flows[flowIndex].dstPort == 25 || flows[flowIndex].dstPort == 587 || flows[flowIndex].dstPort == 465) {
#if SMTP_DEBUG != 0
			inet_ntop(AF_INET, &flows[flowIndex].srcIP.s_addr, (char*) &debug_ipstring, INET6_ADDRSTRLEN);
			fprintf(stdout, "smtpDetector: possible SMTP client flow from %s:%"PRIu16"->", (char*) &debug_ipstring, flows[flowIndex].srcPort);
			inet_ntop(AF_INET, &flows[flowIndex].dstIP.s_addr, (char*) &debug_ipstring, INET6_ADDRSTRLEN);
			fprintf(stdout, "%s:%"PRIu16"\n", (char*) &debug_ipstring, flows[flowIndex].dstPort);
#endif /* SMTP_DEBUG != 0 */
			smtp_flow[flowIndex].type = SMTP_TYPE_UNDEFINED;
		} else if (flows[flowIndex].srcPort == 25 || flows[flowIndex].srcPort == 587 || flows[flowIndex].srcPort == 465) {
#if SMTP_DEBUG != 0
			inet_ntop(AF_INET, &flows[flowIndex].srcIP.s_addr, (char*) &debug_ipstring, INET6_ADDRSTRLEN);
			fprintf(stdout, "smtpDetector: possible SMTP server flow from %s:%"PRIu16"->", (char*) &debug_ipstring, flows[flowIndex].srcPort);
			inet_ntop(AF_INET, &flows[flowIndex].dstIP.s_addr, (char*) &debug_ipstring, INET6_ADDRSTRLEN);
			fprintf(stdout, "%s:%"PRIu16"\n", (char*) &debug_ipstring, flows[flowIndex].dstPort);
#endif /* SMTP_DEBUG != 0 */
		smtp_flow[flowIndex].type = SMTP_TYPE_UNDEFINED;
		}
	}

#else  /* SMTP_STRICT_CHECK != 0 */
	// At least, it should be a TCP flow.
	if (flows[flowIndex].layer4Protocol == IPPROTO_TCP) {
		smtp_flow[flowIndex].type = SMTP_STATE_UNDEFINED;
	} else {
		smtp_flow[flowIndex].type = SMTP_STATE_NO_SMTP;
	}
#endif /* SMTP_STRICT_CHECK != 0 */
}

void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	if (smtp_flow[flowIndex].type == SMTP_TYPE_NO_SMTP) return;
	// print_packet(stdout, packet, 16); // just for debug
	// check if packet contains payload
	u_char* payload = ((u_char*) packet->layer4Header) + 4 * packet->layer4Header->tcpHeader.doff;
	int16_t header_length =(IP_HL(&packet->layer3Header->ipHeader) * 4) + (packet->layer4Header->tcpHeader.doff * 4);
	int16_t payload_length = ntohs(packet->layer3Header->ipHeader.ip_len) - header_length; // if - for whatever strange reasons - payload should be negative, we take a signed var.
#if SMTP_DEBUG >= 3
	printf("smtpDecoder: Header length: %"PRId16" -> payload length: %"PRId16"\n", header_length, payload_length);
#endif /* SMTP_DEBUG >= 3 */

	if (payload_length <= 0) return;

	if (smtp_flow[flowIndex].type == SMTP_TYPE_UNDEFINED) {
		/*
		 * Until now, we don't know if it is a client or a server or no smtp
		 * flow.
		 * A client would indicate itself with the smtp commands (HELO,
		 * EHLO, MAIL, RCPT, DATA, RSET, NOOP, QUIT, VRFY, HELP, AUTH). A
		 * server should start with the 220 (greetings) code.
		 * If we don't see this, We assume that this is no smtp flow or it is
		 * encrypted.
0a		 */
		// check if flow is sent from a smtp client
		if (!strncasecmp(payload, "EHLO", 4) || !strncasecmp(payload, "HELO", 4) || !strncasecmp(payload, "MAIL", 4) ||
				!strncasecmp(payload, "RCPT", 4) || !strncasecmp(payload, "DATA", 4) || !strncasecmp(payload, "RSET", 4) ||
				!strncasecmp(payload, "NOOP", 4) || !strncasecmp(payload, "QUIT", 4) || !strncasecmp(payload, "VRFY", 4) ||
				!strncasecmp(payload, "AUTH", 4) || !strncasecmp(payload, "EXPN", 4) || !strncasecmp(payload, "HELP", 4)
		) {
#if SMTP_DEBUG != 0
			printf("found smtp client flow\n");
#endif /* SMTP_DEBUG != 0 */
			smtp_flow[flowIndex].type = SMTP_TYPE_CLIENT;
			smtp_flow[flowIndex].state = SMTP_STATE_START;
		}

		// check if flow is sent from a smtp server
		else if (!strncasecmp(payload, "220", 3)) {
#if SMTP_DEBUG != 0
			printf("found smtp server flow\n");
#endif /* SMTP_DEBUG != 0 */
			smtp_flow[flowIndex].type = SMTP_TYPE_SERVER;
			smtp_flow[flowIndex].state = SMTP_STATE_START;
		}

		// this is not a smtp flow.
		else {
#if SMTP_DEBUG != 0
			printf("found no smtp flow\n");
#endif /* SMTP_DEBUG != 0 */
			smtp_flow[flowIndex].type = SMTP_TYPE_NO_SMTP;
			return;
		}
	}

	switch (smtp_flow[flowIndex].type) {
	case SMTP_TYPE_CLIENT:
		process_client_payload(payload, payload_length, flowIndex);
		break;
	case SMTP_TYPE_SERVER:
		process_server_payload(payload, payload_length, flowIndex);
		break;
	default:
		fprintf(stderr, "smptDetector: BUG in switch statement in cl4I!\n");
	}
}

void onFlowTerminate(unsigned long flowIndex) {
	// it could happen, that a SMTP session didn't reach the DATA state but the
	// sender and recipients are already given -> write them out, too.
	if ((smtp_flow[flowIndex].act_sender != NULL) || (smtp_flow[flowIndex].act_recipients != NULL)) {
		smtp_mail_rcpt_to_output_string(flowIndex);
		// Set error flag
		smtp_flow[flowIndex].errors |= SMTP_ERR_NO_DATA;
	}

	smtp_email_details_t *curr_email = smtp_flow[flowIndex].out.emails;
	int index;
	char term = '\0';
	char ape = '@';
	char delim = ';';

#if SMTP_OUTPUT_COMPLEXITY < 3
	if (!curr_email) outputBuffer_append(main_output_buffer, SMTP_OUTPUT_MISSING, strlen(SMTP_OUTPUT_MISSING));
	for (;curr_email != NULL; curr_email = curr_email->next) {
		outputBuffer_append(main_output_buffer, SMTP_OUTPUT_LEFT, strlen(SMTP_OUTPUT_LEFT));
		// process sender
		for (index = 0; index < curr_email->num_sender; index++) {
			if (index) outputBuffer_append(main_output_buffer, (char*) &delim, 1);
			outputBuffer_append(main_output_buffer, curr_email->sender[index].name, strlen(curr_email->sender[index].name));
			outputBuffer_append(main_output_buffer, (char*) &ape, 1);
			outputBuffer_append(main_output_buffer, curr_email->sender[index].domain, strlen(curr_email->sender[index].domain));
		}

		// write delimeter
		outputBuffer_append(main_output_buffer, SMTP_OUTPUT_ARROW, strlen(SMTP_OUTPUT_ARROW));

		// process recipients
		for (index = 0; index < curr_email->num_recipients; index++) {
			if (index) outputBuffer_append(main_output_buffer, (char*) &delim, 1);
			outputBuffer_append(main_output_buffer, curr_email->recipients[index].name, strlen(curr_email->recipients[index].name));
			outputBuffer_append(main_output_buffer, (char*) &ape, 1);
			outputBuffer_append(main_output_buffer, curr_email->recipients[index].domain, strlen(curr_email->recipients[index].domain));
		}
		outputBuffer_append(main_output_buffer, SMTP_OUTPUT_RIGHT, strlen(SMTP_OUTPUT_RIGHT));
	}
	// terminate string
	outputBuffer_append(main_output_buffer, (char*) &term, 1);
#endif /* SMTP_OUTPUT_COMPLEXITY < 3 */

#if SMTP_OUTPUT_COMPLEXITY == 3
	// write the role of the flow sender to the output
	// corresponds to the SMTP_TYPE_... definitions
	outputBuffer_append(main_output_buffer, (char*) &(smtp_flow[flowIndex].type), sizeof(uint8_t));

	// Write the emails to the output
	// First the number of eMails
	outputBuffer_append(main_output_buffer, (char*) ((uint32_t*) &(smtp_flow[flowIndex].out.num_emails)), sizeof(uint32_t));
	// Iterate trough emails
	for (;curr_email != NULL; curr_email = curr_email->next) {
		// Process sender
		// Number of sender
		outputBuffer_append(main_output_buffer, (char*) ((uint32_t*) &(curr_email->num_sender)), sizeof(uint32_t));
		// Iterate through sender
		for (index = 0; index < curr_email->num_sender; index++) {
			if (curr_email->sender[index].name)
				outputBuffer_append(main_output_buffer, (char*) curr_email->sender[index].name, strlen(curr_email->sender[index].name));
			else
				outputBuffer_append(main_output_buffer, SMTP_OUTPUT_MISSING, strlen(SMTP_OUTPUT_MISSING));
			outputBuffer_append(main_output_buffer, (char*) &term, 1);
			if (curr_email->sender[index].domain)
				outputBuffer_append(main_output_buffer, (char*) curr_email->sender[index].domain, strlen(curr_email->sender[index].domain));
			else
				outputBuffer_append(main_output_buffer, SMTP_OUTPUT_MISSING, strlen(SMTP_OUTPUT_MISSING));
			outputBuffer_append(main_output_buffer, (char*) &term, 1);
		}
		// Process recipients
		// Number of recipients
		outputBuffer_append(main_output_buffer, (char*) ((uint32_t*) &(curr_email->num_recipients)), sizeof(uint32_t));
		// Iterate through recipients
		for (index = 0; index < curr_email->num_recipients; index++) {
			if (curr_email->recipients[index].name)
				outputBuffer_append(main_output_buffer, (char*) curr_email->recipients[index].name, strlen(curr_email->recipients[index].name));
			else
				outputBuffer_append(main_output_buffer, SMTP_OUTPUT_MISSING, strlen(SMTP_OUTPUT_MISSING));
			outputBuffer_append(main_output_buffer, (char*) &term, 1);
			if (curr_email->recipients[index].domain)
				outputBuffer_append(main_output_buffer, (char*) curr_email->recipients[index].domain, strlen(curr_email->recipients[index].domain));
			else
				outputBuffer_append(main_output_buffer, SMTP_OUTPUT_MISSING, strlen(SMTP_OUTPUT_MISSING));
			outputBuffer_append(main_output_buffer, (char*) &term, 1);
		}
	}

	outputBuffer_append(main_output_buffer, (char*) &(smtp_flow[flowIndex].errors), sizeof(uint8_t));

#endif /* SMTP_OUTPUT_COMPLEXITY == 3 */

	// free pointers in smtp flow
	smtp_flow_free_pointer(flowIndex);

#if SMTP_DEBUG != 0
	if (smtp_flow[flowIndex].type == SMTP_TYPE_CLIENT) {
	}
#endif /* SMTP_DEBUG != 0 */
}

void onApplicationTerminate() {
	// free main struct
	if (smtp_flow) free(smtp_flow);
}

void process_client_payload(char* payload, int16_t payload_length, unsigned long flowIndex) {
	// copy payload into buffer, so we can transform it with strtok_r without
	// disturbing other plugins
	char* smtp_payload_buffer = malloc((payload_length + 1) * sizeof(char)); // init payload buffer
	memcpy(smtp_payload_buffer, payload, payload_length);
	smtp_payload_buffer[payload_length] = '\0'; // terminate string

	char *save_ptr;
	char *current_line = strtok_r(smtp_payload_buffer, "\n\r", &save_ptr);
	while (current_line) {
#if SMTP_DEBUG != 0
		fprintf(stdout, "current line:'%s'\n", current_line);
#endif /* SMTP_DEBUG >= 2 */

		// check if MAIL command is seen. Ignore it if we are in data mode
		if (!strncasecmp(current_line, "MAIL FROM:", 10) && smtp_flow[flowIndex].state != SMTP_STATE_DATA) {
#if SMTP_DEBUG != 0
			fprintf(stdout, "smtpDecoder: MAIL seen. flags: %"PRIu32"\n", smtp_flow[flowIndex].observed_commands);
#endif /* SMTP_DEBUG != 0 */
			// check state
			// if (smtp_flow[flowIndex].)


			// check repeated occurence
			if ((smtp_flow[flowIndex].observed_commands & ((u_int32_t) SMTP_CLIENT_MAIL)) == SMTP_CLIENT_MAIL) {
				smtp_flow[flowIndex].errors |= SMTP_ERR_MULTIPLE_MAIL;
#if SMTP_DEBUG != 0
				fprintf(stdout, "smtpDecoder: Warning: repeated occurence of MAIL\n");
#endif /* SMTP_DEBUG != 0 */
			}
			// set flag
			smtp_flow[flowIndex].observed_commands |= SMTP_CLIENT_MAIL;

			// extract sender and store it in the struct. should be behind the first ':' char
			u_int16_t address_length;
			char *start = strchr(current_line, ':');
			char *mail_save_ptr;
			char *sender = strtok_r(start, SMTP_MAIL_ADDRESS_DELIMETERS "@", &mail_save_ptr);
			char *domain = strtok_r(NULL, SMTP_MAIL_ADDRESS_DELIMETERS, &mail_save_ptr);
			smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_SENDER, sender, domain);
		}

		// check if RCPT command is seen. Ignore it if we are in data mode
		else if (!strncasecmp(current_line, "RCPT TO:", 8) && smtp_flow[flowIndex].state != SMTP_STATE_DATA) {
#if SMTP_DEBUG != 0
				fprintf(stdout, "smtpDecoder: RCPT seen\n");
#endif /* SMTP_DEBUG != 0 */
			// set flag
			smtp_flow[flowIndex].observed_commands |= SMTP_CLIENT_RCPT;

			// extract recipient and store it in the struct. should be behind the first ':' char
			u_int16_t address_length;
			char *start = strchr(current_line, ':');
			char *mail_save_ptr;
			char *sender = strtok_r(start, SMTP_MAIL_ADDRESS_DELIMETERS "@", &mail_save_ptr);
			char *domain = strtok_r(NULL, SMTP_MAIL_ADDRESS_DELIMETERS, &mail_save_ptr);
			smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_RECIPIENT, sender, domain);
		}

		// check if DATA command is seen. Ignore if we are in data mode.
		else if (!strncasecmp(current_line, "DATA", 4) && smtp_flow[flowIndex].state != SMTP_STATE_DATA) {
#if SMTP_DEBUG != 0
				fprintf(stdout, "smtpDecoder: Entering data mode\n");
#endif /* SMTP_DEBUG != 0 */
			// set flag
			smtp_flow[flowIndex].observed_commands |= SMTP_CLIENT_DATA;
			// set mode
			smtp_flow[flowIndex].state = SMTP_STATE_DATA;

			// RFC says, that after DATA mode the eMail is ready to deliver.
			// Therefore we can already move the sender and recipients buffers
			// (if one of them isn't NULL) to the output.
			if ((smtp_flow[flowIndex].act_sender != NULL) || (smtp_flow[flowIndex].act_recipients != NULL))
				smtp_mail_rcpt_to_output_string(flowIndex);

			// Reset the flags:
			// Mail flag
			smtp_flow[flowIndex].observed_commands &= (SMTP_CLIENT_MAIL^UINT32_MAX);
			// rcpt flag
			smtp_flow[flowIndex].observed_commands &= (SMTP_CLIENT_RCPT^UINT32_MAX);
		}

		// if we are in data mode, everything except a line with a single point (".") will be ignored
		else if (!strncasecmp(current_line, ".", 1) && smtp_flow[flowIndex].state == SMTP_STATE_DATA && strlen(current_line) == 1) {
#if SMTP_DEBUG != 0
			fprintf(stdout, "smtpDecoder: leaving data mode\n");
#endif /* SMTP_DEBUG != 0 */
			smtp_flow[flowIndex].state = SMTP_STATE_INIT;
		}

		current_line = strtok_r(NULL, "\n\r", &save_ptr);
	}
	free(smtp_payload_buffer); smtp_payload_buffer = NULL;
}


/*
 * A general note to the next function:
 * Server flow payload is much harder to interpret, because it only reacts on
 * the client. Therefore we focused on the ability to handle correct smtp connections
 * where only one flow - server to client - is seen.
 */
void process_server_payload(char* payload, int16_t payload_length, unsigned long flowIndex) {
	// copy payload into buffer, so we can transform it with strtok_r without
	// disturbing other plugins
	char* smtp_payload_buffer = malloc((payload_length + 1) * sizeof(char)); // init payload buffer
	memcpy(smtp_payload_buffer, payload, payload_length);
	smtp_payload_buffer[payload_length] = '\0'; // terminate string

	char *save_ptr;
	char *current_line = strtok_r(smtp_payload_buffer, "\n\r", &save_ptr);
	while (current_line) {
#if SMTP_DEBUG != 0
		fprintf(stdout, "current line:'%s'\n", current_line);
#endif /* SMTP_DEBUG != 0 */

		if (smtp_flow[flowIndex].state == SMTP_STATE_START && !strncasecmp(current_line, "250", 3)) {
			// most often used init command by clients is EHLO, which is responded by the server with its
			// additional functions separated in lines, this needs to be handled.
			// Solution: Only react on the last line, which is different to the others.
			if (current_line[3] != '-') {
				smtp_flow[flowIndex].state = SMTP_STATE_INIT;
#if SMTP_DEBUG > 0
				fprintf(stdout, "smtpDetector: server state changed to INIT'\n", current_line);
#endif /* SMTP_DEBUG != 0 */
			}
		}

		// server acks a sender or recipient email address
		else if (smtp_flow[flowIndex].state == SMTP_STATE_INIT && !strncasecmp(current_line, "250", 3)) {
#if SMTP_DEBUG > 0
			printf("SERVER acks mail header info\n");
#endif /* SMTP_DEBUG != 0 */
			// Hopefully the server is a little bit chatty and tells us something about the acked address.

			// Check if we find a '@', this would mean there is an email address in it
			char* ape_pos = NULL;
			char *start_pos = ape_pos, *stop_pos = ape_pos;
			char *temp_name = NULL, *temp_domain = NULL;
			if ((ape_pos = strchr(current_line, '@')) != NULL) {
				// search left and right for the first sign that isn't allowed in an email address
				start_pos = ape_pos;
				stop_pos = ape_pos;
				while (!strchr(SMTP_MAIL_ADDRESS_DELIMETERS, *(start_pos - 1))) start_pos = start_pos - 1;
				while (!strchr(SMTP_MAIL_ADDRESS_DELIMETERS, *(stop_pos + 1))) stop_pos = stop_pos + 1;

				// extract the email address
				temp_name = malloc((ape_pos - start_pos + 1) * sizeof(char));
				memcpy(temp_name, start_pos, ape_pos - start_pos);
				temp_name[ape_pos - start_pos] = '\0';

				temp_domain = malloc((stop_pos - ape_pos + 1) * sizeof(char));
				memcpy(temp_domain, ape_pos + 1, (stop_pos - ape_pos) + 1);
				temp_domain[stop_pos - ape_pos] = '\0';
			}

			// Check if we find something that indicates the type (sender or recipient) of the email address
			int type = SMTP_EMAIL_TYPE_UNDEFINED; // no clue until now
			// search for the sender

			/*
			 * Note: strCASEstr produces a strange warning during compiling:
			 * "... comparision between pointer and integer ..."
			 * So far, on Ubuntu we can ignore this, BUT strcasestr is not an
			 * official standard so it may cause problems on other systems
			 * TODO: change configure.ac script for detecting strcasestr
			 */
			if (
					strcasestr(current_line, "send") != 0 // very generic search: Everything with send is matched (i.e. sender, sending, ...)
					|| strcasestr((char*) current_line, (char*) "is syntactically correct") != 0 // specific to web.de
					// || strcasestr(current_line, "OK") != NULL // debug
			) {
				type = SMTP_EMAIL_TYPE_SENDER;
			} else if (
					strcasestr(current_line, "rec") != 0 // very generic search: Everything with rec is matched (i.e. recipient, reciever, ...)
					|| strcasestr(current_line, "verified") != 0 // specific to web.de
			) {
				type = SMTP_EMAIL_TYPE_RECIPIENT;
			}

			// if type = undefined, We assume that the first occurence of 250 is for ack'ing the sender's email address
			if (type == SMTP_EMAIL_TYPE_UNDEFINED) {
				if (!smtp_flow[flowIndex].act_sender) {
					if (temp_name) {
						smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_SENDER, temp_name, temp_domain);
#if SMTP_DEBUG > 0
						fprintf(stdout, "smtpDetector: processing server flow: assuming new sender, address found\n");
#endif /* SMTP_DEBUG != 0 */
					}
					else {
						smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_SENDER, "???", "???");
#if SMTP_DEBUG > 0
		fprintf(stdout, "smtpDetector: processing server flow: assuming new sender, address not found\n");
#endif /* SMTP_DEBUG != 0 */
					}
				} else { // every other are treated as recievers
					if (temp_name) {
						smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_RECIPIENT, temp_name, temp_domain);
#if SMTP_DEBUG > 0
						fprintf(stdout, "smtpDetector: processing server flow: assuming new recipient, address found\n");
#endif /* SMTP_DEBUG != 0 */
					}
					else {
						smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_RECIPIENT, "???", "???");
#if SMTP_DEBUG > 0
						fprintf(stdout, "smtpDetector: processing server flow: assuming new recipient, address not found\n");
#endif /* SMTP_DEBUG != 0 */
					}
				}
			} else if (type == SMTP_EMAIL_TYPE_SENDER) {
				// store it in the sender field, no matter if there is already one inside
				if (temp_name) {
					smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_SENDER, temp_name, temp_domain);
#if SMTP_DEBUG > 0
					fprintf(stdout, "smtpDetector: processing server flow: new sender, address found\n");
#endif /* SMTP_DEBUG != 0 */
				}
				else {
					smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_SENDER, "???", "???");
#if SMTP_DEBUG > 0
					fprintf(stdout, "smtpDetector: processing server flow: new sender, address found\n");
#endif /* SMTP_DEBUG != 0 */
				}
			} else if (type == SMTP_EMAIL_TYPE_RECIPIENT) {
				if (temp_name) {
					smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_RECIPIENT, temp_name, temp_domain);
#if SMTP_DEBUG > 0
					fprintf(stdout, "smtpDetector: processing server flow: new recipient, address found\n");
#endif /* SMTP_DEBUG != 0 */
				}
				else {
					smtp_add_email_address(flowIndex, SMTP_EMAIL_TYPE_RECIPIENT, "???", "???");
#if SMTP_DEBUG > 0
					fprintf(stdout, "smtpDetector: processing server flow: new recipient, address not found\n");
#endif /* SMTP_DEBUG != 0 */
				}
			}

			// free temp fields if necessary
			if (temp_name) free(temp_name);
			if (temp_domain) free(temp_domain);
		}

		// server acked a DATA command, so we can move the info into an email struct
		else if (smtp_flow[flowIndex].state == SMTP_STATE_INIT && !strncasecmp(current_line, "354", 3)) {
#if SMTP_DEBUG > 0
			printf("SERVER ACKS DATA\n");
#endif /* SMTP_DEBUG != 0 */
			smtp_mail_rcpt_to_output_string(flowIndex);
			smtp_flow[flowIndex].state = SMTP_STATE_DATA;
		}

		// DATA part finally acked, reset state
		else if (smtp_flow[flowIndex].state == SMTP_STATE_DATA && !strncasecmp(current_line, "250", 3)) {
#if SMTP_DEBUG > 0
			printf("SERVER ACKS DATA PART END\n");
#endif /* SMTP_DEBUG != 0 */
			smtp_flow[flowIndex].state = SMTP_STATE_INIT;
		}

		current_line = strtok_r(NULL, "\n\r", &save_ptr);
	}
	free(smtp_payload_buffer); smtp_payload_buffer = NULL;
}


void smtp_mail_rcpt_to_output_string(unsigned long flowIndex) {
	//We build a new smtp eMail details struct
	smtp_flow_t *act = &smtp_flow[flowIndex];

	if (act->out.num_emails == UINT32_MAX) {
		// We reached the maximum amount of eMails to handle.
//		fprintf(stderr, "smtpDetector ERROR: flow index: %ld exceeded maximal eMail storage", flows[flowIndex].findex);
		act->errors |= SMTP_ERR_MM_OVRN;
		return;
	}

	if (!act->out.emails) {
		act->out.emails = malloc(sizeof(smtp_email_details_t));
		act->out.last_email = act->out.emails;
	} else {
		act->out.last_email->next = malloc(sizeof(smtp_email_details_t));
		act->out.last_email = act->out.last_email->next;
	}
	act->out.last_email->next = NULL;
	act->out.num_emails++;

	// fill the struct and reset the act->act_... structs
	act->out.last_email->sender = act->act_sender;
	act->act_sender = NULL;
	act->out.last_email->num_sender = act->act_num_sender;
	act->act_num_sender = 0;
	act->out.last_email->recipients = act->act_recipients;
	act->act_recipients = NULL;
	act->out.last_email->num_recipients = act->act_num_recipients;
	act->act_num_recipients = 0;
}

void smtp_add_email_address(unsigned long flowIndex, int type, char* name, char* domain) {
	smtp_email_address_t **address;
	u_int32_t *amount;
	// point to sender or recipient
	if (type == SMTP_EMAIL_TYPE_SENDER) {
		address = &(smtp_flow[flowIndex].act_sender);
		amount = &(smtp_flow[flowIndex].act_num_sender);
	} else if (type == SMTP_EMAIL_TYPE_RECIPIENT){
		address = &(smtp_flow[flowIndex].act_recipients);
		amount = &(smtp_flow[flowIndex].act_num_recipients);
	}

	(*amount)++;
	(*address) = realloc((smtp_email_address_t*) (*address), *amount * sizeof(smtp_email_address_t));
	smtp_email_address_t *curr_addr = *address;
	// copy name
	if (name) {
		curr_addr[*amount - 1].name = malloc((strlen(name) + 1) * sizeof(char));
		strncpy(curr_addr[*amount - 1].name, name, strlen(name) + 1);
	} else {
		curr_addr[*amount - 1].name = malloc((strlen("???") + 1) * sizeof(char));
		strncpy(curr_addr[*amount - 1].name, "???", strlen("???") + 1);
	}
	// copy domain
	if(domain) {
		curr_addr[*amount - 1].domain = malloc((strlen(domain) + 1) * sizeof(char));
		strncpy(curr_addr[*amount - 1].domain, domain, strlen(domain) + 1);
	} else {
		curr_addr[*amount - 1].domain = malloc((strlen("???") + 1) * sizeof(char));
		strncpy(curr_addr[*amount - 1].domain, "???", strlen("???") + 1);
	}
}

/*
 * free everything in a smtp flow
 */
void smtp_flow_free_pointer(unsigned long flowIndex) {
	int index;
	smtp_flow_t *act = &smtp_flow[flowIndex];
	if (act->auth_type) free(act->auth_type);

	if (act->act_num_sender > 0) {
		for (index = 0; index < act->act_num_sender; index++) {
			if (act->act_sender[index].name)	free(act->act_sender[index].name);
			if (act->act_sender[index].domain)	free(act->act_sender[index].domain);
		}
	}

	if (act->act_num_recipients > 0) {
		for (index = 0; index < act->act_num_recipients; index++) {
			if (act->act_recipients[index].name)	free(act->act_recipients[index].name);
			if (act->act_recipients[index].domain)	free(act->act_recipients[index].domain);
		}
	}

	if (act->out.num_emails > 0) {
		smtp_email_details_t *curr_email = act->out.emails;
		smtp_email_details_t *prev_email;
		while (curr_email) {
			for (index = 0; index < curr_email->num_sender; index++) {
				if (curr_email->sender[index].name)		free(curr_email->sender[index].name);
				if (curr_email->sender[index].domain)	free(curr_email->sender[index].domain);
			}

			for (index = 0; index < curr_email->num_recipients; index++) {
				if (curr_email->recipients[index].name)		free(curr_email->recipients[index].name);
				if (curr_email->recipients[index].domain)	free(curr_email->recipients[index].domain);
			}
			prev_email = curr_email;
			curr_email = curr_email->next;
			free(prev_email);
		}
	}
}
