/*
 * SMTP Detector
 * Copyright (C), RUAG Defence, RUAG Electronics AG, Switzerland
 * Author: Torben Ruehl, RDNIZ
 */

#ifndef SMTP_DETECTOR_H_
#define SMTP_DETECTOR_H_

// global includes
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <inttypes.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <strings.h>

// local includes
#include "networkHeaders.h"
#include "main.h"
#include "outputBuffer.h"
#include "binaryValue.h"

#define SMTP_DEBUG			0	// activates DEBUG output for this plugin
#define SMTP_STRICT_CHECK		1	// activates strict checking of SMTP protocol.
#define SMTP_OUTPUT_COMPLEXITY		3	// complexity of output: 1=a single string, 3=split in own subvalues

/* definition of smtp types */
#define SMTP_TYPE_UNDEFINED		0	// we have no clue if it is client or server
#define SMTP_TYPE_NO_SMTP		1	// this is not a smtp flow
#define SMTP_TYPE_CLIENT		2	// seems to be a client
#define SMTP_TYPE_SERVER		3	// seems to be a server
#define SMTP_TYPE_SENDER_ADDRESS	4	// observed address seems to be from the email sender
#define SMTP_TYPE_RECIPIENT_ADDRESS	5	// observed address seems to be from the email recipient


/* definition of different states */
#define SMTP_STATE_START	0	// The really beginning, before the EHLO / HELO
#define SMTP_STATE_INIT		10	// connection is initialized -> C: sent HELO EHLO and S: sent 250-...
#define SMTP_STATE_DATA		20	// sender is in data mode -> client sends, server waits for end indicator
#define SMTP_STATE_DELIVER	30	// a mail is ready to be delivered

/*
 * Definition of command fields for client
 * follows the minimum implementation as described in RFC 5321, section 4.5.1.
 */
#define SMTP_CLIENT_EHLO	0x0001 // 2^0
#define SMTP_CLIENT_HELO	0x0002 // 2^1
#define SMTP_CLIENT_MAIL	0x0004 // 2^2
#define SMTP_CLIENT_RCPT	0x0008 // 2^3
#define SMTP_CLIENT_DATA	0x0010 // 2^5
#define SMTP_CLIENT_RSET	0x0020 // 2^6
#define SMTP_CLIENT_NOOP	0x0040 // 2^7
#define SMTP_CLIENT_QUIT	0x0080 // 2^8
#define SMTP_CLIENT_VRFY	0x0100 // 2^9
//Following fields are "additional" commands not necessary for a working smtp service
#define SMTP_CLIENT_AUTH	0x0200
#define SMTP_CLIENT_EXPN	0x0400
#define SMTP_CLIENT_HELP	0x0800


/* definition of error fields */
#define SMTP_ERR_COMMAND_ORDER	0x01	//
#define SMTP_ERR_MULTIPLE_MAIL	0x02	//
#define SMTP_ERR_NO_DATA	0x04	//
#define SMTP_ERR_NO_INIT	0x08	//
#define SMTP_ERR_MM_OVRN	0x10	// Maximal administratable mail storage exeeded

// All chars not allowed in an eMail address. See RFC 822, section 3.3 and 6.1
#define SMTP_MAIL_ADDRESS_DELIMETERS	"()<>,;:\\\"[] \a\b\f\n\r\t\v"

// Strings for the output
#define SMTP_OUTPUT_ARROW	"->"
#define SMTP_OUTPUT_LEFT	"["
#define SMTP_OUTPUT_RIGHT	"]"
#define SMTP_OUTPUT_MISSING	"---"

// Type of email address, needed in the add_email_address fcn
#define SMTP_EMAIL_TYPE_UNDEFINED	0
#define SMTP_EMAIL_TYPE_SENDER		1
#define SMTP_EMAIL_TYPE_RECIPIENT	2

// output structs:
// address
typedef struct smtp_email_address_s {
	char* name;
	char* domain;
} smtp_email_address_t;

typedef struct smtp_email_details_s {
	smtp_email_address_t *sender; 		// multiple sender are stored as array
	smtp_email_address_t *recipients; 	// multiple recipients are stored as array
	uint32_t num_sender;
	uint32_t num_recipients;
	struct smtp_email_details_s *next;
} smtp_email_details_t;

typedef struct smtp_flow_output{
	smtp_email_details_t *emails;		// a list of all eMail
	smtp_email_details_t *last_email;	// the last seen eMail
	uint32_t num_emails;			// the number of eMails
}smtp_flow_output_t;


// main struct
typedef struct smtp_flow_s {

	// output:
	smtp_email_address_t *act_sender; 	// multiple sender are stored as array
	smtp_email_address_t *act_recipients;	// multiple recipients are stored as array
	uint32_t act_num_sender;
	uint32_t act_num_recipients;
	smtp_flow_output_t out;

	// managing structs:
	char *auth_type;
	uint32_t observed_commands;
	uint8_t errors;
	uint8_t type;
	uint8_t state;
} smtp_flow_t;


// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

// variables
unsigned long plugin_smtpDetector_Index;
smtp_flow_t *smtp_flow;

#endif /* SMTP_DETECTOR_H_ */

