/*
 * tcpStates.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "tcpStates.h"

char* get_plugin_name() {
	return "tcpStates";
}

char* get_plugin_version() {
	return "0.5.6";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
#ifdef __APPLE__
	mainHashMap = getMainHashMap();
#endif // __APPLE__
	pluginTCPStatesIndex = pPluginIndex;

	// malloc memory for the basic struct
	tcpStates_connection_table = malloc(mainHashMap->hashChainTableSize * sizeof(tcp_connection_states_t));
	memset(tcpStates_connection_table, '\0', mainHashMap->hashChainTableSize * sizeof(tcp_connection_states_t));

	// register timeouts
	float timeout_value = TIMEOUT_RESET;
	timeout_handler_add(timeout_value);

	timeout_value = TIMEOUT_NEW;
	timeout_handler_add(timeout_value);

	timeout_value = TIMEOUT_ESTABLISHED;
	timeout_handler_add(timeout_value);

	timeout_value = TIMEOUT_CLOSING;
	timeout_handler_add(timeout_value);

	timeout_value = TIMEOUT_CLOSED;
	timeout_handler_add(timeout_value);

}

#if TCPSTATES_PRINT_OUT != 0
binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;
	binary_value_t *act_bv = bv_new_bv("TCP aggregated protocol state flags", "tcpStates", 0, 1, bt_hex_8);
	main_bv = bv_append_bv(main_bv, act_bv);
	return main_bv;
}
#endif /* TCPSTATES_PRINT_OUT != 0 */

void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	// reset  entry
	memset(&tcpStates_connection_table[flowIndex], '\0', sizeof(tcp_connection_states_t));
}

/*
 * The general state machine approach, not optimized
 * WHY our state machine differs from the "normal" TCP state machine?
 * Because we're sitting somewhere in the middle.
 * This leads to several special cases like
 * - recognizing already opened connections
 * - getting not every packet
 * - seeing only on side of a connection
 * and the most important one:
 * - We don't know the behavior of the internal tcp state machines inside the hosts
 */
void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	if (packet->layer4Type != L3_TCP) return;
#if IPV6_ACTIVATE == 1
	if ( packet->ip6FragHdr && (packet->ip6FragHdr->frag_off & FRAG6ID_N) ) return; // only tcp and 1. frag packet will be processed
#else // IPV6_ACTIVATE == 0
	if ( (((ipHeader_t*)packet->layer3Header)->ip_off) & FRAGID_N ) return; // only tcp and 1. frag packet will be processed
#endif // IPV6_ACTIVATE
	flow_t *actFlowPtr = &flows[flowIndex];
	flow_t *actFlowPtrO;
	
	tcp_connection_states_t *tcpStates_ctP = &tcpStates_connection_table[flowIndex];
	tcp_connection_states_t *tcpStates_ctPO;
	
	if ( actFlowPtr->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) {
		actFlowPtrO = &flows[actFlowPtr->oppositeFlowIndex];
		tcpStates_ctPO = &tcpStates_connection_table[actFlowPtr->oppositeFlowIndex];
	} else {
		actFlowPtrO = NULL;
		tcpStates_ctPO = NULL; 
	}

	tcpHeader_t *tcpHeader = (tcpHeader_t*) &(packet->layer4Header->tcpHeader);

	// extract the TCP flags
	unsigned char tcpFlags =  *((unsigned char*)tcpHeader + 13);


	// Null scan and Christmas ++ scan
	if (((tcpFlags & TH_ALL_FLAGS) == TH_XMAS) || ((tcpFlags & TH_ALL_FLAGS) == TH_NULL)) tcpStates_ctP->bogus_bitField |= EVIL; // evil behavior


	switch (tcpStates_ctP->connection_state) {

	case STATE_NEW:

		// syn flag set
		if ((tcpFlags & TH_SYN_FIN_RST) == TH_SYN) {
			tcpStates_ctP->sender_syn_seen = 1;
			tcpStates_ctP->sender_syn_sequ_nr = ntohl(tcpHeader->seq);
			actFlowPtr->timeout = TIMEOUT_NEW;

			// if (tcpHeader->ack) {
			if ((tcpFlags & TH_ACK) == TH_ACK) {
				// packet acks a syn from the opposite flow
				if (actFlowPtrO) {
					// check ack number
					if (ntohl(tcpHeader->ack_seq) == tcpStates_ctPO->sender_syn_sequ_nr + 1) {
						tcpStates_ctPO->sender_syn_ackd = 1;
					} else {
						// ack number is wrong -> bogus connection establishment, go into STATE ESTABLISHED and set bogus bit zero
						tcpStates_ctP->connection_state = STATE_ESTABLISHED;
						actFlowPtr->timeout = TIMEOUT_ESTABLISHED;
						tcpStates_ctP->bogus_bitField |= MAL_CON_EST;

						tcpStates_ctPO->connection_state = STATE_ESTABLISHED;
						actFlowPtrO->timeout = TIMEOUT_ESTABLISHED;
						tcpStates_ctPO->bogus_bitField |= MAL_CON_EST;
					}

				} else {
					// There is no opposite flow, maybe we don't see it
					// -> set bogus flag zero and set the connection to established
					tcpStates_ctP->connection_state = STATE_ESTABLISHED;
					actFlowPtr->timeout = TIMEOUT_ESTABLISHED;
					tcpStates_ctP->bogus_bitField |= MAL_CON_EST;
				}
			}
			break;
		}

		// Last part of 3 way or simultaneous handshake
		// if (!tcpHeader->syn && tcpHeader->ack && !tcpHeader->fin && !tcpHeader->rst && actFlowPtr->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		if (actFlowPtrO && (tcpFlags & TH_ARSF) == TH_ACK) {
			if (ntohl(tcpHeader->ack_seq) == tcpStates_ctPO->sender_syn_sequ_nr + 1) {
				// correct ack
				tcpStates_ctPO->sender_syn_ackd = 1;
				// check if own syn packet was acked
				if (tcpStates_ctP->sender_syn_ackd == 1) {
					// connection is successfully established -> change to state ESTABLISHED
					tcpStates_ctP->connection_state = STATE_ESTABLISHED;
					actFlowPtr->timeout = TIMEOUT_ESTABLISHED;
					tcpStates_ctPO->connection_state = STATE_ESTABLISHED;
					actFlowPtrO->timeout = TIMEOUT_ESTABLISHED;
				}
			} else {
				// there's something wrong with the ack number -> set bogus flag 0 and set flow to state ESTABLISHED
				tcpStates_ctP->connection_state = STATE_ESTABLISHED;
				actFlowPtr->timeout = TIMEOUT_ESTABLISHED;
				tcpStates_ctP->bogus_bitField |= MAL_CON_EST; // malformed connection establishment

				tcpStates_ctPO->connection_state = STATE_ESTABLISHED;
				actFlowPtrO->timeout = TIMEOUT_ESTABLISHED;
				tcpStates_ctPO->bogus_bitField |= MAL_CON_EST; // malformed connection establishment
			}
			break;
		}

		/*
		 * rst, ack flags set, opposite flow
		 * Normal connection rejection
		 */
		// if (!tcpHeader->syn && tcpHeader->ack && !tcpHeader->fin && tcpHeader->rst && actFlowPtr->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		if (actFlowPtrO && ((tcpFlags & TH_ARSF) == TH_RST_ACK)) {
			// reset flow
			tcpStates_ctP->connection_state = STATE_RESET;
			actFlowPtr->timeout = TIMEOUT_RESET;
			tcpStates_ctP->bogus_bitField |= RST_TRANS; // Reset from sender seen
			// reset opposite flow
			tcpStates_ctPO->connection_state = STATE_RESET;
			actFlowPtrO->timeout = TIMEOUT_RESET;
			break;
		}

		// Every other combination with reset flag
		// if (tcpHeader->rst) {
		if ((tcpFlags & TH_RST) == TH_RST) {
			tcpStates_ctP->bogus_bitField |= (RST_TRANS | MAL_CON_EST); // Reset from sender seen, Malformed Connection estabilshment
			tcpStates_ctP->connection_state = STATE_RESET;
			actFlowPtr->timeout = TIMEOUT_RESET;

			if (actFlowPtrO) {
				// Some strange reset, set bogus flag 0
				tcpStates_ctPO->bogus_bitField |= MAL_CON_EST; // malformed connection establishment
				tcpStates_ctPO->connection_state = STATE_RESET;
				actFlowPtrO->timeout = TIMEOUT_RESET;
			} else {
				// Possible RST scan
				tcpStates_ctP->bogus_bitField |= EVIL; // possible evil behavior
			}
			break;
		}

		/*
		 * A flow starts with a teardown. Possible reasons:
		 * - We didn't see the previous packets of this connection.
		 * - A fin scan
		 */
		// if (tcpHeader->fin && tcpHeader->ack && !tcpHeader->syn && !tcpHeader->rst) {
		if ((tcpFlags & TH_ARSF) == TH_FIN_ACK) {
			tcpStates_ctP->sender_fin_seen = 1;
			tcpStates_ctP->sender_fin_sequ_nr = ntohl(tcpHeader->seq);
			// The connection establishment is definetly malicious or was not seen, set bogus bit field 0
			tcpStates_ctP->bogus_bitField |= MAL_CON_EST; // malformed connection establishment
			tcpStates_ctP->connection_state = STATE_CLOSING;
			actFlowPtr->timeout = TIMEOUT_CLOSING;

			if (actFlowPtrO) {
				tcpStates_ctPO->bogus_bitField |= MAL_CON_EST; // malformed connection establishment

				if ( tcpStates_ctPO->sender_fin_seen && (tcpStates_ctPO->sender_fin_sequ_nr + 1 == ntohl(tcpHeader->ack_seq)) ) {
					tcpStates_ctPO->sender_fin_ackd = 1;
				}
			} else {
				// No opposite flow and this fin packet is the first packet we see? That might be a fin scan
				tcpStates_ctP->sender_fin_scan = 1;
			}
			break;
		}

		// if (tcpHeader->fin && !tcpHeader->ack && !tcpHeader->syn && !tcpHeader->rst) {
		if ((tcpFlags & TH_ARSF) == TH_FIN) {
			// This IS a fin scan!
			tcpStates_ctP->bogus_bitField |= EVIL; // possible evil behavior
			break;
		}

		/*
		 * Every other combination is bogus.
		 * We set the state to ESTABLISHED, because there could be something interesting in the flows :)
		 * TODO: Distinction between more states
		 */
		tcpStates_ctP->connection_state = STATE_ESTABLISHED;
		actFlowPtr->timeout = TIMEOUT_ESTABLISHED;
		tcpStates_ctP->bogus_bitField |= MAL_CON_EST; // malformed connection establishment

		if (actFlowPtrO) {
			tcpStates_ctPO->connection_state = STATE_ESTABLISHED;
			actFlowPtrO->timeout = TIMEOUT_ESTABLISHED;
			tcpStates_ctPO->bogus_bitField |= MAL_CON_EST; // malformed connection establishment
		}
		break;

	case STATE_ESTABLISHED:
		// In this state should be no syn flag seen or no ack missing (even rst packets should ack)
		// if (tcpHeader->syn || !tcpHeader->ack) {
		if ((tcpFlags & TH_SYN_ACK) != TH_ACK) {
			tcpStates_ctP->bogus_bitField |= MAL_FLG2_EST; // Malformed flags during established connection
			if (actFlowPtrO)
				tcpStates_ctPO->bogus_bitField |= MAL_FLG2_EST; // Malformed flags during established connection
		}

		// sender initiates a teardown
		// if (tcpHeader->fin) {
		if ((tcpFlags & TH_FIN) == TH_FIN) {
			tcpStates_ctP->sender_fin_seen = 1;
			tcpStates_ctP->sender_fin_sequ_nr = ntohl(tcpHeader->seq);
			tcpStates_ctP->connection_state = STATE_CLOSING;
			actFlowPtr->timeout = TIMEOUT_CLOSING;

			if (actFlowPtrO && tcpHeader->ack
					&& tcpStates_ctPO->sender_fin_seen
					&& ntohl(tcpHeader->ack_seq) == tcpStates_ctPO->sender_fin_sequ_nr + 1) {
				tcpStates_ctPO->sender_fin_ackd = 1;
			}
		}

		// Connection has been reset
		// if (tcpHeader->rst) {
		if ((tcpFlags & TH_RST) == TH_RST) {
			tcpStates_ctP->connection_state = STATE_RESET;
			actFlowPtr->timeout = TIMEOUT_RESET;
			tcpStates_ctP->bogus_bitField |= (RST_TRANS | MAL_TEARDWN); // Reset from sender seen, Malformed Connection teardown

			if (actFlowPtrO) {
				tcpStates_ctPO->connection_state = STATE_RESET;
				actFlowPtrO->timeout = TIMEOUT_RESET;
				tcpStates_ctPO->bogus_bitField |= MAL_TEARDWN; // Malformed connection teardown
			}
		}
		break;

	case STATE_CLOSING:

		/*
		 * Connection has been resetted
		 */
		// if (tcpHeader->rst) {
		if ((tcpFlags & TH_RST) == TH_RST) {
			tcpStates_ctP->connection_state = STATE_RESET;
			actFlowPtr->timeout = TIMEOUT_RESET;
			tcpStates_ctP->bogus_bitField |= (RST_TRANS | MAL_TEARDWN); // Reset from sender seen, Malformed Connection teardown

			if (actFlowPtrO) {
				tcpStates_ctPO->connection_state = STATE_RESET;
				actFlowPtrO->timeout = TIMEOUT_RESET;
				tcpStates_ctPO->bogus_bitField |= MAL_TEARDWN; // Malformed connection teardown
			}
		}

		// if (tcpHeader->ack
		if ((tcpFlags & TH_ACK) == TH_ACK
				&& actFlowPtrO
				&& tcpStates_ctPO->sender_fin_seen
				&& ntohl(tcpHeader->ack_seq) == tcpStates_ctPO->sender_fin_sequ_nr + 1) {
			tcpStates_ctPO->sender_fin_ackd = 1;
		}

		// Test if teardown is complete
		if (tcpStates_ctP->sender_fin_seen && tcpStates_ctP->sender_fin_ackd
				&& actFlowPtrO
				&& tcpStates_ctPO->sender_fin_seen && tcpStates_ctPO->sender_fin_ackd) {
			tcpStates_ctP->connection_state = STATE_CLOSED;
			actFlowPtr->timeout = TIMEOUT_CLOSED;
			tcpStates_ctPO->connection_state = STATE_CLOSED;
			actFlowPtrO->timeout = TIMEOUT_CLOSED;
		}
		break;

	case STATE_CLOSED:
		// A "normal" connection don't executes this state
		tcpStates_ctP->bogus_bitField |= PKTS_TERM; // more packets from sender after connection closing seen

		// Connection has been reset
		// if (tcpHeader->rst) {
		if ((tcpFlags & TH_RST) == TH_RST) {
			tcpStates_ctP->connection_state = STATE_RESET;
			actFlowPtr->timeout = TIMEOUT_RESET;
			tcpStates_ctP->bogus_bitField |= (RST_TRANS | MAL_TEARDWN); // Reset from sender seen, Malformed Connection teardown

			if (actFlowPtrO) {
				tcpStates_ctPO->connection_state = STATE_RESET;
				actFlowPtrO->timeout = TIMEOUT_RESET;
				tcpStates_ctPO->bogus_bitField |= MAL_TEARDWN; // Malformed connection teardown
			}
		}
		break;

	case STATE_RESET:
		tcpStates_ctP->bogus_bitField |= PKTS_RST; // More packets after reset seen
		break;
	default:
		fprintf(stderr, "BUG in tcpStates.c: Unhandled state: %hhu\n", tcpStates_ctP->connection_state);
		break;
	}
}

#if TCPSTATES_PRINT_OUT != 0
void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	flow_t *actFlowPtr = &flows[flowIndex];

	if (actFlowPtr->layer4Protocol != L3_TCP) {
		outputBuffer_append(main_output_buffer, "\0", 1);
		return; // If it's not TCP, return
	}

	tcp_connection_states_t *tcpStates_ctP = &tcpStates_connection_table[flowIndex];
	tcp_connection_states_t *tcpStates_ctPO;

	if ( actFlowPtr->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) {
		tcpStates_ctPO = &tcpStates_connection_table[actFlowPtr->oppositeFlowIndex];
	} else {
		tcpStates_ctPO = NULL; 
	}
	
	// Test again if connection establishment and teardown where correctly performed
	if (tcpStates_ctPO) {
		if ( !(tcpStates_ctP->sender_syn_seen && tcpStates_ctP->sender_syn_ackd && tcpStates_ctPO->sender_syn_seen && tcpStates_ctPO->sender_syn_ackd) ) {
			tcpStates_ctP->bogus_bitField |= MAL_CON_EST; // Malformed connection establishment
		} 

		if ( !(tcpStates_ctP->sender_fin_seen && tcpStates_ctP->sender_fin_ackd && tcpStates_ctPO->sender_fin_seen && tcpStates_ctPO->sender_fin_ackd) ) {
			tcpStates_ctP->bogus_bitField |= MAL_TEARDWN; // Malformed connection teardown
		} else {
			// A correct teardown implies that no fin scan was performed (see state NEW)
			tcpStates_ctP->sender_fin_scan = 0;
			tcpStates_ctPO->sender_fin_scan = 0;
		}
	} else {
		tcpStates_ctP->bogus_bitField |= (MAL_CON_EST | MAL_TEARDWN); // Malformed connection establishment and teardown
	}

	// If the fin scan bit is still set, set the possible evil behavior bit
	if (tcpStates_ctP->sender_fin_scan) tcpStates_ctP->bogus_bitField |= EVIL;

	// print states info
	outputBuffer_append(main_output_buffer, (char*) &tcpStates_ctP->bogus_bitField, 1);
}
#endif // TCPSTATES_PRINT_OUT != 0

void onApplicationTerminate() {
	if (tcpStates_connection_table) free(tcpStates_connection_table);
	tcpStates_connection_table == NULL;
}


