/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * hashTable.c
 */
#include "hashTable.h"

#if HASHTABLE_DEBUG != 0
void hashTable_print(hashMap_t *hashMap);
#endif /* HASHTABLE_DEBUG != 0 */


/* initializes a hashMap */
/* the scale factor can be submitted to change the base size of the hash table */
hashMap_t *hashTable_initialize(float scaleFactor, unsigned long dataLength) {
	hashMap_t *hashMap = malloc(sizeof(hashMap_t));
	if (!hashMap) {
		fprintf(stdout, "Hashmap Error: Could not allocate more memory!\n");
		terminate();
	}

	if (scaleFactor<1.0f) {
		fprintf(stdout, "Hashmap Warning: Scale factor is smaller than 1.0!\n");
	}

	hashMap->hashTableSize = ceilf((float)HASHTABLE_BASE_SIZE*scaleFactor);
	hashMap->hashChainTableSize = ceilf((float)HASHCHAINTABLE_BASE_SIZE*scaleFactor);
	hashMap->hashTable = (hashBucket_t**) malloc(sizeof(hashBucket_t*)*hashMap->hashTableSize);
	hashMap->hashChainTable = (hashBucket_t*) malloc(sizeof(hashBucket_t)*hashMap->hashChainTableSize);
	hashMap->dataLength = dataLength;

	memset(hashMap->hashTable, '\0', sizeof(hashBucket_t*)*hashMap->hashTableSize);
	memset(hashMap->hashChainTable, '\0', sizeof(hashBucket_t)*hashMap->hashChainTableSize);

	for (hashCounter=0; hashCounter<hashMap->hashChainTableSize; hashCounter++) { // note the missing 'minus one'
		hashMap->hashChainTable[hashCounter].data = (char*) malloc(dataLength);
	}
	for (hashCounter=0; hashCounter<hashMap->hashChainTableSize-1; hashCounter++) {
		hashMap->hashChainTable[hashCounter].nextBucket = &hashMap->hashChainTable[hashCounter+1];
	}
	hashMap->freeBucket = &hashMap->hashChainTable[0];
	hashMap->freeListSize = hashMap->hashChainTableSize;

	currentBucket = NULL;
	previousBucket = NULL;

#if HASHTABLE_DEBUG != 0
	fprintf(stdout, "HashTable after init: -------------------------------------\n");
	hashTable_print(hashMap);
	fprintf(stdout, "-----------------------------------------------------------\n");
#endif /* HASHTABLE_DEBUG != 0 */

	return hashMap; /* successfully return */
}

void hashTable_destroy(hashMap_t *hashMap) {
	if (!hashMap) return;
	for (hashCounter=0; hashCounter<hashMap->hashChainTableSize; hashCounter++) {
		if (hashMap->hashChainTable[hashCounter].data) free(hashMap->hashChainTable[hashCounter].data); hashMap->hashChainTable[hashCounter].data = NULL;
	}
	if (hashMap->hashChainTable) free(hashMap->hashChainTable); hashMap->hashChainTable = NULL;
	if (hashMap->hashTable) free(hashMap->hashTable); hashMap->hashTable = NULL;
	free(hashMap); hashMap = NULL; // NOTE: original pointer still != NULL!!!!
}



/* looks, if a combination of all parameters is already stored in the hashMap,
   if not HASHTABLE_ENTRY_NOT_FOUND is returned */
unsigned long hashTable_lookup(hashMap_t *hashMap, char *data) {
	if (!hashMap) return HASHTABLE_ENTRY_NOT_FOUND;

	currentBucket = hashMap->hashTable[hashTable_hash(hashMap->dataLength, data) % hashMap->hashTableSize];
	while (currentBucket) {
		if (memcmp(currentBucket->data, data, hashMap->dataLength) == 0) {
			return (currentBucket - &hashMap->hashChainTable[0]);
		}
		currentBucket = currentBucket->nextBucket;
	}
	return HASHTABLE_ENTRY_NOT_FOUND;
}


/* the function that generates a hash value, used by the other functions */
unsigned long hashTable_hash(unsigned long dataLength, char *data) {

	uint32_t hash_value_r = 0;
	size_t hash_i_r;

	for (hash_i_r = 0; hash_i_r < dataLength; hash_i_r++) {
		hash_value_r += data[hash_i_r];
		hash_value_r += (hash_value_r << 10);
		hash_value_r ^= (hash_value_r >> 6);
	}

	hash_value_r += (hash_value_r << 3);
	hash_value_r ^= (hash_value_r >> 11);
	hash_value_r += (hash_value_r << 15);

	return hash_value_r;
}

/* inserts a combination of all parameters in the hashMap */
unsigned long hashTable_insert(hashMap_t *hashMap, char *data) {

	if (!hashMap) return HASHTABLE_ENTRY_NOT_FOUND;
	if (hashMap->freeListSize <= 0) {
		fprintf(stdout, "ERROR: Hashtable full! Increase size of HASHTABLE_BASE_SIZE &| HASHCHAINTABLE_BASE_SIZE in tranalyzer.h.\n");
		terminate(); // Hashmap full means: Give me a larger hash, so stop process!
		return HASHTABLE_ENTRY_NOT_FOUND;
	}

	unsigned long hashValue = hashTable_hash(hashMap->dataLength, data) % hashMap->hashTableSize;
	
	/* take a free bucket from the front of the free list */
	currentBucket = hashMap->freeBucket;
	hashMap->freeBucket = hashMap->freeBucket->nextBucket;

	/* point the nextPointer on the current first bucket */
	currentBucket->nextBucket = hashMap->hashTable[hashValue];

	/* place the current bucket at the front of the hashTable */
	hashMap->hashTable[hashValue] = currentBucket;

	/* fill it with the right values */
	memcpy(currentBucket->data, data, hashMap->dataLength);

	hashMap->freeListSize--;

#if HASHTABLE_DEBUG != 0
	fprintf(stdout, "HashTable after insert: -----------------------------------\n");
	hashTable_print(hashMap);
	fprintf(stdout, "-----------------------------------------------------------\n");
#endif /* HASHTABLE_DEBUG != 0 */

	return currentBucket - &hashMap->hashChainTable[0];
}

/* removes a combination of all parameters in the hashMap */
unsigned long hashTable_remove(hashMap_t *hashMap, char *data) {

	if (!hashMap) return -1;
	unsigned long hashValue = hashTable_hash(hashMap->dataLength, data) % hashMap->hashTableSize;
	currentBucket = hashMap->hashTable[hashValue];
	previousBucket = currentBucket;
	while (currentBucket) {
		if (memcmp(currentBucket->data, data, hashMap->dataLength) == 0) {

			/* set the pointers of the hashBucket*/
			if (previousBucket == currentBucket) {
				hashMap->hashTable[hashValue] = currentBucket->nextBucket;
			} else {
				previousBucket->nextBucket = currentBucket->nextBucket;
			}

			/* clean Bucket */
			currentBucket->nextBucket = NULL;
			//memset(currentBucket, '\0', sizeof(hashBucket_t));

			/* reinsert bucket into free list */
			currentBucket->nextBucket = hashMap->freeBucket;
			hashMap->freeBucket = currentBucket;

			hashMap->freeListSize++;

#if HASHTABLE_DEBUG != 0
			fprintf(stdout, "HashTable after remove: -----------------------------------\n");
			hashTable_print(hashMap);
			fprintf(stdout, "-----------------------------------------------------------\n");
#endif /* HASHTABLE_DEBUG != 0 */
			return 0;
		}
		previousBucket = currentBucket;
		currentBucket = currentBucket->nextBucket;
	}
	return HASHTABLE_ENTRY_NOT_FOUND; /* Entry could not be found and therefore wasn't removed */
}

#if HASHTABLE_DEBUG != 0
void hashTable_print(hashMap_t *hashMap) {
	long i = 0, j = 0;
	for (i = 0; i < hashMap->hashTableSize; i++) {
		fprintf(stdout, "%ld:", i);
		j = 0;
		debugBucket = hashMap->hashTable[i];
		while (debugBucket) {
			j++;
			debugBucket = debugBucket->nextBucket;
		}
		fprintf(stdout, "%ld\n", j);
	}
}
#endif /* HASHTABLE_DEBUG != 0 */

