/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * hashTable.h
 *
 * Header file for the HashTable implementation
 *
 * The principal design of this hashTable implementation is as follows:
 * It is a closed hashTable, that means the hashing function returns a
 * bucketnumber which points on an attached list of bucket entries. Each of
 * the bucket entries contains a - hashTable wide - unique number between zero
 * and HASHCHAINTABLESIZE, which can be used for example to access an array. The
 * collection of the free numbers is done by the freeList. Each plugin can
 * define its own hashTable if necessary and use the functions listed below.
 * PLEASE NOTE:
 * The mapping of the hashTable entries on the unique numbers is not explicit!
 * If an entry is removed out of the hashTable and re-inserted later, it will
 * get a different unique number with big probability!
 */

#ifndef HASHTABLE_H_
#define HASHTABLE_H_

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "tranalyzer.h"
#include <inttypes.h>

/* a mix function used by the hashing function */
#define mix(a, b, c) { \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}

/* a debug flag to enable extreme debug */
#define HASHTABLE_DEBUG 0

// functions
extern void terminate();

/* structs */

// The structure for the hashBuckets, 19 Bytes -> 20 Bytes allocated
typedef struct hashBucket {
	struct hashBucket *nextBucket;
	char *data;
} hashBucket_t;

/*typedef struct hashData_s {
	in_addr_t ip_src; // 4 
	in_addr_t ip_dst; // 4
	uint16_t vlan_id; // 2 eigentlich 1.5
	uint16_t port_src; // 2
	uint16_t port_dst; // 2
       	u_int8_t protocol; // 1
} hashData_t;

typedef union hashBucketData_u {
	char data[15];
	hashData_t hashDatas;
} hashBucketData_t;
*/

// The structure to collect all necessary information for a single hashMap
typedef struct hashMap_s {
	unsigned long hashTableSize;
	unsigned long hashChainTableSize;
	hashBucket_t **hashTable;
	hashBucket_t *hashChainTable;
	hashBucket_t *freeBucket; // points on the first free bucket
	unsigned long freeListSize;
	unsigned long dataLength;
} hashMap_t;

/* variables */

hashBucket_t *currentBucket, *previousBucket; /* commonly used pointer to bucket */
#if HASHTABLE_DEBUG != 0
hashBucket_t *debugBucket;
#endif /* HASHTABLE_DEBUG != 0 */
unsigned long hashCounter; /* commonly used counter */
uint32_t hash_value;
size_t hash_i;


/* functions */
// initializes a hashMap
hashMap_t *hashTable_initialize(float scaleFactor, unsigned long dataLength);

// looks, if a combination of all parameters is already stored in the hashMap,
// if not HASHTABLE_ENTRY_NOT_FOUND is returned
unsigned long hashTable_lookup(hashMap_t *hashMap, char *data);

// the function that generates a hash value, used by the other functions
unsigned long hashTable_hash(unsigned long dataLength, char *data);

// inserts a combination of all parameters in the hashMap
unsigned long hashTable_insert(hashMap_t *hashMap, char *data);

// removes a combination of all parameters in the hashMap
unsigned long hashTable_remove(hashMap_t *hashMap, char *data);

// completely destroys a hash table
void hashTable_destroy(hashMap_t *hashMap);

#endif /* HASHTABLE_H_ */
