/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
#include "tranalyzer.h"
#if ENABLE_IO_BUFFERING != 0
#include "ioBuffer.h"

void ioBufferInitialize() {
	threadTerminated = 0;
	
	// create posix thread
	if (pthread_create(&ioBufferThread, NULL, ioBufferThreadFunction, NULL)) {
		fprintf(stderr, "[ioBuffer] ERROR: Could not create ioBuffer thread!\n");
		terminate();
	}
}

void *ioBufferThreadFunction(void *args) {
	int pcapDispatchReturnCode;
	if (captureListFileName && VERBOSE != 0) {
		fprintf(stdout, "processing file no. %lu of %lu: %s\n", act_capture_file_index + 1, capture_file_list->number_of_files, act_capture_list_element->file_name);
	}
	while (!threadTerminated && running) {
		// dispatch
		pcapDispatchReturnCode = pcap_dispatch(captureDescriptor, PACKETS_PER_BURST, ioBufferPush, NULL);
		switch (pcapDispatchReturnCode) {
			case -1:
				// An error occured, terminate
				running = 0;
				break;
			case 0:
				// Terminate if we are reading from a file rather than live-sniffing
				if (captureFileName) {
					threadTerminated = 1;
				} else if (captureListFileName) {
					// set descriptor to next file
					if (act_capture_list_element->next) {
						pcap_close(captureDescriptor);
						act_capture_list_element = act_capture_list_element->next;
						act_capture_file_index++;
						if ((captureDescriptor = pcap_open_offline(act_capture_list_element->file_name, errbuf)) == NULL) {
							fprintf(stderr, "pcap open %s: %s\n", act_capture_list_element->file_name, errbuf);
							running = 0;
						}
						if (VERBOSE != 0) {
							fprintf(stdout, "processing file no. %lu of %lu: %s\n", act_capture_file_index + 1, capture_file_list->number_of_files, act_capture_list_element->file_name);
						}
					} else {
						// there is no next file -> terminate
						threadTerminated = 1;
					}
				} else {
					// reading from live-interface. As we're using non-blocking mode, pcap_dispatch returns zero immediately
					// if no packets are to be read at the moment. This would cost much cpu consumption when no packets can be read,
					// therefore sleep some time
					usleep(NO_PACKETS_WAIT_MICROSEC);
				}
				break;
		}
	}
	return NULL;
}

void ioBufferPush(u_char* inqueue_uchar, const struct pcap_pkthdr *pcapHeader, const u_char* packet) {
	unsigned long nextElement = (writePos + 1) % IO_BUFFER_SIZE;
	while (nextElement == readPos && running) {
		usleep(IO_BUFFER_QUEUE_FULL_WAIT_MICROSEC);
	}
	memcpy((struct pcap_pkthdr*) &headerBuffer[writePos], pcapHeader, sizeof(struct pcap_pkthdr));
	memcpy((mtuPacket_t*) &packetBuffer[writePos], packet, pcapHeader->caplen);
	writePos = nextElement;
}


int ioBufferPop(struct pcap_pkthdr **header, u_char **packet) {

	if (threadTerminated) {
		if (readPos == writePos) {
			running = 0;
			return 0;
		}
	} else {
		while (readPos == writePos && !threadTerminated && running) {
			usleep(IO_BUFFER_QUEUE_FULL_WAIT_MICROSEC);
		}
		if (threadTerminated) {
			running = 0;
			return 0;
		}
	}
	
	unsigned long nextElement = (readPos + 1) % IO_BUFFER_SIZE;
	*packet = (u_char*) &packetBuffer[readPos];
	*header = (struct pcap_pkthdr*) &headerBuffer[readPos];
	readPos = nextElement;
	return 1;
}

void ioBufferTerminate() {
	// dummy at the moment
}
      

#endif

