/*
 * loadTranalyzerPlugins.c
 */

#include "loadTranalyzerPlugins.h"
#include "main.h"
#include "networkHeaders.h"
#include "tranalyzer.h"
#include "outputBuffer.h"

#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <dirent.h>
#include <errno.h>
#include <string.h>

tranalyzer_plugin_array_t* load_tranalyzer_plugins(char* folder) {

	// init output array and fill with zero values
	tranalyzer_plugin_array_t *t_array = malloc(sizeof(tranalyzer_plugin_array_t));
	memset(t_array, '\0', sizeof(tranalyzer_plugin_array_t));

	struct dirent *de;
	DIR* directory;

	tranalyzer_plugin_list_element_t *list_start = NULL;
	tranalyzer_plugin_list_element_t *list_act_element = NULL;
	tranalyzer_plugin_list_element_t *list_new_element = NULL;

	void *tempHandle;
	char* (*tranalyzer_plugin_name)();
	char* (*tranalyzer_plugin_version)();
	unsigned int (*supported_tranalyzerVersion_major)();
	unsigned int (*supported_tranalyzerVersion_minor)();

	char* (*get_dependencies)();

	void (*initialize_func)(unsigned long pPluginIndex);
	binary_value_t* (*printHeader_func)();
	void (*onFlowGenerated_func)(packet_t *packet, unsigned long flowIndex);
	void (*claimLayer2Information_func)(packet_t *packet);
	void (*claimLayer3Information_func)(packet_t *packet);
	void (*claimLayer4Information_func)(packet_t *packet, unsigned long flowIndex);
	void (*onFlowTerminate_func)(unsigned long flowIndex);
	void (*onApplicationTerminate_func)();
	void (*bufferToSink_func)(outputBuffer_t* buffer);


	// init counters
	unsigned int num_of_plugins;
	unsigned int num_of_func_pluginInitialize;
	unsigned int num_of_func_pluginPrintHeader;
	unsigned int num_of_func_pluginOnFlowGenerated;
	unsigned int num_of_func_pluginClaimLayer2Information;
	unsigned int num_of_func_pluginClaimLayer3Information;
	unsigned int num_of_func_pluginClaimLayer4Information;
	unsigned int num_of_func_pluginOnFlowTerminate;
	unsigned int num_of_func_pluginOnApplicationTerminate;
	unsigned int num_of_func_pluginBufferToSink;

	// reset counters
	num_of_plugins = 0;
	num_of_func_pluginInitialize = 0;
	num_of_func_pluginPrintHeader = 0;
	num_of_func_pluginOnFlowGenerated = 0;
	num_of_func_pluginClaimLayer2Information = 0;
	num_of_func_pluginClaimLayer3Information = 0;
	num_of_func_pluginClaimLayer4Information = 0;
	num_of_func_pluginOnFlowTerminate = 0;
	num_of_func_pluginOnApplicationTerminate = 0;
	num_of_func_pluginBufferToSink = 0;

	// Build list of tranalyzer plugins

	directory = opendir(folder);
	if (directory == NULL) {
		if (errno == EACCES) printf("Permission denied.\n");
		if (errno == EMFILE) printf("Too many file descriptors in use by process.\n");
		if (errno == ENFILE) printf("Too many files are currently open in the system.\n");
		if (errno == ENOENT) printf("Plugin directory does not exist.\n");
		if (errno == ENOMEM) printf("Insufficient memory to complete the operation.\n");
		if (errno == ENOTDIR) printf("%s is not a directory.\n", folder);
		return t_array;
	}

	// build list of directory entries in alphabetical order
	dirent_list_entry_t *de_list_start = NULL;
	dirent_list_entry_t *de_list_act = NULL;
	dirent_list_entry_t *de_list_prev = NULL;
	while ((de = readdir(directory)) != NULL) {
#if DEBUG != 0
		fprintf(stdout, "loadTranalyzerPlugins: processing dir entry '%s'...\n", de->d_name);
#endif /* DEBUG != 0 */
		de_list_act = de_list_start;
		de_list_prev = de_list_start;
		while (de_list_act != NULL) {
			// fprintf(stdout, "check '%s'\n", de_list_act->de->d_name);
			if (strcmp(de_list_act->de->d_name, de->d_name) > 0) {
				break;
			}
			de_list_prev = de_list_act;
			de_list_act = de_list_act->next;
		}
		// if (de_list_act != NULL) fprintf(stdout, "insert before '%s'\n", de_list_act->de->d_name);
		// else fprintf(stdout, "insert at end of list\n");
		dirent_list_entry_t *de_list_new = malloc(sizeof(dirent_list_entry_t));
		de_list_new->de = de;
		de_list_new->next = NULL;

		if (de_list_start == NULL) {
			de_list_start = de_list_new;
		}
		else {
			if (de_list_prev == de_list_act) {
				de_list_new->next = de_list_start;
				de_list_start = de_list_new;
			} else {
				de_list_prev->next = de_list_new;
				de_list_new->next = de_list_act;
			}
		}
	}

	/*
	// debug print out;
	de_list_act = de_list_start;

	fprintf(stdout, "Directory entries:\n");
	while (de_list_act != NULL) {
		fprintf(stdout, "   '%s'\n", de_list_act->de->d_name);
		de_list_act = de_list_act->next;
	}
	fprintf(stdout, "\n");
	*/

	// next if clause should never be true, but to be shure...
	if (de_list_start == NULL) {
		fprintf(stdout, "folder %s is empty, this could be a bug!\n", folder);
		return t_array;
	}

#if DEBUG > 0
	de_list_act = de_list_start;
	printf("\nPlugin directory content:\n---------------------\n");
	while(de_list_act != NULL) {
		printf("%s\n", de_list_act->de->d_name);
		de_list_act = de_list_act->next;
	}
	printf("---------------------\n");
#endif /* DEBUG != 0 */

	// because we're using the continue command in the next loop, we build a
	// dummy list entry to have an entry point for the list processing
	de_list_prev = malloc(sizeof(dirent_list_entry_t));
	de_list_prev->next = de_list_start;
	de_list_act = de_list_prev;
	while (de_list_act != NULL) {

		// go to next list entry
		de_list_prev = de_list_act;
		de_list_act = de_list_act->next;
		// free(de_list_prev->de);
		free(de_list_prev);
		if (de_list_act == NULL) break; // last list entry was already processed

		if ((strncmp(de_list_act->de->d_name,".",1) == 0) || (strncmp(de_list_act->de->d_name,"..",2) == 0)) {
#if DEBUG > 0
			fprintf(stdout, "skipping %s\n", de_list_act->de->d_name);
#endif /* VERBOSE >= 2 */
			continue;
		}
#if VERBOSE >= 2
#endif /* VERBOSE >= 2 */
#if DEBUG != 0
#endif /* DEBUG != 0 */

		// check if directory entry is a dynamic library
#if VERBOSE >= 2
		fprintf(stdout, "checking if file %s is a valid tranalyzer plugin...\n", de_list_act->de->d_name);
#endif /* VERBOSE >= 2 */
		char temp_file_name[1024];
		strncpy(temp_file_name, folder, 1024 - MAXNAMLEN - 1);
		// strncat(temp_file_name, "/", 1);
		strncat(temp_file_name, de_list_act->de->d_name, MAXNAMLEN);
		if ((tempHandle = dlopen(temp_file_name, RTLD_GLOBAL|RTLD_LAZY)) == NULL) {
#if VERBOSE >= 2
			fprintf(stdout, "no.\n");
#endif /* VERBOSE >= 2 */
#if DEBUG != 0
			fprintf(stdout, "Error during opening the directory entry:\n%s\n", dlerror());
#endif /* DEBUG != 0 */
			continue; // directory entry is not a dynamic library
		}

		// try getting the name of the plugin
		if ((tranalyzer_plugin_name = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_GET_PLUGIN_NAME)) == NULL) {
#if VERBOSE >= 2
			fprintf(stdout, "no.\n");
#endif /* VERBOSE >= 2 */
#if DEBUG != 0
			fprintf(stdout, "Error during obtaining the address of the symbol for the TRANALYZER_PLUGIN_FUNCTION_GET_PLUGIN_NAME:\n%s\n", dlerror());
#endif /* DEBUG != 0 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}
#if VERBOSE >= 2
		fprintf(stdout, "Plugin name: %s.\n", (*tranalyzer_plugin_name)());
#endif /* VERBOSE >= 2 */

		// try getting the version of the plugin
		if ((tranalyzer_plugin_version = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_GET_PLUGIN_VERSION)) == NULL) {
#if VERBOSE >= 2
			fprintf(stdout, "no.\n");
#endif /* VERBOSE >= 2 */
#if DEBUG != 0
			fprintf(stdout, "Error during obtaining the address of the symbol for the TRANALYZER_PLUGIN_FUNCTION_GET_PLUGIN_VERSION:\n%s\n", dlerror());
#endif /* DEBUG != 0 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}
#if VERBOSE >= 2
		fprintf(stdout, "Plugin version: %s.\n", (*tranalyzer_plugin_version)());
#endif /* VERBOSE >= 2 */

		// try getting the supported version number
		if ((supported_tranalyzerVersion_major = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_GET_VERSION_MAJOR)) == NULL) {
#if VERBOSE >= 2
			fprintf(stdout, "no.\n");
#endif /* VERBOSE >= 2 */
#if DEBUG != 0
			fprintf(stdout, "Error during obtaining the address of the symbol for the TRANALYZER_PLUGIN_FUNCTION_GET_VERSION_MAJOR:\n%s\n", dlerror());
#endif /* DEBUG != 0 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}

		if ((supported_tranalyzerVersion_minor = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_GET_VERSION_MINOR)) == NULL) {
#if VERBOSE >= 2
			fprintf(stdout, "no.\n");
#endif /* VERBOSE >= 2 */
#if DEBUG != 0
			fprintf(stdout, "Error during obtaining the address of the symbol for the TRANALYZER_PLUGIN_FUNCTION_GET_VERSION_MINOR:\n%s\n", dlerror());
#endif /* DEBUG != 0 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}

#if DEBUG > 0
		fprintf(stdout, "\nprocessing %s...\n", tranalyzer_plugin_name());
#endif
#if DEBUG != 0
		fprintf(stdout, "checking supported tranalyzer version...\n");
#endif /* DEBUG != 0 */
		if (TRANALYZER_PLUGIN_SUPPORTED_VERSION_MAJOR > supported_tranalyzerVersion_major()) {
#if VERBOSE >= 2
			fprintf(stdout, "no, supported version number is too low.\n");
#endif /* VERBOSE >= 2 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}

		// TODO: substitute with major version number
		if (0 < supported_tranalyzerVersion_major()) {
#if VERBOSE >= 2
			fprintf(stdout, "no, supported version number is too high.\n");
#endif /* VERBOSE >= 2 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}


		if ((TRANALYZER_PLUGIN_SUPPORTED_VERSION_MAJOR == (*supported_tranalyzerVersion_major)()) &&
				(TRANALYZER_PLUGIN_SUPPORTED_VERSION_MINOR > (*supported_tranalyzerVersion_minor)())) {
#if VERBOSE >= 2
			fprintf(stdout, "no, supported version number is too low.\n");
#endif /* VERBOSE >= 2 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}

		// TODO: substitute with major and minor version number
		if ((TRANALYZER_PLUGIN_CURRENT_VERSION_MAJOR == (*supported_tranalyzerVersion_major)())
				&& (TRANALYZER_PLUGIN_CURRENT_VERSION_MINOR < (*supported_tranalyzerVersion_minor)())) {
#if VERBOSE >= 2
			fprintf(stdout, "no, supported version number is too low.\n");
#endif /* VERBOSE >= 2 */
			continue; // dynamic library is not a valid tranalyzer plugin
		}

		// check which functions are available
#if DEBUG != 0
		fprintf(stdout, "checking plugin function symbols...\n");
#endif /* DEBUG != 0 */


#if DEBUG != 0
		fprintf(stdout, "checking function 'initialize'...");
#endif /* DEBUG != 0 */

		if ((initialize_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_INITIALIZE)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginInitialize++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'printHeader'...");
#endif /* DEBUG != 0 */

		if ((printHeader_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_PRINT_HEADER)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginPrintHeader++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'onFlowGenerated'...");
#endif /* DEBUG != 0 */

		if ((onFlowGenerated_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_ON_FLOW_GENERATED)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginOnFlowGenerated++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'claimLayer2Information'...");
#endif /* DEBUG != 0 */

		if ((claimLayer2Information_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_CLAIM_LAYER_2_INFORMATION)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginClaimLayer2Information++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'claimLayer3Information'...");
#endif /* DEBUG != 0 */

		if ((claimLayer3Information_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_CLAIM_LAYER_3_INFORMATION)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginClaimLayer3Information++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'claimLayer4Information'...");
#endif /* DEBUG != 0 */

		if ((claimLayer4Information_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_CLAIM_LAYER_4_INFORMATION)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginClaimLayer4Information++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'onFlowTerminate'...");
#endif /* DEBUG != 0 */

		if ((onFlowTerminate_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_ON_FLOW_TERMINATE)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginOnFlowTerminate++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'onApplictationTerminate'...");
#endif /* DEBUG != 0 */

		if ((onApplicationTerminate_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_ON_APPLICATION_TERMINATE)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginOnApplicationTerminate++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


#if DEBUG != 0
		fprintf(stdout, "checking function 'bufferToSink'...");
#endif /* DEBUG != 0 */

		if ((bufferToSink_func = dlsym(tempHandle, TRANALYZER_PLUGIN_FUNCTION_BUFFER_TO_SINK)) != NULL) {
#if DEBUG != 0
			fprintf(stdout, "available.\n");
#endif /* DEBUG != 0 */
			num_of_func_pluginBufferToSink++;
		} else {
#if DEBUG != 0
			fprintf(stdout, "not available.\n");
#endif /* DEBUG != 0 */
		}


		// end of check which functions are available

#if VERBOSE >= 2
		fprintf(stdout, "valid plugin, adding it to plugin list... ");
#endif /* VERBOSE >= 2 */
		// build and fill a new list element

		list_new_element = malloc(sizeof(tranalyzer_plugin_list_element_t));
		memset(list_new_element, '\0', sizeof(tranalyzer_plugin_list_element_t));

		list_new_element->plugin.handle = tempHandle;

		strncpy(list_new_element->plugin.name, (*tranalyzer_plugin_name)(), TRANALYZER_PLUGIN_NAME_MAX_LENGTH - 1);
		list_new_element->plugin.name[TRANALYZER_PLUGIN_NAME_MAX_LENGTH - 1] = '\0';

		strncpy(list_new_element->plugin.version, (*tranalyzer_plugin_version)(), TRANALYZER_PLUGIN_VERSION_MAX_LENGTH - 1);
		list_new_element->plugin.version[TRANALYZER_PLUGIN_VERSION_MAX_LENGTH - 1] = '\0';

		list_new_element->plugin.supported_tranalyzer_version_major = (*supported_tranalyzerVersion_major)();
		list_new_element->plugin.supported_tranalyzer_version_minor = (*supported_tranalyzerVersion_minor)();
		list_new_element->plugin.initialize = initialize_func;
		list_new_element->plugin.printHeader = printHeader_func;
		list_new_element->plugin.onFlowGenerated = onFlowGenerated_func;
		list_new_element->plugin.claimLayer2Information = claimLayer2Information_func;
		list_new_element->plugin.claimLayer3Information = claimLayer3Information_func;
		list_new_element->plugin.claimLayer4Information = claimLayer4Information_func;
		list_new_element->plugin.onFlowTerminate = onFlowTerminate_func;
		list_new_element->plugin.onApplicationTerminate = onApplicationTerminate_func;
		list_new_element->plugin.bufferToSink = bufferToSink_func;

		if (list_start == NULL) {
			list_start = list_new_element;
			list_act_element = list_start;
		} else {
			list_act_element->next = list_new_element;
			list_act_element = list_act_element->next;
		}

		num_of_plugins++;

#if VERBOSE >= 2
		fprintf(stdout, "done.\n");
#endif /* VERBOSE >= 2 */
	}

	if (num_of_plugins == 0) {
		fprintf(stderr, "ERROR: No plugins found in folder %s!\n", folder);
		return t_array;
	}


	// check dependencies between plugins
#if VERBOSE >= 2
	fprintf(stdout, "checking dependencies...\n");
#endif /* VERBOSE >= 2 */
	list_act_element = list_start;

	char deps[MAXNAMLEN];
	while (list_act_element != NULL) {
		if ((get_dependencies = dlsym(list_act_element->plugin.handle, TRANALYZER_PLUGIN_FUNCTION_GET_DEPENDENCIES)) != NULL) {
			strncpy(deps, (*get_dependencies)(), MAXNAMLEN - 1);
			char* token = strtok(deps, ",");
			while (token != NULL) {
				int dep_found = 0;
				tranalyzer_plugin_list_element_t *list_temp_element = list_start;
				while (list_temp_element != NULL) {
					if (strcmp(token, list_temp_element->plugin.name) == 0) {
						dep_found = 1;
						break;
					}
					list_temp_element = list_temp_element->next;
				}

				if (dep_found == 0) {
					// dependency wasn't found, so we have to take out this plugin
#if DEBUG != 0
					fprintf(stdout, "%s: dependency '%s' wasn't found\n", list_act_element->plugin.name, token);
#endif /* DEBUG != 0 */
					if (list_act_element == list_start) {
						list_start = list_act_element->next;
					} else {
						list_temp_element = list_start;
						while (list_temp_element->next != list_act_element) list_temp_element = list_temp_element->next;
						list_temp_element->next = list_act_element->next;
					}
#if DEBUG != 0
					fprintf(stdout, "deleting %s\n", list_act_element->plugin.name);
#endif /* DEBUG != 0 */
					if (list_act_element->plugin.initialize) num_of_func_pluginInitialize--;
					if (list_act_element->plugin.printHeader) num_of_func_pluginPrintHeader--;
					if (list_act_element->plugin.onFlowGenerated) num_of_func_pluginOnFlowGenerated--;
					if (list_act_element->plugin.claimLayer2Information) num_of_func_pluginClaimLayer2Information--;
					if (list_act_element->plugin.claimLayer3Information) num_of_func_pluginClaimLayer3Information--;
					if (list_act_element->plugin.claimLayer4Information) num_of_func_pluginClaimLayer4Information--;
					if (list_act_element->plugin.onFlowTerminate) num_of_func_pluginOnFlowTerminate--;
					if (list_act_element->plugin.onApplicationTerminate) num_of_func_pluginOnApplicationTerminate--;
					if (list_act_element->plugin.bufferToSink) num_of_func_pluginBufferToSink--;

					dlclose(list_act_element->plugin.handle);
					free(list_act_element);

					num_of_plugins--;
#if DEBUG != 0
					fprintf(stdout, "resetting dependency check\n");
#endif /* VERBOSE >= 2 */
					// reset set list_act_element to the start, because we have to do the whole check again
					list_act_element = list_start;
					break;
				} else {
#if DEBUG != 0
					fprintf(stdout, "%s: dependency '%s' found\n", list_act_element->plugin.name, list_temp_element->plugin.name);
#endif /* DEBUG != 0 */
				}
				token = strtok(NULL, ",");
			}
			list_act_element = list_act_element->next;
		} else {
			list_act_element = list_act_element->next;
		}
	}

	if (num_of_plugins == 0) {
		fprintf(stderr, "ERROR: No plugins left after dependency check!\n");
		return t_array;
	}

	// Now fill array
	t_array->num_of_plugins = num_of_plugins;
	t_array->plugin = malloc(t_array->num_of_plugins * sizeof(tranalyzer_plugin_t));

	list_act_element = list_start;

	int i;
	for(i = 0; i < t_array->num_of_plugins; i++) {
		t_array->plugin[i].handle = list_act_element->plugin.handle;
		strncpy(t_array->plugin[i].name, list_act_element->plugin.name, TRANALYZER_PLUGIN_NAME_MAX_LENGTH);
		strncpy(t_array->plugin[i].version, list_act_element->plugin.version, TRANALYZER_PLUGIN_NAME_MAX_LENGTH);
		t_array->plugin[i].supported_tranalyzer_version_major = list_act_element->plugin.supported_tranalyzer_version_major;
		t_array->plugin[i].supported_tranalyzer_version_minor = list_act_element->plugin.supported_tranalyzer_version_minor;
		t_array->plugin[i].initialize = list_act_element->plugin.initialize;
		t_array->plugin[i].printHeader = list_act_element->plugin.printHeader;
		t_array->plugin[i].onFlowGenerated = list_act_element->plugin.onFlowGenerated;
		t_array->plugin[i].claimLayer2Information = list_act_element->plugin.claimLayer2Information;
		t_array->plugin[i].claimLayer3Information = list_act_element->plugin.claimLayer3Information;
		t_array->plugin[i].claimLayer4Information = list_act_element->plugin.claimLayer4Information;
		t_array->plugin[i].onFlowTerminate = list_act_element->plugin.onFlowTerminate;
		t_array->plugin[i].onApplicationTerminate = list_act_element->plugin.onApplicationTerminate;
		t_array->plugin[i].bufferToSink = list_act_element->plugin.bufferToSink;

		list_new_element = list_act_element;
		list_act_element = list_act_element->next;
		free(list_new_element);
	}

	// fill counters
	t_array->num_of_func_pluginInitialize = num_of_func_pluginInitialize;
	t_array->num_of_func_pluginPrintHeader = num_of_func_pluginPrintHeader;
	t_array->num_of_func_pluginOnFlowGenerated = num_of_func_pluginOnFlowGenerated;
	t_array->num_of_func_pluginClaimLayer2Information = num_of_func_pluginClaimLayer2Information;
	t_array->num_of_func_pluginClaimLayer3Information = num_of_func_pluginClaimLayer3Information;
	t_array->num_of_func_pluginClaimLayer4Information = num_of_func_pluginClaimLayer4Information;
	t_array->num_of_func_pluginOnFlowTerminate = num_of_func_pluginOnFlowTerminate;
	t_array->num_of_func_pluginOnApplicationTerminate = num_of_func_pluginOnApplicationTerminate;
	t_array->num_of_func_bufferToSink = num_of_func_pluginBufferToSink;

	// malloc function pointer arrays
	if (num_of_func_pluginInitialize > 0)
		t_array->pluginInitialize = malloc(sizeof(initialize_func) * num_of_func_pluginInitialize);
	if (num_of_func_pluginPrintHeader > 0)
		t_array->pluginPrintHeader = malloc(sizeof(printHeader_func) * num_of_func_pluginPrintHeader);
	if (num_of_func_pluginOnFlowGenerated > 0)
		t_array->pluginOnFlowGenerated = malloc(sizeof(onFlowGenerated_func) * num_of_func_pluginOnFlowGenerated);
	if (num_of_func_pluginClaimLayer2Information > 0)
		t_array->pluginClaimLayer2Information = malloc(sizeof(claimLayer2Information_func) * num_of_func_pluginClaimLayer2Information);
	if (num_of_func_pluginClaimLayer3Information > 0)
		t_array->pluginClaimLayer3Information = malloc(sizeof(claimLayer3Information_func) * num_of_func_pluginClaimLayer3Information);
	if (num_of_func_pluginClaimLayer4Information > 0)
		t_array->pluginClaimLayer4Information = malloc(sizeof(claimLayer4Information_func) * num_of_func_pluginClaimLayer4Information);
	if (num_of_func_pluginOnFlowTerminate > 0)
		t_array->pluginOnFlowTerminate = malloc(sizeof(onFlowTerminate_func) * num_of_func_pluginOnFlowTerminate);
	if (num_of_func_pluginOnApplicationTerminate > 0)
		t_array->pluginOnApplicationTerminate = malloc(sizeof(onApplicationTerminate_func) * num_of_func_pluginOnApplicationTerminate);
	if (num_of_func_pluginBufferToSink > 0)
		t_array->pluginBufferToSink = malloc(sizeof(bufferToSink_func) * num_of_func_pluginBufferToSink);

	// fill pointer arrays
	// reset counters
	num_of_func_pluginInitialize = 0;
	num_of_func_pluginPrintHeader = 0;
	num_of_func_pluginOnFlowGenerated = 0;
	num_of_func_pluginClaimLayer2Information = 0;
	num_of_func_pluginClaimLayer3Information = 0;
	num_of_func_pluginClaimLayer4Information = 0;
	num_of_func_pluginOnFlowTerminate = 0;
	num_of_func_pluginOnApplicationTerminate = 0;
	num_of_func_pluginBufferToSink = 0;

	for (i = 0; i < num_of_plugins; i++) {
		if (t_array->plugin[i].initialize != NULL) t_array->pluginInitialize[num_of_func_pluginInitialize++] = t_array->plugin[i].initialize;
		if (t_array->plugin[i].printHeader != NULL) t_array->pluginPrintHeader[num_of_func_pluginPrintHeader++] = t_array->plugin[i].printHeader;
		if (t_array->plugin[i].onFlowGenerated != NULL) t_array->pluginOnFlowGenerated[num_of_func_pluginOnFlowGenerated++] = t_array->plugin[i].onFlowGenerated;
		if (t_array->plugin[i].claimLayer2Information != NULL) t_array->pluginClaimLayer2Information[num_of_func_pluginClaimLayer2Information++] = t_array->plugin[i].claimLayer2Information;
		if (t_array->plugin[i].claimLayer3Information != NULL) t_array->pluginClaimLayer3Information[num_of_func_pluginClaimLayer3Information++] = t_array->plugin[i].claimLayer3Information;
		if (t_array->plugin[i].claimLayer4Information != NULL) t_array->pluginClaimLayer4Information[num_of_func_pluginClaimLayer4Information++] = t_array->plugin[i].claimLayer4Information;
		if (t_array->plugin[i].onFlowTerminate != NULL) t_array->pluginOnFlowTerminate[num_of_func_pluginOnFlowTerminate++] = t_array->plugin[i].onFlowTerminate;
		if (t_array->plugin[i].onApplicationTerminate != NULL) t_array->pluginOnApplicationTerminate[num_of_func_pluginOnApplicationTerminate++] = t_array->plugin[i].onApplicationTerminate;
		if (t_array->plugin[i].bufferToSink != NULL) t_array->pluginBufferToSink[num_of_func_pluginBufferToSink++] = t_array->plugin[i].bufferToSink;
	}
	closedir(directory);
	tempHandle = NULL;
	return t_array;
}

int unload_tranalyzer_plugins(tranalyzer_plugin_array_t *plugin_array) {
	//int i;

	// free the function arrays
	/*for (i = 0; i < plugin_array->num_of_func_pluginInitialize; i++) {
		free(plugin_array->pluginInitialize[i]);
	}*/
	free(plugin_array->pluginInitialize);

	/*for (i = 0; i < plugin_array->num_of_func_pluginPrintHeader; i++) {
		free(plugin_array->pluginPrintHeader[i]);
	}*/
	free(plugin_array->pluginPrintHeader);

	/*for (i = 0; i < plugin_array->num_of_func_pluginOnFlowGenerated; i++) {
		free(plugin_array->pluginOnFlowGenerated[i]);
	}*/
	free(plugin_array->pluginOnFlowGenerated);

	/*for (i = 0; i < plugin_array->num_of_func_pluginClaimLayer2Information; i++) {
		free(plugin_array->pluginClaimLayer2Information[i]);
	}*/
	free(plugin_array->pluginClaimLayer2Information);

	/*for (i = 0; i < plugin_array->num_of_func_pluginClaimLayer3Information; i++) {
		free(plugin_array->pluginClaimLayer3Information[i]);
	}*/
	free(plugin_array->pluginClaimLayer3Information);

	/*for (i = 0; i < plugin_array->num_of_func_pluginClaimLayer4Information; i++) {
		free(plugin_array->pluginClaimLayer4Information[i]);
	}*/
	free(plugin_array->pluginClaimLayer4Information);

	/*for (i = 0; i < plugin_array->num_of_func_pluginOnFlowTerminate; i++) {
		free(plugin_array->pluginOnFlowTerminate[i]);
	}*/
	free(plugin_array->pluginOnFlowTerminate);

	/*for (i = 0; i < plugin_array->num_of_func_pluginOnApplicationTerminate; i++) {
		free(plugin_array->pluginOnApplicationTerminate[i]);
	}*/
	free(plugin_array->pluginOnApplicationTerminate);

	free(plugin_array->pluginBufferToSink);

	unsigned int ltp_i;
	for (ltp_i = 0; ltp_i < plugin_array->num_of_plugins; ltp_i++) {
		dlclose(plugin_array->plugin[ltp_i].handle);
	}

	// free the plugins
	free(plugin_array->plugin);


	// now free the structure itself
	free(plugin_array);

	return 0; // unload was successful
}
