/*
 * loadTranalyzerPlugins.h
 */

#ifndef LOADTRANALYZERPLUGINS_H_
#define LOADTRANALYZERPLUGINS_H_

#include <stdlib.h>
#include <stdio.h>
#include "networkHeaders.h"
#include "outputBuffer.h"
#include "binaryValue.h"

#define TRANALYZER_PLUGIN_CURRENT_VERSION_MAJOR 0 // defines the current major version of the plugin architecture
#define TRANALYZER_PLUGIN_CURRENT_VERSION_MINOR 5 // defines the current minor version of the plugin architecture

#define TRANALYZER_PLUGIN_SUPPORTED_VERSION_MAJOR 0 // defines the lowest major version of tranalyzer whose plugin system definition is supported
#define TRANALYZER_PLUGIN_SUPPORTED_VERSION_MINOR 5 // defines the lowest minor version of tranalyzer whose plugin system definition is supported

#define TRANALYZER_PLUGIN_NAME_MAX_LENGTH 256 // the maximum length of the plugin name
#define TRANALYZER_PLUGIN_VERSION_MAX_LENGTH 256 // the maximum length of the plugin version

#define TRANALYZER_PLUGIN_FUNCTION_GET_PLUGIN_NAME "get_plugin_name" // the name of the plugin function that returns the plugin name
#define TRANALYZER_PLUGIN_FUNCTION_GET_PLUGIN_VERSION "get_plugin_version" // the name of the plugin function that returns the plugin version (as string)
#define TRANALYZER_PLUGIN_FUNCTION_GET_VERSION_MAJOR "get_supported_tranalyzer_version_major" // the name of the plugin function that returns the major version of tranalyzer whose plugin system definition is supported
#define TRANALYZER_PLUGIN_FUNCTION_GET_VERSION_MINOR "get_supported_tranalyzer_version_minor" // the name of the plugin function that returns the major version of tranalyzer whose plugin system definition is supported
#define TRANALYZER_PLUGIN_FUNCTION_GET_DEPENDENCIES "get_dependencies" // the name of the plugin function that returns the name of the plugins to which the actual plugin has dependencies and therefore have to be present
#define TRANALYZER_PLUGIN_FUNCTION_INITIALIZE "initialize" // the name of the plugin function that initializes the plugin
#define TRANALYZER_PLUGIN_FUNCTION_PRINT_HEADER "printHeader" // the name of the plugin function that prints information for the header
#define TRANALYZER_PLUGIN_FUNCTION_ON_FLOW_GENERATED "onFlowGenerated"
#define TRANALYZER_PLUGIN_FUNCTION_CLAIM_LAYER_2_INFORMATION "claimLayer2Information"
#define TRANALYZER_PLUGIN_FUNCTION_CLAIM_LAYER_3_INFORMATION "claimLayer3Information"
#define TRANALYZER_PLUGIN_FUNCTION_CLAIM_LAYER_4_INFORMATION "claimLayer4Information"
#define TRANALYZER_PLUGIN_FUNCTION_ON_FLOW_TERMINATE "onFlowTerminate"
#define TRANALYZER_PLUGIN_FUNCTION_ON_APPLICATION_TERMINATE "onApplicationTerminate"
#define TRANALYZER_PLUGIN_FUNCTION_BUFFER_TO_SINK "bufferToSink"

typedef struct dirent_list_entry_s {
	struct dirent *de;
	struct dirent_list_entry_s *next;
} dirent_list_entry_t;

/*
// dummy struct, only needed for testing
typedef struct packet_s {
	int value;
} packet_t;
*/

typedef struct tranalyzer_plugin {
	char name[TRANALYZER_PLUGIN_NAME_MAX_LENGTH];
	char version[TRANALYZER_PLUGIN_VERSION_MAX_LENGTH];
	unsigned int supported_tranalyzer_version_major;
	unsigned int supported_tranalyzer_version_minor;
	void *handle;

	void (*initialize)(unsigned long pPluginIndex); // the function to initialize the plugin
	binary_value_t* (*printHeader)();
	void (*onFlowGenerated)(packet_t *packet, unsigned long flowIndex);
	void (*claimLayer2Information)(packet_t *packet);
	void (*claimLayer3Information)(packet_t *packet);
	void (*claimLayer4Information)(packet_t *packet, unsigned long flowIndex);
	void (*onFlowTerminate)(unsigned long flowIndex);
	void (*onApplicationTerminate)();
	void (*bufferToSink)(outputBuffer_t* buffer);
} tranalyzer_plugin_t;

typedef struct tranalyzer_plugin_list_element {
	tranalyzer_plugin_t plugin;
	struct tranalyzer_plugin_list_element *next;
} tranalyzer_plugin_list_element_t;

typedef struct tranalyzer_plugin_array {
	unsigned int num_of_plugins;
	unsigned int num_of_func_pluginInitialize;
	unsigned int num_of_func_pluginPrintHeader;
	unsigned int num_of_func_pluginOnFlowGenerated;
	unsigned int num_of_func_pluginClaimLayer2Information;
	unsigned int num_of_func_pluginClaimLayer3Information;
	unsigned int num_of_func_pluginClaimLayer4Information;
	unsigned int num_of_func_pluginOnFlowTerminate;
	unsigned int num_of_func_pluginOnApplicationTerminate;
	unsigned int num_of_func_bufferToSink;

	tranalyzer_plugin_t *plugin;

	void (**pluginInitialize)(unsigned long pPluginIndex);
	binary_value_t* (**pluginPrintHeader)();
	void (**pluginOnFlowGenerated)(packet_t *packet, unsigned long flowIndex);
	void (**pluginClaimLayer2Information)(packet_t *packet);
	void (**pluginClaimLayer3Information)(packet_t *packet);
	void (**pluginClaimLayer4Information)(packet_t *packet, unsigned long flowIndex);
	void (**pluginOnFlowTerminate)(unsigned long flowIndex);
	void (**pluginOnApplicationTerminate)();
	void (**pluginBufferToSink)(outputBuffer_t *buffer);
} tranalyzer_plugin_array_t;


/*
 * Loads the dynamic librarys from a given folder and returns a
 * tranalyzer_plugin_array struct or NULL if no appropriate dynamic library was
 * found.
 */
tranalyzer_plugin_array_t* load_tranalyzer_plugins(char* folder);


/*
 * unloads all plugins in a tranalyzer_plugin_array struct.
 */
int unload_tranalyzer_plugins(tranalyzer_plugin_array_t *plugin_array);

#endif /* LOADTRANALYZERPLUGINS_H_ */
