/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __MAIN_H__
#define __MAIN_H__

// includes
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <pcap.h>
#include <time.h>
#include <sys/stat.h>
#include <limits.h>
#include <math.h>
#include <sys/ioctl.h>
#include <libgen.h>
#include <inttypes.h>
#include <float.h>
#ifdef __APPLE__
#include <netinet/if_ether.h>
#include <net/if_dl.h>
#include <ifaddrs.h>
#else
#include <netinet/ether.h>
#endif //__APPLE__
#include <net/if.h>

// local includes
#include "tranalyzer.h"
#include "networkHeaders.h"
#include "packetCapture.h"
#include "hashTable.h"
#include "tranalyzer_utils.h"
#include "loadTranalyzerPlugins.h"
#include "ioBuffer.h"
#include "redBlackTree.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "getter.h"
// global defines
// globalWarn, packet and flow status variables

#define	L3FLOWINVERT		0x00000001	// flowstatus: Invert Flow, not client flow

#define	L2SNAPLENGTH		0x00000001	// globalWarn: acquired packet length < minimal L2 datagram
#define	L3SNAPLENGTH		0x00000002	// acquired packet length < packet length in L3 header
#define	L3HDRSHRTLEN		0x00000004	// acquired packet length < minimal L3 Header
#define	L4HDRSHRTLEN		0x00000008	// acquired packet length < minimal L4 Header
#define	IPV4_FRAG		0x00000010	// IPv4 Fragmentation present
#define IPV4_FRAG_ERR		0x00000020	// IPv4 Fragmentation Error (detailed err s. tcpFlags plugin)
#define IPV4_FRAG_HDSEQ_ERR	0x00000040	// IPv4 1. fragment out of sequence or missing
#define IPV4_FRAG_PENDING	0x00000080	// fragmentation sequence not completed when flow timeout

#define	L2_PPPoE_D		0x00000100	//
#define L2_PPPoE_S		0x00000200	//
#define L2_LLDP			0x00000400	//
#define L2_ARP			0x00000800	// ARP present
#define L2_RARP			0x00001000	// reverse ARP present
#define L2_VLAN			0x00002000	// Vlans present
#define L2_MPLS_UCAST		0x00004000	// MPLS unicast present
#define L2_MPLS_MCAST		0x00008000	// MPLS multicast present
#define L2_L2TP			0x00010000	// L2TP v2 present
#define L2_GRE			0x00020000	// GRE V1/2 present
#define L2_PPP			0x00040000	// PPP header present after L2TP or GRE
#define L2_IPV6			0x00080000	// IPv6 packets present

#define RMFLOW			0x00800000	// Netflow or Packet Alarm Mode flag: remove this flow from chain instantly
#define PPP_NRHD		0x10000000	// PPPL3 header not readable, compressed
#define	LANDATTACK		0x40000000	// same src IP && dst IP and src port && dst port
#define	TIMEJUMP		0x80000000	// time slip in the pcap file mostly due to NTP operations on the capture machine

//#define ALARM			0x80000000

#define	SNAPLENGTH	(L2SNAPLENGTH | L3SNAPLENGTH)
#define L2_MPLS		(L2_MPLS_UCAST | L2_MPLS_MCAST)

#define MAX_TM_BUF	35

// variables

// pcap
pcap_t* captureDescriptor; // pcap handler
struct bpf_program bpfProgram; // if bpf filters set, the compiled program will be stored here

// general use
char* tranalyzer_logo;
char* ants_ascii;
volatile int running;
unsigned long srvFlowIndex, clntFlowIndex; // the 'server' (initiator) and 'client' (receiver) flow index
char errbuf[PCAP_ERRBUF_SIZE]; // error buffer for output uses
char headerFileName[MAX_FILENAME_LEN+1]; // filename of the header file
long readPackets; // indicates how many packets we have dispatched with pcap_dispatch
flow_t *lruPointer, *lruPrevFlow, *lruNextFlow, *srvFlow, *clntFlow; // convenient pointers so that we don't have to push them on the stack during runtime
struct timeval startTime, endTime, duration, currTime; // currTime only used for checking whether the lru list has only one item (special case)
struct timeval actTime, startTStamp;
float timeDiff; // the difference between the time of the last packet or actual time and the time of the last packet of the actual inspected flow in the lru list
unsigned long long captureFileSize; // If we're capturing traffic from a savefile, this is the filesize
unsigned long long bytesProcessed, bytesonWire, rawbytesonWire; // If we're capturing traffic from a file, these are bytes already processed
uint32_t totalfIndex; 
struct pcap_stat *pcap_statistics; // pcap statistics for received and dropped packets
struct stat fileStats;
hashMap_t *mainHashMap, *fragPendMap;

unsigned long numOfPackets, numOfV4Packets, numOfV6Packets, numOfFragV4Packets, numOfFragV6Packets, numOfARPPackets, numOfRARPPackets;
unsigned long maxNumFlows, maxNumFlowsPeak;
unsigned long numAlarms;

#if FORCE_MODE == 1
unsigned long num_rm_flows;
flow_t *rm_flows[HASHCHAINTABLE_BASE_SIZE];
#endif // FORCE_MODE == 1

tranalyzer_plugin_array_t *tranalyzer_plugins;
FILE *logFile; // a file for the logging information before exiting
char logFileName[MAX_FILENAME_LEN+1]; // filename of the log file
uint32_t globalWarn; // packetCapture snaplength and timing warning
unsigned char startTS, supOut; // acquire timestamp of first packet
char timeBuf[MAX_TM_BUF]; // char time buffer for tm conversion

outputBuffer_t *main_output_buffer, *header_output_buffer;
binary_value_t *main_header_bv;
binary_header_t *main_header;

/*
 * A variable to store the HASHCHAINTABLE_SIZE
 *
 * Why do we need this:
 * A plugin needs to determine AT RUNTIME, how big the HashTable is.
 * If a plugin uses the define, it is static and therefore not portable.
 * Thus a change in the size of the HashTable invokes a COMPLETE re-compilation
 * of ALL plugins, including proprietary ones -> fail.
 * volatile prevents from optimizing out.
 */
volatile unsigned long size_of_rb_tree_nodepool; // size_of_main_hashtable

typedef struct outputString_s {
	unsigned short stringSize;
	char outputString[4096]; // a String to store all the textual output of the plugins
} outputString_t;

outputString_t *pluginOutputString;
unsigned short pluginOutputStringPos; // a counter to determine the writing pos in the output string
unsigned char exitPriority; // determines how importantly we want to terminate the application. Every CTRL+C keystroke increments the counter, starting at zero

typedef struct timeout_handler_s {
	float timeout; // the timeout value in seconds
	flow_t flow; // a sentinel flow
	struct timeout_handler_s *next;
} timeout_handler_t;


typedef struct timeout_manager_s {
	timeout_handler_t *timeout_handler_list;
} timeout_manager_t;

timeout_manager_t main_timeout_manager;
timeout_handler_t *main_timeout_handler;

typedef struct file_list_element_s {
	char *file_name;
	struct file_list_element_s *next;
	unsigned long long file_size;
} file_list_element_t;

typedef struct file_list_s {
	unsigned long number_of_files;
	unsigned long long size_of_files;
	file_list_element_t *file_list;
} file_list_t;

file_list_t *capture_file_list;
file_list_element_t *act_capture_list_element;
unsigned long act_capture_file_index;

// parsing parameters
char *captureInterface; // the interface from which the packets get sniffed
char *captureFileName; // if not read from an interface, read from a stored dump file
char *captureListFileName; // the name of the file with a list of
char *pluginFolder; // if the p option is used, the name of plugin folder is stored in here
char *baseFileName; // the base file name for all generated files
char *bpfCommand; // bpf filter command
char *esomFileName; // 
char specialOutput;

// functions
int main(int argc, char *argv[]);
void removeFlow(flow_t *flow);
void lruFlowPrintout();
void printUsage();
void terminate();
void mainLoop();
void catchInterrupt(int signal);
char *copy_argv(char **argv);
void prepareSniffing();
void printFlow(unsigned long serverFlowIndex, unsigned long clientFlowIndex);
void userInterrupt(int signal);
void buildHeaders();
void cycleLRULists();
void cycleLRUList(timeout_handler_t *th);
int timeout_handler_add(float timeout_in_sec);
void write_log_file();
void print_IP_interfaces(FILE * file);
file_list_t* get_capture_file_list(FILE *list_file);

#endif

