/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "packetCapture.h"
#include "main.h"

#if ENABLE_IO_BUFFERING == 0
// callback function triggered every time we receive/read a new packet from the pcap descriptor
void perPacketCallback(u_char* inqueue_uchar, const struct pcap_pkthdr *pcapHeader, const u_char* packet) {
	actTime.tv_sec = pcapHeader->ts.tv_sec;
	actTime.tv_usec = pcapHeader->ts.tv_usec;
	cycleLRULists();
	processPacket(inqueue_uchar, pcapHeader, packet);
}
#endif // ENABLE_IO_BUFFERING == 0

// the function that starts the processing of a packet
inline void processPacket(u_char* inqueue_uchar, const struct pcap_pkthdr *pcapHeader, const u_char* packet) {
	uint16_t i;
	_8021Q_t *shape;

	numOfPackets++; // increment packet counter
	memset(&newPacket, '\0', sizeof(packet_t)); // reset packet structure
 
	newPacket.raw_packet = packet;

	newPacket.pcapHeader = pcapHeader; // basically to store the packet arrival time in milliseconds

	newPacket.snapL2Length = pcapHeader->caplen;
	newPacket.packetRawLength = pcapHeader->len;
	
	bytesProcessed += newPacket.snapL2Length;
	rawbytesonWire += newPacket.packetRawLength;
	
	newPacket.layer2Header = (l2Header_t*)packet;

	if (newPacket.snapL2Length < sizeof(ethernetHeader_t)) {
//		newPacket.status = L2SNAPLENGTH;
		if (!(globalWarn & L2SNAPLENGTH)) { // Snap length warning
			globalWarn |= L2SNAPLENGTH;
			printf("Warning: Ethernet header: %lu > Snaplength: %d, Packets ignored\n", sizeof(ethernetHeader_t), newPacket.snapL2Length);
		}
		return; 
	}

#if NOLAYER2 == 0 // real traffic
	newPacket.layer2Type = L2_ETHERNET;


	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginClaimLayer2Information; i++) {
		tranalyzer_plugins->pluginClaimLayer2Information[i](&newPacket);
	}

	// check embedded Protocols 
	shape = (_8021Q_t*) (packet + 12); // advance 12 bytes to ether type
	if (shape->identifier == ETHERTYPE_VLANn) { // check for 802.1ad signature
#if (AGGREGATIONFLAG & VLANID) == 0
		newPacket.vlans = (uint32_t*)shape; 
#endif
		newPacket.status |= L2_VLAN;
		globalWarn |= L2_VLAN;
		while (shape->identifier == ETHERTYPE_VLANn) shape++; // scroll all vlan headers
		
#if (AGGREGATIONFLAG & VLANID) == 0
		innerVLANID = ntohs((shape-1)->vlanID) & VLANID_MASK16;
	} else
#else
	}
#endif
		innerVLANID = 0;

	//newPacket.layer3Type = ntohs(*(uint16_t*)shape);
	switch (*(uint16_t*)shape) {
		case ETHERTYPE_IPn: // IP v4 packet
			newPacket.layer3Header = (const l3Header_t*) ((u_char*)shape + 2); // adjust offset
#if IPV6_ACTIVATE == 0
			dissembleIPv4Packet(&newPacket);
#else // IPV6_ACTIVATE == 1
			newPacket.layer3Type = ETHERTYPE_IP;
			numOfV4Packets++;
#endif // IPV6_ACTIVATE
			break;
		case ETHERTYPE_IPV6n: // IPv6 packet
			newPacket.layer3Header = (const l3Header_t*) ((u_char*)shape + 2); // adjust offset
#if IPV6_ACTIVATE == 1
			dissembleIPv6Packet(&newPacket);
#else // IPV6_ACTIVATE == 0
			newPacket.layer3Type = ETHERTYPE_IPV6;
			newPacket.status |= L2_IPV6;
			numOfV6Packets++;
#endif // IPV6_ACTIVATE
			break;
		case ETHERTYPE_ARPn:
		 	newPacket.status |= L2_ARP;
			newPacket.layer3Type = ETHERTYPE_ARP;
			numOfARPPackets++;
			break;
		case ETHERTYPE_RARPn:
		 	newPacket.status |= L2_RARP;
	//		newPacket.layer3Type = ETHERTYPE_RARP;
			numOfRARPPackets++;
			break;
		case ETHERTYPE_MPLS_MULTICASTn: // 0x8848, MPLS Multicast
			newPacket.status |= L2_MPLS_MCAST;
			goto mpls_ucast;
		case ETHERTYPE_MPLS_UNICASTn: // 0x8847, MPLS Unicast
			newPacket.status |= L2_MPLS_UCAST;
	mpls_ucast:	globalWarn |= L2_MPLS;
			shape = (_8021Q_t*)&shape->vlanID;
			newPacket.mpls = (uint32_t*)shape;
			while ( !(shape->vlanID & BTM_MPLS_STKn16) ) shape++; // test MPLS end of stack bit
			shape++; // advance 4 bytes to IP Header
			newPacket.layer3Header = (const l3Header_t*) (u_char*)shape;	
				
			if ( (shape->identifier & 0x0040) == 0x0040) { // IPv4 packet
#if IPV6_ACTIVATE == 0
				dissembleIPv4Packet(&newPacket);
#else // IPV6_ACTIVATE == 1
				newPacket.layer3Type = ETHERTYPE_IP;
				numOfV4Packets++;
#endif // IPV6_ACTIVATE
			} else if ( (shape->identifier & 0x0060) == 0x0060 ) { // IPv6 packet
#if IPV6_ACTIVATE == 1
				dissembleIPv6Packet(&newPacket);
#else // IPV6_ACTIVATE == 0
				newPacket.layer3Type = ETHERTYPE_IPV6;
				newPacket.status |= L2_IPV6;
				numOfV6Packets++;
#endif // IPV6_ACTIVATE
			}
			break;
		case ETHERTYPE_PPPoE_Dn:
		 	newPacket.status |= L2_PPPoE_D;
			break;
		case ETHERTYPE_PPPoE_Sn:
		 	newPacket.status |= L2_PPPoE_S;
			break;
		case ETHERTYPE_LLDP_n:
		 	newPacket.status |= L2_LLDP;
			break;
		default:
			break;
	} // Switch end

#else // NOLAYER2 == 1 // artificial crap

	newPacket.layer2Type = 0;
	innerVLANID = 0;
	newPacket.layer3Header = (const l3Header_t*) (((u_char*)packet) + NOL2_L3HDROFFSET);  
	newPacket.layer3Type = ETHERTYPE_IP;
	dissembleIPv4Packet(&newPacket);

#endif // NOLAYER2

}


#if IPV6_ACTIVATE == 0

inline void dissembleIPv4Packet(packet_t *packet) {
	int32_t packetLen;
	uint16_t l2Len; // uint32_t
	uint16_t i, l3Len;
#if FRAGMENTATION == 1
#if FRAG_HLST_CRFT == 1
	uint8_t sw_fnohead = 0;
#endif // FRAG_HLST_CRFT
#if FRAG_ERROR_DUMP == 1
	char srcIP[MAX_IP_PRINT_LEN];
        char dstIP[MAX_IP_PRINT_LEN];
#endif // FRAG_ERROR_DUMP
#endif // FRAGMENTATION
	flow_t *flow;
	unsigned long flowIndex;

	packet->layer3Type = ETHERTYPE_IP;
//	packet->status |= L2_IPV4;
	numOfV4Packets++;

	ipHeader = (ipHeader_t*) packet->layer3Header;

	packet->l3HdrLen = IP_HL(ipHeader) << 2;
// TODO: L3 encapsulation IP in IP or ipv6 in IP

#if L2TP == 1
	uint16_t *l2TPPP, *l2TPH;
	if (ipHeader->ip_p == L3_UDP ) { // udp
		packet->layer4Header = (const l4Header_t*) ((u_char*)ipHeader + (IP_HL(ipHeader)<<2)); // adjust header to the beginning of the encapsulated protocol
		l2TPPP = (uint16_t*) packet->layer4Header;
		if ( *l2TPPP == L2TP_PORT_N && *(l2TPPP+1) == L2TP_PORT_N) { // L2TP port 1701, 0xa506 hex network order
			packet->l2tpLayer3Hdr = packet->layer3Header;
			globalWarn |= L2_L2TP;
			packet->status |= L2_L2TP;
			l2TPPP += 4; // advance to L2TP 
			l2TPH = l2TPPP;
			packet->l2TPHdr = (l2tpv2Header_t*)l2TPPP;
			if ( (*l2TPH & (L2TP_TYP | L2TP_VER)) == L2TP_V2 ) { // if data then advance to l3 header, later version: supply l2tp, ppp parameters for plugins
				l2TPPP++;
				if ( *l2TPH & L2TP_LEN ) l2TPPP++;
				l2TPPP += 2; // tunnel / session ID
				if ( *l2TPH & L2TP_SQN ) l2TPPP += 2;
				if ( *l2TPH & L2TP_OFF ) l2TPPP += (ntohs(*l2TPPP)>>1) + 1;

//				if ( *l2TPPP ==  PPP_ADD_CTL ) { // HDLC ppp present 0xff03	
					packet->pppHdr = (pppHu_t*)l2TPPP; // save ppp header
					l2TPPP++;	 // advance HDLC ppp header add field 0xff03, following HDLC ppp encapsulated prot code
   					if (*l2TPPP == PPP_PROT_IP) { // PPP IP encapsulation?
						packet->status |= L2_PPP;
					//	packet->layer2Type = L2TP_V2;
						packet->layer3Header = (const l3Header_t*) (u_char*)(++l2TPPP);
						packet->layer3Type = ETHERTYPE_IP;
						ipHeader = (ipHeader_t*) packet->layer3Header;
					} else return; // only IP
/*				} else { // TODO: check for IPC etc 
					return;  
				} */
			} else return; // only data
		} // L2TP ?
	} // UDP ?
#endif // L2TP

#if GRE == 1
	uint32_t *grePPP, *greHD;
	if (ipHeader->ip_p == L3_GRE) {
		packet->layer4Header = (const l4Header_t*) ((u_char*)ipHeader + (IP_HL(ipHeader)<<2)); // adjust header to the beginning of the encapsulated protocol
		grePPP = (uint32_t*) packet->layer4Header;
		globalWarn |= L2_GRE;
		packet->status |= L2_GRE;
		greHD = grePPP++;
		packet->greHdr = (greHeader_t*)greHD;
		packet->greLayer3Hdr = packet->layer3Header;
		if (*greHD & GRE_CKSMn) grePPP++;
		if (*greHD & GRE_RTn) grePPP++;
		if (*greHD & GRE_KEYn) grePPP++;
		if (*greHD & GRE_SQn) grePPP++;
		if (*greHD & GRE_SSRn) grePPP++;
		if (*greHD & GRE_ACKn) grePPP++;
		if ((*greHD & GRE_PROTOn) == GRE_PPPn) {
			packet->status |= L2_PPP;
			if ( (*grePPP  & 0x000000ff) == GRE_PPP_CMPRSS ) {
				packet->status |= PPP_NRHD; // compressed, no readable header; info for later processing of flow
			}
			if ( ((greHeader_t*)greHD)->plength ) {
				
			} else {
					
			}
		} else { 
			packet->layer3Header = (const l3Header_t*) (u_char*)(grePPP);
			packet->layer3Type = (uint16_t)(*greHD & GRE_PROTOn);
			ipHeader = (ipHeader_t*)packet->layer3Header;
		}
	}
#endif // GRE == 1

#if MULTIPKTSUP == 1
	if (flow) {
		if (ipHeader->ip_id == flow->lastIPID) return;
		else flow->lastIPID = ipHeader->ip_id;
	}
#endif //
	
#if FRAGMENTATION == 0 
	// do not handle fragmented packets

	if (ipHeader->ip_off & FRAGID_N) {
		globalWarn |= IPV4_FRAG;
		numOfFragV4Packets++;
		return; // fragmentation switch off: ignore fragmented packets except the 1. protocol header 
	}
#endif //FRAGMENTATION	

	i = (uint16_t)((uint8_t*) packet->layer3Header - (uint8_t*) packet->layer2Header); // L2,VLAN length
	packet->snapL3Length = packet->snapL2Length - i; // L3 Packet length
	packet->l2HdrLen = i;
	l3Len = ntohs(ipHeader->ip_len);  // get IP packet length from IP header
	l2Len = l3Len + i;
 	packet->packetL2Length = l2Len; 
	bytesonWire += l2Len;   // estimate all Ethernet & IP bytes seen on wire
	
	if ( packet->snapL3Length < l3Len ) { // Layer3 snaplength too short or IP packet too short?
		packet->status |= L3SNAPLENGTH;
 		if (!(globalWarn & L3SNAPLENGTH)) { // Snap length warning
			globalWarn |= L3SNAPLENGTH;
			printf("Warning: snapL2Length: %d - snapL3Length: %d - IpLength in Header: %d\n", packet->snapL2Length, packet->snapL3Length, l3Len);
		} 
	} else if ( packet->snapL3Length > l3Len ) {
		packet->snapL2Length = l2Len; 
		packet->snapL3Length = l3Len;
	}

	if ( l3Len < 20) { // Layer3 snaplength too short or IP packet too short?
		packet->status |= L3HDRSHRTLEN;	
	}

#if PACKETLENGTH == 0 
	packetLen = l2Len;
#else
	packetLen = l3Len;
#endif

	// -------------------- layer3 ------------------------

#if  PACKETLENGTH <= 1 
#if ( FRGIPPKTLENVIEW == 1 && FRAGMENTATION == 1 ) // IP packet view mode in case of fragmentation

	if (ipHeader->ip_off & FRAGID_N) packetLen -= packet->l3HdrLen; // remove IP header len only from 2nd++ frag if whole packet statistical view is required: default
#endif // / ( FRGPKTLENVIEW == 1 && FRAGMENTATION == 1 )

	if (packetLen >= 0) packet->packetLength = packetLen; 
	else { 	
		packet->packetLength = 0;
		packet->status |= L3HDRSHRTLEN;	
	}
#endif // PACKETLENGTH <= 1 

	packet->layer4Type = ipHeader->ip_p; // set layer4Type already for global plugins such as protocolStatistics
	
	// if desired calculate statistics with the IP protocol. Note that the data of the encapsulated protocol is still unknown.
	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginClaimLayer3Information; i++) {
		tranalyzer_plugins->pluginClaimLayer3Information[i](packet);
	}

	// -------------------- layer 4 ------------------------


	uint16_t l4HdrOff = 8;
	packet->layer4Header = (const l4Header_t*) ((u_char*)ipHeader + packet->l3HdrLen); // adjust header to the beginning of the encapsulated protocol
#if FRAGMENTATION == 1
	if ( (ipHeader->ip_off & FRAGID_N) == FRAGID_1P_N ) { 
#endif
	    switch (ipHeader->ip_p) {
		case 1:
			l4HdrOff = 4;
			break; 
		case 6: 
			l4HdrOff = packet->layer4Header->tcpHeader.doff << 2; 
			break;
		case 47:
			if ( *(uint32_t*)packet->layer4Header & 0x000080f0 ) l4HdrOff = 8;			
			break;
		case 89:
			l4HdrOff = 16;
			break;
		case 132:
			l4HdrOff = 12;
			break;
	    }
#if FRAGMENTATION == 1
	}  else l4HdrOff = 0;
#endif // FRAGMENTATION == 1

 	packet->l4HdrLen = l4HdrOff;

#if PACKETLENGTH >= 2
	packetLen -= IP_HL(ipHeader) << 2;  
#if PACKETLENGTH == 3 // sustract L4 header
	packetLen -= l4HdrOff;	
#endif // PACKETLENGTH == 3
	if (packetLen >= 0) packet->packetLength = packetLen;
	else { 
		packet->packetLength = 0;
		packet->status |= L4HDRSHRTLEN;	
	}
#endif // PACKETLENGTH >= 2


	// -------------------- layer 7 info for packet structure ------------------------
	
	packet->layer7Header = (uint8_t*)packet->layer4Header + l4HdrOff;
	if ( packet->snapL3Length >= l3Len) { 					// L3 length not snapped
		packet->snapL4Length = l3Len - packet->l3HdrLen;
                packet->snapL7Length = packet->snapL4Length - l4HdrOff; 	// Protocol L3/4 Hdr lengths are valid 
        } else { 									// L3 length snapped so calculate real hdr L7 length
		packet->snapL4Length = packet->snapL3Length + packet->l3HdrLen;
                packet->snapL7Length = (uint16_t)(packet->layer7Header - (uint8_t*)packet->layer3Header); // offset between L3 and L7
                packet->snapL7Length = packet->snapL3Length - packet->snapL7Length; 	// real L7 length
        }

// source and destination port of a layer 4 header
	
#if AGGREGATIONFLAG & SUBNET
//	if (subnet_table) {
//		ipHeader->ip_src.s_addr = subnet_testP(subnet_table, flow->srcIP.s_addr); // subnet test src ip
//		ipHeader->ip_dst.s_addr = subnet_testP(subnet_table, flow->dstIP.s_addr); // subnet test dst ip
//	}
#else // AGGREGATIONFLAG & SUBNET == 0

#if AGGREGATIONFLAG & SRCIP
	ipHeader->ip_src.s_addr &= SRCIPMASKn;
#endif // AGGREGATIONFLAG & SRCIP
#if AGGREGATIONFLAG & DSTIP
	ipHeader->ip_dst.s_addr &= DSTIPMASKn;
#endif // AGGREGATIONFLAG & DSTIP

#endif // AGGREGATIONFLAG & SUBNET

	hashHelper.vlanID = innerVLANID;
	hashHelper.srcIP = ipHeader->ip_src;
	hashHelper.dstIP = ipHeader->ip_dst;
	
#if FRAGMENTATION == 1

	unsigned long fragPendIndex;	
	
	if (ipHeader->ip_off & FRAGID_N) { // if 2nd++ fragmented packet 
	
//		numOfFragV4Packets++;
		hashHelper.srcPort = ipHeader->ip_id;
		fragPendIndex = hashTable_lookup(fragPendMap, (char*)&hashHelper.srcIP);
	
		if (fragPendIndex == HASHTABLE_ENTRY_NOT_FOUND) { // probably missed  1. frag packet or packet mangling 
			globalWarn |= (IPV4_FRAG | IPV4_FRAG_HDSEQ_ERR);
#if FRAG_ERROR_DUMP == 1 
			inet_ntop(AF_INET, &(hashHelper.srcIP), srcIP, MAX_IP_PRINT_LEN);
                	inet_ntop(AF_INET, &(hashHelper.dstIP), dstIP, MAX_IP_PRINT_LEN);
			printf("1.frag not found @ %ld.%ld %d %s %d %s %d %d - 0x%04X 0x%04X\n", packet->pcapHeader->ts.tv_sec, packet->pcapHeader->ts.tv_usec, hashHelper.vlanID, srcIP, ntohs(packet->layer4Header->tcpHeader.source), dstIP, ntohs(packet->layer4Header->tcpHeader.dest), packet->layer4Type, ntohs(ipHeader->ip_id), ntohs(ipHeader->ip_off) );
#endif // FRAG_ERROR_DUMP

#if FRAG_HLST_CRFT == 1
			sw_fnohead = IPV4_FRAG_HDSEQ_ERR;
			goto create_packetF; // we dont know the flow, but create one anyway, because might be interesting crafted packet
#else 
			return; // we don't know the flow, so ignore packet
#endif // FRAG_HLST_LCFT
		} else { 
			numOfFragV4Packets++;
			flowIndex = fragPend[fragPendIndex];
			flow = &flows[flowIndex];
			if (!(ipHeader->ip_off & MORE_FRAG_N)) { // remove packet from frag queue when last fragment received
				if (hashTable_remove(fragPendMap, (char*) &hashHelper.srcIP) == HASHTABLE_ENTRY_NOT_FOUND) printf("- fragPend remove failed -"); 
				if (flow->status & IPV4_FRAG_PENDING) flow->status &= ~IPV4_FRAG_PENDING; 
			}	
		}

	} else { // not fragmented or 1. fragmented packet

create_packetF:
#endif //FRAGMENTATION == 1

		srcPort = 0;
		dstPort = 0;

		// encapsulated packet 
		switch (packet->layer4Type) {
			case 4: // IP in IP TODO: move pointer to next IP header and goto beginning
				break;
			case 5: // ST  
				break;
			case 6: // TCP
#if (FRAG_HLST_CRFT == 1 && FRAGMENTATION == 1 ) // 
				if ( sw_fnohead ) break;	
#endif // FRAG_HLST_CRFT == 1 && FRAGMENTATION == 1 )
				if (l3Len <  40 ) packet->status |= L4HDRSHRTLEN;
				else {		
					srcPort = ntohs(packet->layer4Header->tcpHeader.source);
					dstPort = ntohs(packet->layer4Header->tcpHeader.dest);
				}
				break;
			case 8: // EGP 
			case 9: // IGRP
			case 11: // Network Voice Protocol 
				break;
			case 17: // UDP
#if ( FRAG_HLST_CRFT == 1 && FRAGMENTATION == 1 ) // 
				if ( sw_fnohead ) break;	
#endif // FRAG_HLST_CRFT == 1 && FRAGMENTATION == 1 )
				if (l3Len <  28 ) packet->status |= L4HDRSHRTLEN;	
				else {
					srcPort = ntohs(packet->layer4Header->udpHeader.source);
					dstPort = ntohs(packet->layer4Header->udpHeader.dest);
#if L2TP == 0
					if (srcPort == 1701 && dstPort == 1701) { // L2TP ports
						packet->layer3Type = L2TP_V2;
						packet->status |= L2_L2TP;
						globalWarn |= L2_L2TP;
					}
#endif	
				}
				break;
			case 41: // IPv6 encapsulation in ipv4: TODO: move pointer and goto IPV6 processing 
				break;
			case 46: // RSVP
			case 47: // GRE
				packet->status |= L2_GRE;
				globalWarn |= L2_GRE;
			case 50:	 // ESP IPSEC TODO: 
				break;
			case 51: // AH IPSEC TODO: check encapsulated protocol
				break;
			case 89: // OSPF TODO: prepare content inspection
				break;
			case 103: // PIM 
				break;
			case 115: // L2TPV3
				packet->layer3Type = L2TP_V3;
				break;
			case 132: // SCTP, ports at the same postion as TCP
#if ( FRAG_HLST_CRFT == 1 && FRAGMENTATION == 1 ) // 
				if ( sw_fnohead ) break;	
#endif // FRAG_HLST_CRFT == 1 && FRAGMENTATION == 1 )
				if (l3Len <  40 ) packet->status |= L4HDRSHRTLEN;
				else {		
					srcPort = ntohs(packet->layer4Header->tcpHeader.source);
					dstPort = ntohs(packet->layer4Header->tcpHeader.dest);
				}
				break;
			default: // every other fucking shit port = 0
				break;
		}

#if AGGREGATIONFLAG & SRCPORT
		if (srcPort >= SRCPORTLW && srcPort <= SRCPORTHW ) srcPort = 1; else srcPort = 0;
#endif
		hashHelper.srcPort = srcPort; 
#if AGGREGATIONFLAG & DSTPORT
		if (dstPort >= DSTPORTLW && dstPort <= DSTPORTHW ) dstPort = 1; else dstPort = 0;
#endif
		hashHelper.dstPort = dstPort;
#if AGGREGATIONFLAG & L4PROT
		packet->layer4Type = 0;
#endif
		hashHelper.layer4Protocol = packet->layer4Type;

		flowIndex = hashTable_lookup(mainHashMap, (char*)&hashHelper.srcIP);
		if (flowIndex == HASHTABLE_ENTRY_NOT_FOUND) {
			if (exitPriority > 0) return;
			flowIndex = flowCreate(packet);
		}

		flow = &flows[flowIndex];

#if FRAGMENTATION >= 1		
		if ((ipHeader->ip_off & FRAGIDM_N) == MORE_FRAG_N 
#if FRAG_HLST_CRFT == 1
				|| sw_fnohead 
#endif
						) { // if 1. fragmented packet or mangled fragment		
#if FRAG_HLST_CRFT == 1		
			flow->status |= sw_fnohead;
			sw_fnohead = 0; // reset error state of fragmentation machine
#endif		
			numOfFragV4Packets++;
			globalWarn |= IPV4_FRAG;
			flow->status |= IPV4_FRAG;
			hashHelper.srcIP = ipHeader->ip_src; // flowCreate looked into reverse flow
			hashHelper.dstIP = ipHeader->ip_dst; // so set orig flow again
			if (flow->status & IPV4_FRAG_PENDING) {
				hashHelper.srcPort = flow->lastFragIPID;
        			if (hashTable_remove(fragPendMap, (char*) &hashHelper.srcIP) == HASHTABLE_ENTRY_NOT_FOUND) printf("remove IPID notfound \n");
				flow->status |= IPV4_FRAG_ERR;
			}
			hashHelper.srcPort = ipHeader->ip_id;
			fragPendIndex = hashTable_insert(fragPendMap, (char*)&hashHelper.srcIP);
        		if (fragPendIndex == HASHTABLE_ENTRY_NOT_FOUND) {
                                printf("frag insert failed!\n");
                                exit(3);
                        }
			fragPend[fragPendIndex] = flowIndex;
			flow->lastFragIPID = ipHeader->ip_id; 
			flow->status |= IPV4_FRAG_PENDING;  
		} else if ( flow->status & IPV4_FRAG_PENDING ) flow->status |= IPV4_FRAG_ERR;
	}
#endif //FRAGMENTATION

	flow->status |= packet->status;

	// update lru list
	updateLRUList(flow);

	flow->lastSeen = packet->pcapHeader->ts;
	

	// layer 4
	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginClaimLayer4Information; i++) {
		tranalyzer_plugins->pluginClaimLayer4Information[i](packet, flowIndex);
	}
}


inline unsigned long flowCreate(packet_t *packet) {
	uint32_t i;
	unsigned long flowIndex = hashTable_insert(mainHashMap, (char*)&hashHelper.srcIP);

	if (flowIndex == HASHTABLE_ENTRY_NOT_FOUND) return HASHTABLE_ENTRY_NOT_FOUND;
	flow_t *flow = &flows[flowIndex];
	flow->lruNextFlow = NULL;
	flow->lruPrevFlow = NULL;
	flow->lastSeen = packet->pcapHeader->ts;
	flow->firstSeen = packet->pcapHeader->ts;
	flow->layer4Protocol = packet->layer4Type;
	flow->flowIndex = flowIndex;
	flow->oppositeFlowIndex = HASHTABLE_ENTRY_NOT_FOUND;
	flow->srcIP = ipHeader->ip_src;
	flow->dstIP = ipHeader->ip_dst;
	flow->srcPort = srcPort;
	flow->dstPort = dstPort;
	flow->vlanID = innerVLANID;
	flow->timeout = FLOW_TIMEOUT; // fill with the standard value
	flow->status = 0;
	//flow->IPVer = 0x10; // binary 2⁴
	totalFlows++;


	// check whether the reverse flow exists and link both flows
	//struct in_addr tmpAddress = hashHelper.srcIP;
	//i = hashHelper.srcPort;
	hashHelper.srcIP = hashHelper.dstIP;
	hashHelper.dstIP = ipHeader->ip_src;
	hashHelper.srcPort = dstPort;
	hashHelper.dstPort = srcPort;
	
	if (!startTS) {
		startTStamp = flow->firstSeen;
		startTS = 1;
		strftime(timeBuf, MAX_TM_BUF, "%a %d %b %Y %X", localtime(&startTStamp.tv_sec));
		printf("Dump start: %ld.%06ld sec : %s.%06ld\n", (long int)startTStamp.tv_sec, (long int)startTStamp.tv_usec, timeBuf, (long int)startTStamp.tv_usec);
	 }

	unsigned long reverseFlowIndex = hashTable_lookup(mainHashMap, (char*)&hashHelper.srcIP);
	flow_t *revflow = &flows[reverseFlowIndex];
	
	if (reverseFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		if ( reverseFlowIndex != flowIndex) { // a flow must not be its own opposit
			// yes, reverse flow is in the hashTable
			flow->oppositeFlowIndex = reverseFlowIndex;
			revflow->oppositeFlowIndex = flowIndex;
			flow->findex = revflow->findex;

			if ( !(revflow->status & L3FLOWINVERT) ) flow->status = L3FLOWINVERT; 

		} else flow->status |= LANDATTACK;

	} else { 
		flow->findex = ++totalfIndex;
		if ( ( packet->layer4Type == 6 && ( ( *( (char*)packet->layer4Header + 13 ) & TH_SYN_ACK ) == TH_SYN_ACK ) ) || ( srcPort < 1024 && srcPort < dstPort ) ) flow->status = L3FLOWINVERT;
	}

	maxNumFlows++;
	if (maxNumFlows > maxNumFlowsPeak) maxNumFlowsPeak = maxNumFlows;

	// call tasklets functions
	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginOnFlowGenerated; i++) {
		tranalyzer_plugins->pluginOnFlowGenerated[i](packet, flowIndex);
	}

	return flowIndex;
}
#endif // IPV6_ACTIVATE == 0


#if IPV6_ACTIVATE == 1
inline void dissembleIPv6Packet(packet_t *packet) {
	int32_t packetLen;
	uint16_t l2Len; // uint32_t
	uint16_t i, l3Len;
	uint16_t ip6HDLen = 40;
	uint8_t nxtHdr, *n6Hdr = NULL;
#if FRAGMENTATION == 1
#if FRAG_HLST_CRFT == 1
	uint8_t sw_fnohead = 0;
#endif // FRAG_HLST_CRFT
#endif // FRAGMENTATION

	flow_t *flow;
	unsigned long flowIndex;
	ip6Header_t *ip6Header = (ip6Header_t*) packet->layer3Header;
	ip6FragHdr_t *ip6FragHdr = NULL;
	ip6OptHdr_t *ip6OptHdr = NULL;
	ip6RouteHdr_t *ip6RouteHdr = NULL;
	

	packet->layer3Type = ETHERTYPE_IPV6;
	packet->status |= L2_IPV6;
	numOfV6Packets++;
	
	if ( ip6Header->next_header == 44 ) {
		ip6FragHdr = (ip6FragHdr_t*)(++ip6Header); 
#if FRAGMENTATION == 0 
		// do not handle fragmented packets
		if ( ip6FragHdr->frag_off & FRAG6ID_N ) {
			globalWarn |= IPV4_FRAG;
			numOfFragV6Packets++;
			return; // fragmentation switch off: ignore fragmented packets except the 1. protocol header 
		}
#endif //FRAGMENTATION	
	}

	i = (uint16_t)((uint8_t*) packet->layer3Header - (uint8_t*) packet->layer2Header); // L2,VLAN length
	packet->snapL3Length = packet->snapL2Length - i; // L3 Packet length
	packet->l2HdrLen = i;
	l3Len = ntohs(ip6Header->payload_len) + 40;  // get IP packet length from IP header
	l2Len = l3Len + i;
 	packet->packetL2Length = l2Len; 
	bytesonWire += l2Len;   // estimate all Ethernet & IP bytes seen on wire
	
	if ( packet->snapL3Length < l3Len ) { // Layer3 snaplength too short or IP packet too short?
		packet->status |= L3SNAPLENGTH;
 		if (!(globalWarn & L3SNAPLENGTH)) { // Snap length warning
			globalWarn |= L3SNAPLENGTH;
			printf("Warning: snapL2Length: %d - snapL3Length: %d - IpLength in Header: %d\n", packet->snapL2Length, packet->snapL3Length, l3Len);
		} 
	} else if ( packet->snapL3Length > l3Len ) {
		packet->snapL2Length = l2Len; 
		packet->snapL3Length = l3Len;
	}

#if PACKETLENGTH == 0 
	packetLen = l2Len;
#else
	packetLen = l3Len;
#endif

	// -------------------- layer3 ------------------------

#if  PACKETLENGTH <= 1 

	packet->packetLength = packetLen; 
#endif // PACKETLENGTH <= 1 

	packet->layer4Type = nxtHdr = ip6Header->next_header; // set layer4Type already for global plugins such as protocolStatistics
	

	uint16_t l4HdrOff = 8;
	srcPort = 0;
	dstPort = 0;
	for ( i = 0; i < MAXHDRCNT; i++ ) {
	    packet->layer4Header = (const l4Header_t*) ((unsigned char*)ip6Header + ip6HDLen); // adjust header to the beginning of the encapsulated protocol
	    switch ( nxtHdr ) { 
		case 0: //options
		case 60:
			ip6OptHdr = (ip6OptHdr_t*)packet->layer4Header;
			nxtHdr = ip6OptHdr->next_header;
			ip6HDLen += (ip6OptHdr->len + 1)<<3; 
			packet->ip6OptHdr = ip6OptHdr;
			continue;
		case 58: // icmp
			l4HdrOff = 4;
			i = MAXHDRCNT;
			break;
		case 6: // tcp
			l4HdrOff = packet->layer4Header->tcpHeader.doff << 2; 
			if (l3Len < 40 ) packet->status |= L4HDRSHRTLEN;
			else {		
				srcPort = ntohs(packet->layer4Header->tcpHeader.source);
				dstPort = ntohs(packet->layer4Header->tcpHeader.dest);
			}
			i = MAXHDRCNT;
			break;
		case 17: // udp
			if (l3Len < 40 ) packet->status |= L4HDRSHRTLEN;	
			else {
				srcPort = ntohs(packet->layer4Header->udpHeader.source);
				dstPort = ntohs(packet->layer4Header->udpHeader.dest);
#if L2TP == 1
				uint16_t *l2TPPP, *l2TPH;
				l2TPPP = (uint16_t*) packet->layer4Header;
				if ( *l2TPPP == L2TP_PORT_N && *(l2TPPP+1) == L2TP_PORT_N) { // L2TP port 1701, 0xa506 hex network order
					packet->l2tpLayer3Hdr = packet->layer3Header;
					globalWarn |= L2_L2TP;
					packet->status |= L2_L2TP;
					l2TPPP += 4; // advance to L2TP 
					l2TPH = l2TPPP;
					packet->l2TPHdr = (l2tpv2Header_t*)l2TPPP;
					if ( (*l2TPH & (L2TP_TYP | L2TP_VER)) == L2TP_V2 ) { // if data then advance to l3 header, later version: supply l2tp, ppp parameters for plugins
						l2TPPP++;
						if ( *l2TPH & L2TP_LEN ) 
						l2TPPP++;
						l2TPPP += 2; // tunnel / session ID
						if ( *l2TPH & L2TP_SQN ) 
							l2TPPP += 2;
						if ( *l2TPH & L2TP_OFF ) 
  							l2TPPP += (ntohs(*l2TPPP)>>1) + 1;

//						if ( *l2TPPP ==  PPP_ADD_CTL ) { // HDLC ppp present 0xff03	
							packet->pppHdr = (pppHu_t*)l2TPPP; // save ppp header
							l2TPPP++;	 // advance HDLC ppp header add field 0xff03, following HDLC ppp encapsulated prot code
   							if (*l2TPPP == PPP_PROT_IP) { // PPP IP encapsulation?
								packet->status |= L2_PPP;
								packet->layer2Type = L2TP_V2;
								packet->layer3Header = (const l3Header_t*) (u_char*)(++l2TPPP);
								packet->layer3Type = ETHERTYPE_IP;
								ip6Header = (ip6Header_t*) packet->layer3Header;
							} else return; // only IP
/*						} else { // TODO: check for IPC etc 
							return;  
						} */
					} else return; // only data
				} // L2TP ?

#else // L2TP == 0
				if (srcPort == 1701 && dstPort == 1701) { // L2TP ports
					packet->layer3Type = L2TP_V2;
					packet->status |= L2_L2TP;
					globalWarn |= L2_L2TP;
				}
#endif // L2TP
			}
			i = MAXHDRCNT;
			break;
		case 43: // routing
			ip6RouteHdr = (ip6RouteHdr_t*)packet->layer4Header; 
			nxtHdr = ip6RouteHdr->next_header;
			ip6HDLen += (ip6RouteHdr->len + 1)<<3; 
			continue;
		case 44: // fragmentation
			ip6FragHdr = (ip6FragHdr_t*)packet->layer4Header; 
			numOfFragV6Packets++;
#if FRAGMENTATION == 0 
			// do not handle fragmented packets
			if ( ip6FragHdr->frag_off & FRAG6ID_N ) {
				globalWarn |= IPV4_FRAG;
				return; // fragmentation switch off: ignore fragmented packets except the 1. protocol header 
			}
#endif //FRAGMENTATION	
#if  ( PACKETLENGTH <= 1 && FRGIPPKTLENVIEW == 1 && FRAGMENTATION == 1 ) // IP packet view mode in case of fragmentation

			if (ip6FragHdr->frag_off & FRAG6ID_N ) packetLen -= ip6HDLen; // remove IP header len only from 2nd++ frag if whole packet statistical view is required: default
#endif // ( PACKETLENGTH <= 1 && FRGPKTLENVIEW == 1 && FRAGMENTATION == 1 )

			nxtHdr = ip6FragHdr->next_header;
			packet->ip6FragHdr = ip6FragHdr;
			ip6HDLen += 8; 
			continue;
		case 47:
#if GRE == 1
			uint32_t *grePPP, *greHD;
			if (ipHeader->ip_p == L3_GRE) {
				grePPP = (uint32_t*) packet->layer4Header;
				globalWarn |= L2_GRE;
				packet->status |= L2_GRE;
				greHD = grePPP++;
				packet->greHdr = (greHeader_t*)greHD;
				packet->greLayer3Hdr = packet->layer3Header;
				if (*greHD & GRE_CKSMn) grePPP++;
				if (*greHD & GRE_RTn) grePPP++;
				if (*greHD & GRE_KEYn) grePPP++;
				if (*greHD & GRE_SQn) grePPP++;
				if (*greHD & GRE_SSRn) grePPP++;
				if (*greHD & GRE_ACKn) grePPP++;
				if ((*greHD & GRE_PROTOn) == GRE_PPPn) {
					packet->status |= L2_PPP;
					if ( (*grePPP  & 0x000000ff) == GRE_PPP_CMPRSS ) {
						packet->status |= PPP_NRHD; // compressed, no readable header; info for later processing of flow
					}
					if ( ((greHeader_t*)greHD)->plength ) {
						
					} else {
						
					}
				} else { 
					packet->layer3Header = (const l3Header_t*) (u_char*)(grePPP);
					packet->layer3Type = (uint16_t)(*greHD & GRE_PROTOn);
					ipHeader = (ipHeader_t*)packet->layer3Header;
				}
			}
#else // GRE == 0
		
			if ( !(*(uint16_t*)packet->layer4Header & 0x000080f0) ) l4HdrOff = 8;
#endif // GRE 
			i = MAXHDRCNT;
			break;
		case 59:
			i = MAXHDRCNT;			
			break;
		case 89:
			l4HdrOff = 16;
			i = MAXHDRCNT;			
			break;
		case 115: // L2TPV3
			packet->layer3Type = L2TP_V3;
			i = MAXHDRCNT;			
			break;
		case 132:
			l4HdrOff = 12;
			if (l3Len <  40 ) packet->status |= L4HDRSHRTLEN;
			else {		
				srcPort = ntohs(packet->layer4Header->tcpHeader.source);
				dstPort = ntohs(packet->layer4Header->tcpHeader.dest);
			}
			i = MAXHDRCNT;			
			break;
		default: // all other protocols not implemented yet
			i = MAXHDRCNT;			
			break;
	    }
	}

	packet->layer4Type = nxtHdr; // set layer4Type already for global plugins such as protocolStatistics
	packet->l3HdrLen = ip6HDLen;

	// if desired calculate statistics with the IP protocol. Note that the data of the encapsulated protocol is still unknown.
	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginClaimLayer3Information; i++) {
		tranalyzer_plugins->pluginClaimLayer3Information[i](packet);
	}

	// -------------------- layer 4 ------------------------

#if PACKETLENGTH >= 2

	packetLen -= ip6HDLen;
#if PACKETLENGTH == 3 // sustract L4 header

	packetLen -= l4HdrOff;	
#endif // PACKETLENGTH == 3
	if (packetLen >= 0) packet->packetLength = packetLen;
	else { 
		packet->packetLength = 0;
		packet->status |= L4HDRSHRTLEN;	
	}
#endif // PACKETLENGTH >= 2


	// -------------------- layer 7 info for packet structure ------------------------
	
	packet->layer7Header = (uint8_t*)packet->layer4Header + l4HdrOff;
	if ( packet->snapL3Length >= l3Len ) { 					// L3 length not snapped
                packet->snapL7Length =  l3Len - ip6HDLen - l4HdrOff; 	// Protocol L3/4 Hdr lengths are valid 
        } else { 									// L3 length snapped so calculate real hdr L7 length
                packet->snapL7Length = (uint16_t)(packet->layer7Header - (uint8_t*)packet->layer3Header); // offset between L3 and L7
                packet->snapL7Length = packet->snapL3Length - packet->snapL7Length; 	// real L7 length
        }

// source and destination port of a layer 4 header
	
#if AGGREGATIONFLAG & SUBNET
//	if (subnet_table) {
//		ipHeader->ip_src.s_addr = subnet_testP(subnet_table, flow->srcIP.s_addr); // subnet test src ip
//		ipHeader->ip_dst.s_addr = subnet_testP(subnet_table, flow->dstIP.s_addr); // subnet test dst ip
//	}
#else // AGGREGATIONFLAG & SUBNET == 0

#if AGGREGATIONFLAG & (SRCIP | DSTIP)
	//for (i = 0; i < 2; i++) {
#if AGGREGATIONFLAG & SRCIP
	//	ip6Header->ip_src.IPv6L[i] &= SRCIP6MASKn[i];
#endif // AGGREGATIONFLAG & SRCIP
#if AGGREGATIONFLAG & DSTIP
	//	ip6Header->ip_dst.IPv6L[i] &= DSTIP6MASKn[i];
#endif // AGGREGATIONFLAG & DSTIP

#endif // AGGREGATIONFLAG & SUBNET
	//}
#endif // AGGREGATIONFLAG & (SRCIP | DSTIP)
 
	hashHelper.IPVer = 0x40; // binary 2⁶
	hashHelper.vlanID = innerVLANID;
	hashHelper.srcIP = ip6Header->ip_src;
	hashHelper.dstIP = ip6Header->ip_dst;
	
#if FRAGMENTATION == 1

	unsigned long fragPendIndex;	
	
	if ( ip6FragHdr && (ip6FragHdr->frag_off & FRAG6ID_N) ) { // 2nd++ fragmented packet
	
		hashHelper.fragID6 = ip6FragHdr->id;
		fragPendIndex = hashTable_lookup(fragPendMap, (char*)&hashHelper.fragID6);
	
		if (fragPendIndex == HASHTABLE_ENTRY_NOT_FOUND) { // probably missed  1. frag packet or packet mangling 
			globalWarn |= (IPV4_FRAG | IPV4_FRAG_HDSEQ_ERR);
#if FRAG_ERROR_DUMP == 1 
			printf("1.frag not found @ %ld.%ld %d %d %d %d - 0x%08X 0x%04X\n", packet->pcapHeader->ts.tv_sec, packet->pcapHeader->ts.tv_usec, hashHelper.vlanID, srcIP, ntohs(packet->layer4Header->tcpHeader.source), dstIP, ntohs(packet->layer4Header->tcpHeader.dest), packet->layer4Type, ntohs(hashHelper.fragID6), ntohs(ip6FragHdr->frag_off) );
#endif // FRAG_ERROR_DUMP

#if FRAG_HLST_CRFT == 1
			sw_fnohead = IPV4_FRAG_HDSEQ_ERR;
			goto create_packetF; // we dont know the flow, but create one anyway, because might be interesting crafted packet
#else 
			return; // we don't know the flow, so ignore packet
#endif // FRAG_HLST_LCFT
		} else { 
			numOfFragV6Packets++;
			flowIndex = fragPend[fragPendIndex];
			flow = &flows[flowIndex];
			if (!(ip6FragHdr->frag_off & MORE_FRAG6_N)) { // remove packet from frag queue when last fragment received
				if (hashTable_remove(fragPendMap, (char*) &hashHelper.fragID6) == HASHTABLE_ENTRY_NOT_FOUND) printf("- fragPend remove failed -"); 
				if (flow->status & IPV4_FRAG_PENDING) flow->status &= ~IPV4_FRAG_PENDING; 
			}	
		}

	} else { // not fragmented or 1. fragmented packet

#endif //FRAGMENTATION == 1

create_packetF:

#if AGGREGATIONFLAG & SRCPORT
		if (srcPort >= SRCPORTLW && srcPort <= SRCPORTHW ) srcPort = 1; else srcPort = 0;
#endif
		hashHelper.srcPort = srcPort; 
#if AGGREGATIONFLAG & DSTPORT
		if (dstPort >= DSTPORTLW && dstPort <= DSTPORTHW ) dstPort = 1; else dstPort = 0;
#endif
		hashHelper.dstPort = dstPort;
#if AGGREGATIONFLAG & L4PROT
		packet->layer4Type = 0;
#endif
		hashHelper.layer4Protocol = packet->layer4Type;

		flowIndex = hashTable_lookup(mainHashMap, (char*)&hashHelper.srcIP);
		if (flowIndex == HASHTABLE_ENTRY_NOT_FOUND) {
			if (exitPriority > 0) return;
			flowIndex = flowCreate6(packet);
		}

		flow = &flows[flowIndex];

#if FRAGMENTATION >= 1		
		if ( (ip6FragHdr && (ip6FragHdr->frag_off & MORE_FRAG6_N) == MORE_FRAG6_N)  // more flag set
#if FRAG_HLST_CRFT == 1
				|| sw_fnohead 
#endif
						) { // if 1. fragmented packet or mangled fragment		
#if FRAG_HLST_CRFT == 1		
			flow->status |= sw_fnohead;
			sw_fnohead = 0; // reset error state of fragmentation machine
#endif		
			numOfFragV6Packets++;
			globalWarn |= IPV4_FRAG;
			flow->status |= IPV4_FRAG;
			hashHelper.srcIP = ip6Header->ip_src; // flowCreate looked into reverse flow
			hashHelper.dstIP = ip6Header->ip_dst; // so set orig flow again
			if (flow->status & IPV4_FRAG_PENDING) {
				hashHelper.fragID6 = flow->lastFragIPID;
        			if (hashTable_remove(fragPendMap, (char*) &hashHelper.srcIP) == HASHTABLE_ENTRY_NOT_FOUND) printf("remove IPID notfound \n");
				flow->status |= IPV4_FRAG_ERR;
			}
			hashHelper.fragID6 = ip6FragHdr->id;
			fragPendIndex = hashTable_insert(fragPendMap, (char*)&hashHelper.srcIP);
        		if (fragPendIndex == HASHTABLE_ENTRY_NOT_FOUND) {
                                printf("frag insert failed!\n");
                                exit(3);
                        }
			fragPend[fragPendIndex] = flowIndex;
			flow->lastFragIPID = ip6FragHdr->id; 
			flow->status |= IPV4_FRAG_PENDING;  
		} else if ( flow->status & IPV4_FRAG_PENDING ) flow->status |= IPV4_FRAG_ERR;
	}
#endif //FRAGMENTATION

	flow->status |= packet->status;

	// pdate lru list
	updateLRUList(flow);

	flow->lastSeen = packet->pcapHeader->ts;
	

	// layer 4
	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginClaimLayer4Information; i++) {
		tranalyzer_plugins->pluginClaimLayer4Information[i](packet, flowIndex);
	}


}

inline unsigned long flowCreate6(packet_t *packet) {
	uint16_t i;
	unsigned long flowIndex = hashTable_insert(mainHashMap, (char*)&hashHelper.srcIP);

	if (flowIndex == HASHTABLE_ENTRY_NOT_FOUND) return HASHTABLE_ENTRY_NOT_FOUND;
	
	ip6Header_t *ip6Header = (ip6Header_t*)packet->layer3Header;
	flow_t *flow = &flows[flowIndex];
	flow->lruNextFlow = NULL;
	flow->lruPrevFlow = NULL;
	flow->lastSeen = packet->pcapHeader->ts;
	flow->firstSeen = packet->pcapHeader->ts;
	flow->layer4Protocol = packet->layer4Type;
	flow->flowIndex = flowIndex;
	flow->oppositeFlowIndex = HASHTABLE_ENTRY_NOT_FOUND;
	flow->srcIP = ip6Header->ip_src;
	flow->dstIP = ip6Header->ip_dst;
	flow->srcPort = srcPort;
	flow->dstPort = dstPort;
	flow->vlanID = innerVLANID;
	flow->timeout = FLOW_TIMEOUT; // fill with the standard value
	flow->status = L2_IPV6;
	flow->IPVer = 0x40; // binary 2⁶
	totalFlows++;


	// check whether the reverse flow exists and link both flows
	hashHelper.srcIP = hashHelper.dstIP;
	hashHelper.dstIP = ip6Header->ip_src;
	hashHelper.srcPort = dstPort;
	hashHelper.dstPort = srcPort;
	
	if (!startTS) {
		startTStamp = flow->firstSeen;
		startTS = 1;
		strftime(timeBuf, MAX_TM_BUF, "%a %d %b %Y %X", localtime(&startTStamp.tv_sec));
		printf("Dump start: %ld.%06ld sec : %s.%06ld\n", startTStamp.tv_sec, startTStamp.tv_usec, timeBuf, startTStamp.tv_usec);
	 }

	unsigned long reverseFlowIndex = hashTable_lookup(mainHashMap, (char*)&hashHelper.srcIP);
	flow_t *revflow = &flows[reverseFlowIndex];
	/*
	 * A note to the following line:
	 * The term "reverseFlowIndex != flowIndex" indicates that we don't allow a flow to be its own opposite
	 */
	if (reverseFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		if ( reverseFlowIndex != flowIndex) {
			// yes, reverse flow is in the hashTable
			flow->oppositeFlowIndex = reverseFlowIndex;
			revflow->oppositeFlowIndex = flowIndex;
			flow->findex = revflow->findex;

			if ( !(revflow->status & L3FLOWINVERT) )
				 flow->status = L3FLOWINVERT; 

		} else flow->status |= LANDATTACK;

	} else { 
		flow->findex = ++totalfIndex;
		if ( ( packet->layer4Type == 6 && ( ( *( (char*)packet->layer4Header + 13 ) & TH_SYN_ACK ) == TH_SYN_ACK ) ) || ( srcPort < 1024 && srcPort < dstPort ) ) 
			flow->status = L3FLOWINVERT;
	}


	// call tasklets functions
	for (i = 0; i < tranalyzer_plugins->num_of_func_pluginOnFlowGenerated; i++) {
		tranalyzer_plugins->pluginOnFlowGenerated[i](packet, flowIndex);
	}

	return flowIndex;
}
#endif // IPV6_ACTIVATE == 1



inline void updateLRUList(flow_t *flowToUpdate) {
	if (lruHead.lruNextFlow != flowToUpdate) {
		// we have work to do, move flow to the front (head)

		// remove flow from its current position
		if (flowToUpdate->lruPrevFlow) flowToUpdate->lruPrevFlow->lruNextFlow = flowToUpdate->lruNextFlow;
		if (flowToUpdate->lruNextFlow) flowToUpdate->lruNextFlow->lruPrevFlow = flowToUpdate->lruPrevFlow;

		// append it at the head of the lru list
		flowToUpdate->lruNextFlow = lruHead.lruNextFlow;
		lruHead.lruNextFlow->lruPrevFlow = flowToUpdate;
		lruHead.lruNextFlow = flowToUpdate;
		flowToUpdate->lruPrevFlow = &lruHead;
	}
}

