/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * RedBlackTree.c
 */

#include "redBlackTree.h"
#include <string.h>
#include <inttypes.h>

void insert_case1(redBlackNode_t *n);
void insert_case2(redBlackNode_t *n);
void insert_case3(redBlackNode_t *n);
void insert_case4(redBlackNode_t *n);
void insert_case5(redBlackNode_t *n);
redBlackNode_t* grandparent(redBlackNode_t *node);
redBlackNode_t* uncle(redBlackNode_t *node);
unsigned int rekursiveGetDepth(redBlackNode_t *tree, unsigned int currentDepth);

/*
 * The following functions handles the rebalancing of the tree, if necessary.
 * Shamelessly stolen from the german Wikipedia:
 * http://de.wikipedia.org/wiki/Rot-Schwarz-Baum
 * TODO: improve speed
 */
void rotate_left(redBlackNode_t *node) {
#if DEBUG > 3
		fprintf(stdout,"rotate LEFT\n:");
		if (!node->parent) fprintf(stdout,"-> parent node is NULL");
#endif
	if (node->parent) {
		node->parent->rightChild = node->leftChild;
		if (node->leftChild) node->leftChild->parent = node->parent;

		node->leftChild = node->parent;
		node->parent = node->parent->parent;

		node->leftChild->parent = node;
		if (node->parent) {
			if (node->parent->leftChild == node->leftChild) node->parent->leftChild = node;
			else node->parent->rightChild = node;
		}
	}
}

void rotate_right(redBlackNode_t *node) {
#if DEBUG > 3
		fprintf(stdout,"rotate RIGHT\n");
		if (!node->parent) fprintf(stdout,"-> parent node is NULL");
#endif
	if (node->parent) {
		node->parent->leftChild = node->rightChild;
		if (node->rightChild) node->rightChild->parent = node->parent;

		node->rightChild = node->parent;
		node->parent = node->parent->parent;

		node->rightChild->parent = node;
		if (node->parent) {
			if (node->parent->leftChild == node->rightChild) node->parent->leftChild = node;
			else node->parent->rightChild = node;
		}
	}
}

/* case 1: Node is root */
void insert_case1(redBlackNode_t *n) {
	if (n->parent == NULL) n->color = NODE_COLOR_BLACK;
	else insert_case2(n);
 }

void insert_case2(redBlackNode_t *n) {
	if (n->parent->color == NODE_COLOR_BLACK) return; // tree is still OK
	else if (grandparent(n)) insert_case3(n);
 }

void insert_case3(redBlackNode_t *n) {
	if (uncle(n) != NULL && uncle(n)->color == NODE_COLOR_RED) {
		n->parent->color = NODE_COLOR_BLACK;
		uncle(n)->color = NODE_COLOR_BLACK;
		grandparent(n)->color = NODE_COLOR_RED;
		insert_case1(grandparent(n));
	} else insert_case4(n);
 }

void insert_case4(redBlackNode_t *n) {
	if (n == n->parent->rightChild && n->parent == grandparent(n)->leftChild) {
		rotate_left(n);
		// rotate_left(n->parent);
		n = n->leftChild;
	} else if (n == n->parent->leftChild && n->parent == grandparent(n)->rightChild) {
		rotate_right(n);
		// rotate_right(n->parent);
		n = n->rightChild;
	}
	if (n) insert_case5(n);
}

void insert_case5(redBlackNode_t *n) {
	n->parent->color = NODE_COLOR_BLACK;
	grandparent(n)->color = NODE_COLOR_RED;
	if (n == n->parent->leftChild && n->parent == grandparent(n)->leftChild) {
		rotate_right(n->parent);
		// rotate_right(grandparent(n));
	} else {
		rotate_left(n->parent);
		// rotate_left(grandparent(n));
	}
}

redBlackNode_t* grandparent(redBlackNode_t *node) {
	if (node->parent != NULL) return node->parent->parent;
	else return NULL;
}

redBlackNode_t* uncle(redBlackNode_t *node) {
	if (grandparent(node) != NULL) {
		if (node->parent == grandparent(node)->leftChild) return grandparent(node)->rightChild;
		else return grandparent(node)->leftChild;
	} else return NULL;
}

redBlackTreeNodePool_t* redBlackTree_initTreeNodePool(uint32_t rb_tree_nodepool_size) {
	redBlackTreeNodePool_t *pool;
	if ((pool = malloc(sizeof(redBlackTreeNodePool_t))) == NULL) return NULL;
	if ((pool->nodePool = malloc(rb_tree_nodepool_size * sizeof(redBlackNode_t))) == NULL) {
		free(pool);
		return NULL;
	}

	memset(pool->nodePool, '\0', rb_tree_nodepool_size * sizeof(redBlackNode_t));
	pool->freeNode = &pool->nodePool[0];
	pool->numOfFreeNodes = rb_tree_nodepool_size;
	pool->pool_size = rb_tree_nodepool_size;

	/* set the rightChild pointer of a node to the next node of the array to
	 * construct the free node list*/
	for (treeCounter = 0; treeCounter < rb_tree_nodepool_size - 1; treeCounter++) {
		pool->nodePool[treeCounter].rightChild = &pool->nodePool[treeCounter + 1];
	}

	return pool;
}

redBlackNode_t* redBlackTree_search_insert(redBlackNode_t *tree, redBlackTreeNodePool_t *treeNodePool, int32_t value, unsigned short doInsert, uint8_t *entryExists) {
	if (!tree && (doInsert == NOT_INSERT_NODE)) {
		*entryExists = ENTRY_NOT_EXISTS;
#if DEBUG > 1
		fprintf(stdout, "Tree does not exist an nothing was created\n");
#endif
		return NULL;
	}

	/* look for a entry with the given value */
	currentNode = tree;
	previousNode = NULL;
	while (currentNode) {
		if (currentNode->value == value) {
			*entryExists = ENTRY_EXISTS;
			return currentNode;
		}

		previousNode = currentNode;
		if (value < currentNode->value) currentNode = currentNode->leftChild;
		else currentNode = currentNode->rightChild;
	}

	/* No node with the value was found */
	*entryExists = ENTRY_NOT_EXISTS;

	if (doInsert == INSERT_NODE) {
#if DEBUG > 3
		fprintf(stdout, "creating node\n");
#endif
		if (treeNodePool->numOfFreeNodes == 0) {
			fprintf(stdout, "ERROR: No free nodes in treeNodePol !!!");
			return NULL;
		}
		/* So generate a new one. Pick a free node out of the free list */
		currentNode = treeNodePool->freeNode;
		treeNodePool->freeNode = treeNodePool->freeNode->rightChild;

		// fill it with the right values
		currentNode->value = value;
		currentNode->parent = previousNode;
		currentNode->leftChild = NULL;
		currentNode->rightChild = NULL;
		currentNode->color = NODE_COLOR_RED;

		// if there is no parent, then this is a new tree
		if (previousNode == NULL) {
			// do nothing;
		}
		/* else connect it to the correct side of the parent */
		else if (value < previousNode->value) previousNode->leftChild = currentNode;
		else previousNode->rightChild = currentNode;

		// decrease the nodePool's freeNode counter
		treeNodePool->numOfFreeNodes--;

		// Test if a reconfiguration of the tree is necessary
#if DEBUG > 3
		fprintf(stdout,"Tree BEFORE:\n:");
		if (tree) redBlackTree_printTree(tree, 5);
		else redBlackTree_printTree(currentNode, 5);
#endif

#if REDBLACKTREE_ENABLE_ROTATION != 0
			insert_case1(currentNode);
#endif

#if DEBUG > 3
		previousNode = currentNode;
		while (previousNode->parent) previousNode = previousNode->parent;
		tree = previousNode;
		fprintf(stdout,"Tree AFTER:\n:");
		if (tree) redBlackTree_printTree(tree, 5);
		else redBlackTree_printTree(currentNode, 5);
#endif
		return currentNode;
	}
	return NULL;
}

int redBlackTree_destroy(redBlackNode_t *tree, redBlackTreeNodePool_t *treeNodePool) {
	if (!tree) return 0;
	if (tree->parent != NULL) return 1; // tree wasn't destroyed because it was not the root node. TODO: switch to delete function
	// Note: This destroying function is programmed iteratevly (not recursive)
	// because of speed

	currentNode = tree;
	while (currentNode) {
		if (currentNode->leftChild != NULL) {
			// go deeper in the tree;
			currentNode = currentNode->leftChild;

		} else if (currentNode->rightChild != NULL) {
			// go deeper in the tree;
			currentNode = currentNode->rightChild;
		} else {
			// current node leaves are NULL so put the current Node back in the
			// free list
			currentNode->rightChild = treeNodePool->freeNode;
			treeNodePool->freeNode = currentNode;

			// increase the free list counter
			treeNodePool->numOfFreeNodes++;

			// check if the currentNode equals the tree root and return 0 if
			// this is true
			if (currentNode == tree) return 0;

			// disconnect the parent left- or rightChild pointer from the
			// currentNode and set the currentNode to its parent
			if (currentNode->parent->leftChild == currentNode) {
				currentNode = currentNode->parent;
				currentNode->leftChild->parent = NULL;
				currentNode->leftChild = NULL;
			} else {
				currentNode = currentNode->parent;
				currentNode->rightChild->parent = NULL;
				currentNode->rightChild = NULL;
			}
		}
	}
	return 1;
}

/*
 * PLEASE NOTE:
 * This function is NOT designed to be fast, it should only be used to perform
 * a debug output!!!
 */

void redBlackTree_printTree(redBlackNode_t *tree, unsigned long maxDepth) {
	fprintf(stdout, "Red-Black tree printing: ----------------------------------\n");
	// first get max depth of tree
	unsigned int treeDepth = rekursiveGetDepth(tree, 0);
	fprintf(stdout, "Tree depth is %d\n", treeDepth);
	if (treeDepth > maxDepth) {
		fprintf(stdout, "NOTE: Tree output is pruned at height %lu\n", maxDepth);
		treeDepth = maxDepth;
	}

	// initialize a queue
	redBlackTreeNodeQueue_t queue;
	redBlackTreeNodeQueueBin_t *currentQueueBin = malloc(sizeof(redBlackTreeNodeQueueBin_t));
	redBlackTreeNodeQueueBin_t *childQueueBin;
	currentQueueBin->node = tree;
	currentQueueBin->nextInQueue = NULL;
	currentQueueBin->level = 1;
	currentQueueBin->printPos = 0;
	queue.queueStart = currentQueueBin;
	queue.queueEnd = currentQueueBin;

	// Determine printing pos
	int i;
	for (i = 1; i < treeDepth; i++) {
		currentQueueBin->printPos = 2 * currentQueueBin->printPos + 3;
	}

	// start processing
	int currentLevel = 0;
	int currentPrintingPos = 0;
	while (currentQueueBin) {
		// reset printingPos if tree level changed, also new line
		if (currentLevel < currentQueueBin->level) {
			currentLevel = currentQueueBin->level;
			currentPrintingPos = 0;
			fprintf(stdout, "\n");
		}

		// skip to next printing pos
		for (i = currentPrintingPos; i < currentQueueBin->printPos; i++) {
			currentPrintingPos++;
			fprintf(stdout," ");
		}

		// print out current bin
		fprintf(stdout, "%05"PRId32"", currentQueueBin->node->value);
		currentPrintingPos = currentPrintingPos + 5; // TODO: What if output of previous line is longer than 5 ?

		// put left and right child of node into queue
		if ((currentLevel < treeDepth) && currentQueueBin->node->leftChild) {
			childQueueBin = malloc(sizeof(redBlackTreeNodeQueueBin_t));
			childQueueBin->node = currentQueueBin->node->leftChild;
			childQueueBin->level = currentQueueBin->level + 1;

			// determine difference between current printPos and printPos of child
			int j = 0;
			for (i = currentLevel; i < treeDepth; i++) {
				j = 2 * j + 3;
			}

			int k = 0;
			for (i = currentLevel + 1; i < treeDepth; i++) {
				k = 2 * k + 3;
			}
			int offset = j - k;

			childQueueBin->printPos = currentQueueBin->printPos - offset;
			childQueueBin->nextInQueue = NULL;
			queue.queueEnd->nextInQueue = childQueueBin;
			queue.queueEnd = childQueueBin;
		}

		if ((currentLevel < treeDepth) && currentQueueBin->node->rightChild) {
			childQueueBin = malloc(sizeof(redBlackTreeNodeQueueBin_t));
			childQueueBin->node = currentQueueBin->node->rightChild;
			childQueueBin->level = currentQueueBin->level + 1;

			// determine difference between current printPos and printPos of child
			int j = 0;
			for (i = currentLevel; i < treeDepth; i++) {
				j = 2 * j + 3;
			}

			int k = 0;
			for (i = currentLevel + 1; i < treeDepth; i++) {
				k = 2 * k + 3;
			}
			int offset = j - k;

			childQueueBin->printPos = currentQueueBin->printPos + offset;
			childQueueBin->nextInQueue = NULL;
			queue.queueEnd->nextInQueue = childQueueBin;
			queue.queueEnd = childQueueBin;
		}

		// currentQueueBin isn't needed anymore, so free it
		queue.queueStart = queue.queueStart->nextInQueue;
		free(currentQueueBin);
		// point to next entry
		currentQueueBin = queue.queueStart;
	}
	fprintf(stdout, "\n\n-----------------------------------------------------------\n\n");
}

unsigned int rekursiveGetDepth(redBlackNode_t *tree, unsigned int currentDepth) {
	unsigned int maxDepth = currentDepth + 1, tempDepth = 0;
	if (tree->leftChild) {
		tempDepth = rekursiveGetDepth(tree->leftChild, currentDepth + 1);
		if (tempDepth > maxDepth) maxDepth = tempDepth;
	}
	if (tree->rightChild) {
		tempDepth = rekursiveGetDepth(tree->rightChild, currentDepth + 1);
		if (tempDepth > maxDepth) maxDepth = tempDepth;
	}

	return maxDepth;
}


