/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * RedBlackTree.h
 * implements a Red-Black tree.
 * This implementation provides the construction of multiple RB trees out of a
 * "pool" of preallocated tree nodes.
 */

#ifndef REDBLACKTREE_H_
#define REDBLACKTREE_H_

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>

/* local includes */
#include "tranalyzer.h"

#define REDBLACKTREE_ENABLE_ROTATION 0 // a value of greater than zero enables the Red-Black Tree feature of rotating an unbalanced tree
#define NODE_COLOR_RED 0
#define NODE_COLOR_BLACK 1
#define INSERT_NODE 1
#define NOT_INSERT_NODE 0
#define ENTRY_EXISTS 1
#define ENTRY_NOT_EXISTS 0

typedef struct redBlackNode_s {
	int32_t value; // the value of a node. 4 bytes
	struct redBlackNode_s *parent; // the parent node of a node. 4 bytes
	struct redBlackNode_s *leftChild; // the left child node of a node. 4
	struct redBlackNode_s *rightChild; // the right child node of a node. 4
	// void *userObject; // a user-defined object, for example an additional value. 4
	unsigned short color; // a value of 0 means, the node is red, 1 means, it is black. 2 bytes
} redBlackNode_t; // 22 bytes -> 24 allocated

typedef struct redBlackTreeNodePool_s {
	redBlackNode_t *nodePool; // a pointer to an array of nodes. 4 bytes
	redBlackNode_t *freeNode; // a pointer to the first free Node. 4
	unsigned long numOfFreeNodes; // the number of free nodes in this treeNodePool. 4
	uint32_t pool_size;
} redBlackTreeNodePool_t; // 12 bytes

typedef struct redBlackTreeNodeQueueBin {
	redBlackNode_t *node; // 4
	unsigned int level; // mostly 4
	struct redBlackTreeNodeQueueBin *nextInQueue; // 4
	unsigned short printPos; // 2
} redBlackTreeNodeQueueBin_t; // 14 bytes -> 16 allocated

/* a queue for performing level-order tree traversing */
typedef struct redBlackTreeNodeQueue {
	redBlackTreeNodeQueueBin_t *queueStart; // 4
	redBlackTreeNodeQueueBin_t *queueEnd; // 4
} redBlackTreeNodeQueue_t; // 8 bytes

/* commonly used variables */
redBlackNode_t *currentNode;
redBlackNode_t *previousNode;
unsigned long treeCounter;

/*
 * initializes the redBlackTreeNodePool treeNodePool;
 */
redBlackTreeNodePool_t* redBlackTree_initTreeNodePool(uint32_t rb_tree_nodepool_size);

/*
 * searches a value in the tree and - if the doInsert flag is set - inserts a
 * new node out of the treenode pool into the tree. If the value already exists,
 * the entryExists pointer is set ENTRY_EXISTS.
 *
 * PLEASE NOTE: If the root changed (which can happen because of the rotation in
 * Red-Black trees, you may need to update your the tree pointer. This is NOT
 * done in this function!!! An example code snipped could look like:
 *
 * while (tree->parent) tree = tree->parent;
 *
 * with tree as the tree pointer
 */
redBlackNode_t* redBlackTree_search_insert(redBlackNode_t *tree, redBlackTreeNodePool_t *treeNodePool, int32_t value, unsigned short doInsert, uint8_t *entryExists);

/* destroys a tree and puts its nodes back into the free node list */
int redBlackTree_destroy(redBlackNode_t *tree, redBlackTreeNodePool_t *treeNodePool);

/* prints a whole tree in nice ascii art :) */
void redBlackTree_printTree(redBlackNode_t *tree, unsigned long maxDepth);

/* TODO: define / implement delete function */

#endif /* REDBLACKTREE_H_ */
